# Filters added to this controller apply to all controllers in the application.
# Likewise, all the methods added will be available for all controllers.


class ApplicationController < ActionController::Base
  # helper :all # include all helpers, all the time
  protect_from_forgery # See ActionController::RequestForgeryProtection for details
  include AuthenticatedSystem
  include AsperaController::ApplicationController
  NAME_IDENTIFIER_FORMAT_UNSPECIFIED = "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified"
  SKIP_HTML_SANITIZATION = []

  helper :form, :navigation, :job_and_step_link

  before_filter :disable_browser_cache
  #AC-808
  before_filter :hostname_authenticity
  before_filter :license_check
  before_filter :login_required
  before_filter :locked_account
  before_filter :check_expiry_time
  before_filter :check_concurrent_login
  before_filter :active_user
  before_filter :santize_user_inputs
  before_filter :password_expired
  before_filter :set_user_time_zone

  filter_parameter_logging :password, :passphrase, :entitlement_node_password, :v3_node_password

  rescue_from WillPaginate::InvalidPage, :with => :invalid_page if RAILS_ENV=='production'
  rescue_from  ActiveRecord::RecordNotFound, :with => :invalid_page if RAILS_ENV=='production'
  rescue_from "NodeConfiguration::AsconfiguratorError", :with => :handle_config_exception

  protected

  #AC-808 Start
  # Validate host from Request Referer, send back 404 for unknown host
  def hostname_authenticity
    requestreferer = request.referer
    if requestreferer.present?
      host = URI.parse(requestreferer).host
      unless accepted_hosts.include?(host)
        render :file => "#{RAILS_ROOT}/public/404.html",  :status => 404
      end
    end
  end

  # Fetch Accepted Hosts from Console.yml
  def accepted_hosts
    accepted_hosts ||= (CONSOLE_CONFIG['AcceptedHosts'] || []) << '127.0.0.1' <<  'localhost'
  end
  #AC-808 End

  def license_edit_only?
    return !!session[:license_edit_only]
  end

  def license_check
    session[:license_edit_only] = !Utils.licensed?
  end

  def invalid_page
    flash[:error] = "Invalid page requested"
    redirect_to(error_url)
  end

  # not logged in and not on login/logout pages
  def active_user
    if !logged_in? && controller_name != 'sessions'
      logout and return
    end
  end
  
  # Force user to either update password or log out
  # License update takes precedence
  def password_expired
    if !license_edit_only?
      current_path = request.request_uri
      if logged_in? && current_user.password_expired?
        if current_path != edit_password_path && current_path != logout_path
          flash[:error] = "Your password has expired. Please enter a new password." unless flash[:error].present?
          redirect_to edit_password_path
        end
      end
    end
  end

  def logout
    was_saml_user = current_user && current_user.saml_remote?
    reset_session
    redirect_to_login(was_saml_user)
  end

  def redirect_to_login(was_saml_user=false)
   if request.xhr?
     session[:return_to] = request.referer
     render :js => %|document.location.href = "#{(ConsoleConfig::SamlConfig.enable_saml && was_saml_user) ? saml_login_path : login_path(:local => 'true')}";|
   else
     if request.get?
       session[:return_to] = request.request_uri
     else
       session[:return_to] = root_path #request.referer
     end

     if ConsoleConfig::SamlConfig.enable_saml && was_saml_user
       redirect_to saml_login_path
     else
       redirect_to login_path(:local => 'true')
     end
   end
  end

  def login_success(current_user, options={})
    return_to = session[:return_to]
    reset_session
    session[:return_to] = return_to
    current_user.process_login(options[:remote_ip])
    session[:user_id] = current_user.id
    session[:login_uuid] = current_user.login_uuid
    reset_session_expiration
    if !options[:license_edit_only]
      redirect_back_or_default(options[:return_to] || url_for(:controller => 'dashboard', :action => 'show'))
    end
  end

  # Returns +true+ if user is logged in and is an admin.
  #
  # This is a helper method available in Views and Helpers.
  def admin?
    logged_in? && current_user.admin?
  end
  helper_method :admin?

  # If user is not an admin?, she is redirected to "/" with an error message.
  def admin_filter
    unless admin?
      flash[:error] = "Not authorized"
      redirect_to(error_url) and return
    end
  end

  def santize_user_inputs
    return if !params.has_key?("commit") && !(params.has_key?("action") && (["save", "update", "save_smart"].include? params["action"]))
    
    sanitize_params(params)
  end

  def sanitize_params(hash)
    hash.each do |key, value|
      next if (self.class::SKIP_HTML_SANITIZATION && self.class::SKIP_HTML_SANITIZATION.include?(key)) || value.nil?
      
      if value.is_a? Hash
        sanitize_params(value)
      elsif value.is_a?(Array) && value[0].is_a?(Hash)
        sanitize_params(value[0])
      elsif value.is_a? String
        hash[key] = sanitize(value)
      elsif [Tempfile, File].member? value.class
        value.original_path = sanitize(value.original_path)
      end
      
    end
  end

  private

  # whether to use HTTP basic auth (API)
  def api_request?
    false
  end

  def basic_auth_challenge
    return unless api_request?
    # Challenge service requests that don't have HTTP-Basic auth headers
    if request.headers['HTTP_AUTHORIZATION'].nil?
      headers['WWW-Authenticate'] = "Basic 'REST API'"
      render_api_error(:unauthorized)
      return
    end

    # Return 401 if authentication failed
    if current_user.nil?
      render_api_error(:unauthorized)
    end
  end

  def render_api_error(status, message = nil)
    response = {
        :error => {
        :code => status_code(status),
        :user_message => message || status_message(status) || ""
        }
    }
    render :json => response, :status => status
  end

  def status_code(status)
    status.is_a?(Fixnum) ? status : Rack::Utils.status_code(status)
  end

  def status_message(status)
    Rack::Utils::HTTP_STATUS_CODES[status_code(status)]
  end

  def basic_auth_user
    found_user = authenticate_with_http_basic do |username, password|
      user = authenticate(username, password, request.remote_ip)
      if user
        if user.active? && user.licensed?
          reset_session
          user.process_login(request.remote_ip)
          session[:user_id] = user.id
          session[:login_uuid] = user.login_uuid
          reset_session_expiration
          user
        else
          nil
        end
      else
        nil
      end
    end
    if found_user.nil? || !found_user.errors.empty?
      nil
    else
      found_user
    end
  end

  # convert array of elements into javascript associative array {'object.id': 'object',...}
  def json_hash(array, options={})
    elements = array.collect{|element|
      %Q["#{element.id}":#{element.to_json(options)}]
    }
    "{#{elements.join(",")}}"
  end

  def javascript_error_display(error_string_or_array)
    render(:update) { |page|
      message = ([error_string_or_array].flatten.map {|e| escape_javascript(e)}).join("\\n")
      page << "alert('#{message}')"
    } and return false
  end

  def set_user_time_zone
    Time.zone = current_user.time_zone if logged_in?
  end

  #feed in a string with a class name, get back a reference to the Class
  def class_object(class_name)
      Object::const_get(class_name).new().class
  end

  def current_model_class_object
      return class_object(self.controller_name.camelize)
  end

  def do_create(redirect_action = 'show')
    _my_model = current_model_class_object
    @record = _my_model.new(params[:record])

    if @record.save
      flash[:notice] = "#{_my_model.human_name} successfully created."
      redirect_to :action => redirect_action, :id => @record
    else
      # create failed
      render :action => 'new'
    end
  end

  def do_destroy
    _my_model = current_model_class_object
    @record = _my_model.find(params[:id])
    if @record.destroy
      #go back to list, try to show approx same place in list if there is a page position remembered
      redirect_to :action => 'list'
    else
      # destroy failed
      render :action => 'show'
    end
  end

  def do_list

  end

  def do_update(redirect_action = 'show')
    _my_model = current_model_class_object
    @record = _my_model.find(params[:id])
    @record.attributes = params[:record]
    if @record.save
      flash[:notice] = "#{_my_model.human_name} successfully updated."
      redirect_to :action => redirect_action, :id => (redirect_action=='list' ? nil : @record)
    else
      # save failed
      render :action => 'edit'
    end
  end

  # Duplicate of function defined in ApplicationHelper - need to move to lib
  # express a hash as a javascript object
  def hash_to_javascript(hash_to_convert)
    _item_array = []
    hash_to_convert.each_pair do |_key, _value|
      if _value.is_a? String
        _item_array << "\"#{escape_javascript(_key)}\": \"#{escape_javascript(_value)}\""
      elsif _value.respond_to? :each_pair
        _item_array << "\"#{escape_javascript(_key)}\": #{hash_to_javascript(_value)}"
      else
        raise "Attempt to convert hash value of unsupported type to javascript: type=#{_value.class.name}"
      end
    end
    "{#{_item_array.join(", ")}}"
  end

  def paginated_records_get(model, options)
    _result = model.paginate(
      {
        :conditions => options[:conditions],
        :order      => options[:order],
        :page       => options[:page],
        :per_page   => options[:per_page],
      }
    )
    return _result
  end

  # Duplicate of function defined in ApplicationHelper - need to move to lib
  # clone parameter to protect against inadvertently changing the original
  # also protect against nil
  def param_clone(key)
    if params[key].nil?
      return nil
    else
      return params[key].clone
    end
  end

  # generate a SQL 'WHERE' clause
  def param_filter_clause(model, prefix, base_clause = '')
      _filters = []
      if base_clause != ''
          _filters << base_clause.clone
      end

      for _key in params.keys
          # if the parameter name starts with prefix
          if _key.to_s.index("#{prefix}_")==0
              # then it's either a sort or a filter for the session grid
              _param_suffix = _key.to_s[(prefix.length+1)..-1]
              # if this is not a sort specification
              if !['order','sort'].include? _param_suffix
                #look up the filter clause
                if _param_suffix == 'filter_field'
                  _clause = model.get_filter_clause(params[_key], params["#{prefix}_filter_value"])
                elsif _param_suffix != 'filter_value'
                  _clause = model.get_filter_clause(_param_suffix, params[_key])
                end
                if !_clause.nil?
                    _filters << _clause
                end
              end
          end
      end

      _filter_clause = _filters.join(" AND ")
      if _filter_clause.length == 0
          return nil
      else
          return _filter_clause
      end
  end

  # Duplicate of function defined in ApplicationHelper - need to move to lib
  # express params as javascript object
  def params_as_javascript()
      hash_to_javascript(params)
  end

  def rescue_action_in_public(exception)
      super
  end

  # Returns cached instance of Browser which incapsulates browser-sniffing.
  def browser
    @browser ||= Browser.new(request.user_agent)
  end
  helper_method :browser

  # Handle exceptions connecting to node for configuration
  def handle_config_exception(e)
    case e.class.to_s
    when "Net::SSH::AuthenticationFailed"
      flash[:error] = "Authentication failed connecting to node: #{@node.node_address}"
    when "Net::SSH::HostKeyMismatch"
      flash[:error] = e.message
    when 'NodeConfiguration::AdminUserDocrootedError'
      flash[:error] = "Can't run asconfigurator.  Make sure admin user is not docrooted."
      redirect_to edit_credentials_node_url(@node) and return
    else
      flash[:error] = e.message
      logger.error "#{$!}" + $@.join("\n")
    end
    redirect_to edit_node_url(@node)
  end

  def node_config_flash_success
    flash.now[:notice] = "Changes saved."
    unless @node.asconfigurator_errors.empty?
      flash.now[:error] = @node.asconfigurator_errors.join("\n")[0, 1000]

    end
  end

  def node_config_flash_failure(error_message = "There were errors.  Changes not saved.")
    flash.now[:notice] = nil
    flash.now[:error] = error_message[0, 1000]
  end

  def asconfigurator_error_flash
    if !@node.asconfigurator_errors.empty?
      flash[:error] = @node.asconfigurator_errors.join("\n")[0, 1000]
    end
  end

  def node_is_configurable?
    unless @node.is_configurable?
      flash[:error] = "Need Admin Credentials set to configure this node."
      redirect_to edit_credentials_node_url(@node)
    end
  end

  def ami?
    Ami::Base.ami?
  end

  def sanitize(str)
    ActionController::Base.helpers.sanitize(str)
  end

  def saml_settings_object
    @saml_config ||= ConsoleConfig::SamlConfig.load_settings
    @saml_settings ||= OneLogin::RubySaml::Settings.new.tap do |settings|
      settings.assertion_consumer_service_url = saml_cb_url
      settings.issuer                         = saml_metadata_url
      settings.name_identifier_format         = NAME_IDENTIFIER_FORMAT_UNSPECIFIED
      settings.name_identifier_value          = current_user.email if defined?(current_user) && current_user.present?
      settings.idp_sso_target_url             = @saml_config.present? ? @saml_config.idp_sso_target_url : ""
      settings.idp_slo_target_url             = @saml_config.present? ? @saml_config.idp_slo_target_url : ""
      settings.idp_cert_fingerprint           = @saml_config.present? ? @saml_config.idp_cert_fingerprint : ""
      settings.idp_cert_fingerprint_algorithm = @saml_config.present? ? @saml_config.idp_cert_fingerprint_algorithm : XMLSecurity::Document::SHA1
    end
  end

  helper_method :ami?

  def ami_new_install?
    Ami::Base.new_install?
  end
  helper_method :ami_new_install?

  def ami_with_shares?
    Ami::Base.ami_with_shares?
  end
  helper_method :ami_with_shares?

  def remote_auth?
    ConsoleConfig::DirectoryService.load_settings.authenticate_remotely
  end
  helper_method :remote_auth?

  # Even when the user clicks the browser's  back button, the browser
  # will request the page instead of displaying the cached page.
  
  def disable_browser_cache
    # add pragma in the layout for IE:
    @aspera_prevent_caching = true
    # other browsers:
    headers["Cache-Control"] = "no-cache, no-store, max-age=0, must-revalidate"
    headers["Pragma"]        = "no-cache"
    headers["Expires"]       = "Fri, 01 Jan 1990 00:00:00 GMT"
  end

  def check_for_database_messages
    database_version_error_messages = Utils.database_version_error_messages
    license_error_message = nil
    
    if ConsoleConfig::EntitlementLicenseMode.enabled?
      ec = ConsoleConfig::EntitlementConfig.load_settings
      
      if !ec.safe_net_console_enabled || ec.safe_net_failed
        license_error_message = "This server is not currently entitled"
        if ec.safe_net_failure_message.present?
          license_error_message+=": #{ec.safe_net_failure_message}"
        end
      end

      if ConsoleConfig::EntitlementConfig.license_expired?
        license_error_message+="\n" if license_error_message.present?
        license_error_message = "License has expired on #{ConsoleConfig::EntitlementConfig.safe_net_expiration_date}."
      end
    else
      license_error_message = Utils.license_error_message
    end
    if license_error_message.present? && Utils.recoverable_license_error?
      msg = "Only admin logins permitted until this problem is resolved."
    else
      msg = "No logins permitted until this problem is resolved."
    end

    if database_version_error_messages || license_error_message
      flash[:notice] = nil
      flash[:warning] = nil
      flash.now[:error] = [
        database_version_error_messages,
        license_error_message,
        msg
      ].join("\n")
      return true
    end
    return false
  end

end
