class AspRpt::Template < ActiveRecord::Base

  SKIP_SANITIZATION = true

  set_table_name :rpt_templates

  default_scope :order => 'name'

  named_scope :by_name, :order => 'name'
  named_scope :built_in, :conditions => 'owner_id=0'
  named_scope :user_created, :conditions => 'owner_id!=0'

  concerned_with :prepackaged_reports

  before_save :trim_description

  belongs_to :owner, :class_name => "User"
  has_many :params,
           :class_name => "AspRpt::TemplateParam",
           :dependent => :destroy,
           :foreign_key => :rpt_template_id
  accepts_nested_attributes_for :params

  has_many :requests,
           :class_name => "AspRpt::Request",
           :dependent => :nullify,
           :foreign_key => :rpt_template_id

  has_many :scheduled_requests,
           :class_name => "AspRpt::Request",
           :dependent => :nullify,
           :foreign_key => :rpt_template_id,
           :conditions => ['status=?', Console::STATUS_SCHEDULED]
  
  validates_presence_of :name, :sql_text, :owner_id
  validates_uniqueness_of :name

  validate :validate_has_final_result_table
  validate :validate_source_tables
  before_save :sql_strip_trailing_spaces
  after_save :set_params

  FINAL_RESULT_TABLE = 'FINAL_RESULT_TABLE'
  REPORT_PERIOD_START = 'REPORT_PERIOD_START'
  REPORT_PERIOD_END = 'REPORT_PERIOD_END'
  REPORT_TIME_ZONE = 'REPORT_TIME_ZONE'
  USER_ID = 'USER_ID'

  default_value_for :sql_text, [
    %[CREATE TABLE $#{FINAL_RESULT_TABLE}],
    %[],
    %[SELECT DISTINCT],
    %[  t.name],
    %[  , t.contact],
    %[  , t.last_source_ip AS `from`],
    %[  , t.last_dest_ip AS `to`],
    %[  , t.started_at],
    %[  , t.stopped_at],
    %[  , t.status],
    %[  , t.bytes_transferred],
    %[  , t.files_complete],
    %[  , t.files_failed],
    %[  , t.files_skipped],
    %[],
    %[FROM],
    %[  $TBL_TRANSFERS t],
    %[],
    %[WHERE],
    %[  t.started_at < '$REPORT_PERIOD_END'],
    %[  AND (],
    %[    t.stopped_at >= '$REPORT_PERIOD_START'],
    %[    OR t.stopped_at IS NULL],
    %[  )],
    %[],
    %[ORDER BY],
    %[  t.started_at],
    %[  , t.id],
    %[;]
  ].join("\n")

  def self.options_for_select
    options = []
    unless built_in.empty?
      options << ['Built in',built_in.collect{|template| [template.name, template.id]}]
    end
    unless user_created.empty?
      options << ['Custom', user_created.collect{|template| [template.name, template.id]}]
    end
    options
  end

  # is this a built-in template? (non-editable, non-deletable)
  def built_in?
    owner_id==0
  end

  def advanced?
    report_class == 'Advanced'
  end

  def basic?
    report_class == 'Basic'
  end

  def sql_strip_trailing_spaces
    self.sql_text = sql_text && sql_text.split("\n").collect{|line| line.rstrip}.join("\n")
  end
  
  # Ensures the params association matches substitution variables in sql_text.
  def set_params
    if sql_text_changed?
      sql_param_variables.each do |sql_variable|
        params.find_or_create_by_name(sql_variable)
      end
      params_to_delete = params.select { |param| !sql_param_variables.include?(param.name) }
      params.destroy(params_to_delete)
      move_report_period_params_to_top
      set_types_for_builtin_params
    end
  end

  def move_report_period_params_to_top
    (params_report_period + params_report_period_not).each_with_index do |param, index|
      param.update_attribute(:position, (index + 1)) unless param.frozen?
    end
  end

  def report_period_variable_names
    @report_period_variables ||= [REPORT_PERIOD_START, REPORT_PERIOD_END, REPORT_TIME_ZONE]
  end

  def params_report_period
    @params_report_period ||= report_period_variable_names.map { |var_name| params.find_by_name(var_name) }.compact
  end

  def params_report_period_not
    @params_report_period_not ||= params - params_report_period
  end

  def set_types_for_builtin_params
    params.update_all("param_type = '#{AspRpt::TemplateParam::PARAM_TYPE_DATE}'", "name in ('#{REPORT_PERIOD_START}', '#{REPORT_PERIOD_END}')")
    params.update_all("param_type = 'time_zone'", "name = '#{REPORT_TIME_ZONE}'")
    params.update_all("required = FALSE", "name LIKE 'TMP_%'")
  end

  # Returns all $VARS found in sql_text (Array of String w/o "$").
  # Note that the default value for sql_text apparently doesn't
  # kick in here so we have to OR with an empty string.
  def sql_variables
    @sql_variables ||= begin
      sql_vars_with_dollar = (sql_text || '').scan(%r{\$[A-Z0-9_]+}).uniq
      sql_vars_with_dollar.map { |var| var[1, var.size-1] }
    end
  end

  # Support sql() to match the basic template's sql()
  def sql
    sql_text
  end

  def use_report_period_start?
      sql_variables.include?(REPORT_PERIOD_START)
  end

  def use_report_period_end?
      sql_variables.include?(REPORT_PERIOD_END)
  end

  # Returns list of parameters found in sql_text that are configurable
  # by admin (Array of AspRpt::TemplateParam)
  def configurable_params
    unconfigurable_var_names = unconfigurable_variables.map { |hash| hash[:name] }
    params.reject { |param| unconfigurable_var_names.include?(param.name) }
  end

  # Returns array of sql variables that get an entry in rpt_template_params.
  # These are variables that (can) get input from user at report request time.
  def sql_param_variables
    @sql_param_variables ||= begin
      vars = sql_variables.reject do |variable|
        variable == FINAL_RESULT_TABLE ||
        variable == USER_ID ||
        variable.starts_with?('TEMP_') ||
        variable.starts_with?('TBL_')
      end
      # if any of the report period variables are used, include all of them
      if (vars & [REPORT_PERIOD_START, REPORT_PERIOD_END, REPORT_TIME_ZONE]).present?
        vars |= [REPORT_PERIOD_START, REPORT_PERIOD_END, REPORT_TIME_ZONE]
      end
      vars
    end
  end

  def trim_description
    self.description = description && description[0, 2000]
  end

  # Returns variables whose UI is not configurable, either because they aren't
  # set by user input or are builtin.
  def unconfigurable_variables
    (source_table_variables +
      temp_table_variables +
      final_result_table_variable +
      user_variable +
      builtin_variables).flatten
  end

  # TODO: make configurable
  def source_tables
    @source_tables ||= {
      'TBL_NODES' => {:description => 'Nodes', :name => 'rpt_transfer_nodes'},
      'TBL_TRANSFERS' => {:description => 'Transfers', :name => 'rpt_transfers'},
      'TBL_FILES' => {:description => 'Files', :name => 'rpt_transfer_files'},
      'TBL_TRANSFER_SESSIONS' => {:description =>'Transfer sessions', :name => 'rpt_transfer_sessions'},
      'TBL_TRANSFER_SESSION_FILES' => {:description =>'Transfer session files', :name => 'rpt_transfer_session_files'},
    }
  end

  # Returns source table name.  Raise error if not found as the generated
  # sql text would be invalid.
  def source_table_name(var_with_dollar)
    var_without_dollar = var_with_dollar.sub(/^\$/, '')
    source_table = source_tables[var_without_dollar] || raise("Invalid source table var: #{var_with_dollar}")
    source_table[:name]
  end


  def source_table_variables
    sql_variables.
        select { |var| var.starts_with?("TBL_") }.
        map { |var| {:name => var, :description => ("Source table: #{source_tables[var][:description]}" rescue nil) } }
  end

  def temp_table_variables
    sql_variables.
        select { |var| var.starts_with?("TMP_") }.
        map { |var| {:name => var, :description => 'Temporary table'} }
  end

  def final_result_table_variable
    [{:name => FINAL_RESULT_TABLE, :description => 'Final results table'}]
  end

  def user_variable
    if sql_variables.include?(USER_ID)
      [{:name => USER_ID, :description => 'User id'}]
    else
      []
    end
  end

  def builtin_variables
    returning [] do |vars|
      vars << report_period_variables
      # TODO: make configurable
      # would add non-report period builtins (if we ever have any)
    end
  end

  def report_period_variables
    returning [] do |vars|
      vars << {:name => REPORT_PERIOD_START, :description => 'Builtin variable - Report period start'} if sql_variables.include?(REPORT_PERIOD_START)
      vars << {:name => REPORT_PERIOD_END, :description => 'Builtin variable - Report period end'} if sql_variables.include?(REPORT_PERIOD_END)
      vars << {:name => REPORT_TIME_ZONE, :description => 'Builtin - Added because of presence of REPORT_PERIOD_* variable'} if vars.present?
    end
  end

  def validate_has_final_result_table
    return unless sql_text.present?
    unless sql_text.include?("$#{FINAL_RESULT_TABLE}")
      errors.add(:sql_text, "must contain the variable $#{FINAL_RESULT_TABLE}")
    end
  end

  def validate_source_tables
    unknown_source_table_vars = source_table_variables.
        select { |hash| hash[:description].blank? }.
        map { |hash| "$#{hash[:name]}" }

    if unknown_source_table_vars.present?
      errors.add(:sql_text, "contains unrecognized source table variables: #{unknown_source_table_vars.join("\n")}")
    end
  end

  # Returns new template that is a clone of original.  We want to copy params
  # from original to pick up any customization rather than keep the generated
  # params from the #set_params() callback.
  def copy
    returning clone do |new_template|
      new_template.created_at = nil
      new_template.updated_at = nil
      new_template.name = unique_name_for_copy
      new_template.save!

      new_template.params = params.map { |param| param.clone }.
          each { |new_param| new_param.created_at = nil }.
          each { |new_param| new_param.updated_at = nil }
    end
  end

  # If name is "Foo", returns names like "Copy of Foo", "Copy 2 of Foo", etc.
  def unique_name_for_copy
    base_name = name.sub(/^Copy.*?of /, '')
    ([''] + (2..99).to_a).each do |counter|
      new_name = "Copy #{counter} of #{base_name}".squish
      return new_name unless AspRpt::Template.find_by_name(new_name)
    end
  end

  def sort_params(param_ids)
    param_ids.each_with_index do |id, index|
      params.find(id).update_attribute(:position, index)
    end
    move_report_period_params_to_top
  end

end