class AsperaSync < ActiveRecord::Base
  belongs_to :fasp_node
  validates_presence_of :name
  validates_presence_of :fasp_node
  validates_presence_of :local_dir
  validates_presence_of :remote_host_ip
  validates_presence_of :remote_dir

  named_scope :active, :conditions => { :active => true }
  named_scope :deactivated, :conditions => { :active => false }
  named_scope :not_marked_for_deletion, :conditions => { :mark_for_deletion => false }
  named_scope :marked_for_deletion, :conditions => { :mark_for_deletion => true }
  named_scope :valid_snid, :conditions => "snid IS NOT NULL && snid > 0"

  def summary
    an = AsperaNoded.new(self.fasp_node)
    ret = an.summary(self.snid.to_i)
    @json_errors = an.errors
    ret
  end
  
  def inode_format=(val)
    # old node versions still report this
    # need to handle the input, but do
    # not report this data via console
  end

  def bw_graph_data
    an = AsperaNoded.new(self.fasp_node)
    ret = an.bw_graph_data(self.snid)
    @json_errors = an.errors
    ret
  end
  

  def files(search=nil, status=nil, skip=nil)
    an = AsperaNoded.new(self.fasp_node)
    files,count = an.files(self.snid.to_s, search, status, skip)
    @json_errors = an.errors
    [files,count]
  end

  def json_errors
    @json_errors||[]
  end

  def self.serialize_v2_json(sync_summary)
    ret = {}
    remove_fields = [:cumulative_error_files, :session_syncd_files_transferred, :cumulative_syncd_files, :session_syncd_files_same, 
    :session_syncd_files_deduped, :cumulative_conflict_files, :cumulative_total_files, :cumulative_pot_conflict_files, :cumulative_pending_files, 
    :ascp_session_ids, :session_syncd_files_skipped, :partially_completed_files, :cumulative_error_bytes]
    
    sync_summary.each do |k,v|
      ret[k.to_sym] = v
    end
    ret[:start_time] = Time.at(ret[:start_time].to_i/1000)
    ret[:snid] = ret[:snid].to_i
    ret[:json_api_version] = '2'

    ret.except(*remove_fields)
  end

  def self.serialize_v1_json(sync_summary)
    { :sid => sync_summary[:sid],
      :snid => sync_summary["sync_id"].to_i,
      :fasp_node_id => sync_summary[:fasp_node_id],
      :transferred_bytes => sync_summary["total_bytes_transferred"],
      :deleted_bytes => sync_summary["deleted_bytes"],
      :deduped_bytes => sync_summary["deduped_bytes"],
      :cumulative_transferred_bytes => sync_summary["cumulative_total_bytes_transferred"],
      :dedupe_enabled => (sync_summary["dedupe_enabled"].to_i == 1),
      :cumulative_deleted_bytes => sync_summary["cumulative_deleted_bytes"],
      :fasp_session_id => sync_summary["session_id"],
      :cumulative_deduped_bytes => sync_summary["cumulative_deduped_bytes"],
      :name => sync_summary["name"],
      :local_dir => sync_summary["local_dir"],
      :local_host => sync_summary["local_host"],
      :local_user => sync_summary["local_user"],
      :remote_dir => sync_summary["remote_dir"],
      :remote_host => sync_summary["remote_host"],
      :remote_host_ip => sync_summary["remote_host_ip"],
      :direction => sync_summary["direction"],
      :status => sync_summary["status"],
      :checksum_type => sync_summary["checksum"],
      :total_paths => sync_summary["paths_total"],
      :syncd_paths => sync_summary["paths_syncd"],
      :pending_paths => sync_summary["paths_pending"],
      :conflict_paths => sync_summary["paths_conflict"],
      :deleted_paths => sync_summary["paths_deleted"],
      :error_paths => sync_summary["paths_error"],
      :remote_user => sync_summary["remote_user"],
      :start_time => Time.at(sync_summary["start_time"].to_i/1000),
      :json_api_version => '1'
    }
  end

  def active?
    !!active
  end

  def deactivated?
    !active
  end

  def display_status
    !active? ? "deactivated" : status
  end

  def stopped?
    status == 'stopped'
  end

  def defunct?
    status == 'defunct'
  end

  def dedupe_enabled?
    !!dedupe_enabled
  end

  def deactivate!
    update_attribute(:active, false)
    update_attribute(:snid, nil)
    update_attribute(:snid, nil)
    update_attribute(:pending_paths, 0)
  end

  def ok_to_remove?
    stopped? || deactivated? || defunct?
  end

  def self.ok_to_delete?(status)
    %w(stopped defunct).include?(status)
  end

  def cookie_string
    "aspera:async:#{self.fasp_session_id}:#{self.name}"
  end

  def self.delete_marked_for_deletion
    self.transaction do
      return if self.marked_for_deletion.count == 0
      self.marked_for_deletion.each do |sync|
        an = AsperaNoded.new(sync.fasp_node)
        if sync.active?
          success,response = an.delete(sync.snid)
          if success
            sync.destroy
            logger.info("Successfully deleted sync from redis. snid=#{sync.snid}, id=#{sync.id}, msg=#{response[:message]}, code=#{response[:code]}")
          else
            logger.error("Error deleting sync from redis. snid=#{sync.snid}, id=#{sync.id}, msg=#{response[:message]}, code=#{response[:code]}")
          end
        else
          sync.destroy
        end
      end
    end
  end

  def is_v1?
    self.json_api_version == '1'
  end

  def is_v2?
    self.json_api_version == '2'
  end

end