class BackgroundProcess < ActiveRecord::Base

  concerned_with :status

  validate :validate_daily_restart
  validates_numericality_of :sleep_seconds
  validates_numericality_of :max_startup_interval_seconds
  validates_numericality_of :max_heartbeat_interval_seconds

  delegate :batch_size_used?, :startup_sequence, :description, :controllable_thru_ui?, :log_file_path, :output_file_path,
           :start, :kill, :daily_restart_used?,
           :start_process, :stop_process, :command_line,
           :to => :background_process_class

  enum_helper :status, ['starting', 'running', 'stopped'], :prefix => :none, :allow_blank => true
  enum_helper :pid_status, ['running', 'stopped'], :allow_blank => true
  enum_helper :request, ['start', 'stop', 'kill'], :allow_blank => true
  enum_helper :startup_type, ['manual', 'automatic', 'disabled']

  named_scope :auto, :conditions => {:startup_type => STARTUP_TYPE_AUTOMATIC}

  SKIP_SANITIZATION = true

  # Note: Called by MCP process.  Only MCP should log.
  def monitor
    if needs_to_be_started?
      mcp_logger.info "....needs to be started"
      if no_other_process_starting?
        mcp_logger.info "......no process currently starting"
        mcp_logger.info "......Starting process #{name}"
        start
      else
        mcp_logger.info "......Waiting to start #{name} (other process(es) starting)"
        mcp_logger.info "........other process(es): #{BackgroundProcess.processes_starting.map(&:name).join(", ")}"
      end
    elsif needs_to_be_killed?
      mcp_logger.info "....needs to be killed"
      mcp_logger.warn "......Killing #{name}"
      kill
    elsif restart?
      restart_process
    elsif time_to_restart?
      update_attribute(:next_restart_at, next_restart_at + 1.day)
      request_restart
    else
      mcp_logger.info "....nothing needed"
#      cleanup_request
    end
  end

  def produces_output_file?
    background_process_class.produces_output_file?
  end

  def time_to_restart?
    pid_status_running? && next_restart_at? && (Time.now > next_restart_at)
  end

  def needs_to_be_started?
    pid_status_stopped? && request_start?
  end

  def no_process_starting?
    !self.class.any_process_starting?
  end

  def no_other_process_starting?
    process_names = BackgroundProcess.processes_starting.map(&:name)
    process_names.delete(self.name)
    process_names.empty?
  end

  def needs_to_be_killed?
    request_kill? || (request_stop? && hung?)#|| (starting? && startup_interval_exceeded?)
  end

  def hung?
    (running? && heartbeat_not_current?) #|| (starting? && startup_interval_exceeded?)
  end

  def heartbeat_current?
    heartbeat? && (Time.now - heartbeat) <= max_heartbeat_interval_seconds
  end

  def heartbeat_not_current?
    !heartbeat_current?
  end

  def startup_interval_exceeded?
    exceeded = (Time.now - (last_attempted_start_time || updated_at)) > max_startup_interval_seconds
    mcp_logger.info "startup timeout check: last_attempted_start_time=#{last_attempted_start_time}, updated_at=#{updated_at}"
    mcp_logger.warn "startup interval exceeded! #{(Time.now - (last_attempted_start_time || updated_at))} seconds exceeds configured maximum (#{max_startup_interval_seconds})" if exceeded
    exceeded
  end

  def request_start
    mcp_logger.debug "requesting start, id=#{id}, name=#{name}"
    result = update_attribute(:request, REQUEST_START)
    mcp_logger.debug "requesting start, id=#{id}, name=#{name}, result=#{result}"
  end

  def request_stop
    mcp_logger.debug "requesting stop, id=#{id}, name=#{name}"
    result = update_attribute(:request, REQUEST_STOP)
    mcp_logger.debug "requesting stop, id=#{id}, name=#{name}, result=#{result}"
  end

  def request_kill
    mcp_logger.debug "requesting kill, id=#{id}, name=#{name}"
    result = update_attribute(:request, REQUEST_KILL)
    mcp_logger.debug "requesting kill, id=#{id}, name=#{name}, result=#{result}"
  end

  def request_restart
    mcp_logger.debug "requesting restart, id=#{id}, name=#{name}"
    result = update_attribute(:restart, true)
    mcp_logger.debug "requesting restart, id=#{id}, name=#{name}, result=#{result}"
  end

  def restart_process
    if running?
      request_stop
    elsif hung?
      mcp_logger.warn "restart_process: process hung, killing: id=#{id}, name=#{name}"
      kill
    elsif stopped?
      request_start
    end
  end

  def update_attributes_for_stopped
    mcp_logger.debug "update_attributes_for_stopped, id=#{id}, name=#{name}"
    result = update_attributes({
                          :request => nil,
                          :status => BackgroundProcess::STOPPED,
                          :heartbeat => nil,
                          :pid_status => BackgroundProcess::PID_STATUS_STOPPED,
                          :pid => nil
                      })
    mcp_logger.debug "update_attributes_for_stopped, id=#{id}, name=#{name}, result=#{result}"
  end

  def update_attributes_for_running
    mcp_logger.debug "update_attributes_for_running, id=#{id}, name=#{name}"
    result = update_attributes({
                          :status => BackgroundProcess::RUNNING,
                          :pid_status => BackgroundProcess::RUNNING,
                          :heartbeat => Time.now,
                          :pid => $$
                      })
    mcp_logger.debug "update_attributes_for_running, id=#{id}, name=#{name}, result=#{result}"
  end

  def initial_heartbeat
    sql = "SELECT MAX(initial_heartbeat) FROM #{heartbeat_table_name}"
    result = self.connection.select_all(sql).first['MAX(initial_heartbeat)']
    result && result.to_datetime
  end

  def last_heartbeat
    sql = "SELECT MAX(heartbeat) FROM #{heartbeat_table_name}"
    result = self.connection.select_all(sql).first['MAX(heartbeat)']
    result && result.to_datetime
  end

  def raw_loop_time_seconds
    sql = "SELECT MAX(loop_time_seconds) FROM #{heartbeat_table_name}"
    result = self.connection.select_all(sql).first['MAX(loop_time_seconds)']
    result && result.to_i
  end

  def heartbeat_table_name
    name.gsub('AsperaBackground::','ab').underscore
  end

  def display_status
    if restart?
      'restarting'
    elsif starting?
      'starting'
    elsif request_stop?
      'stopping'
    elsif hung?
      'slow'
    elsif stopped?
      'stopped'
    else
      pid_status
    end
  end

  def background_process_class
    name.split('/').map { |part| part.classify }.join("::").constantize
  end


  # Validates that #daily_restart_time time is of the form HH:MM.
  # Has side-affect of resetting #next_restart_at (which will be
  # cleared by master process if process is not running).
  def validate_daily_restart
    if daily_restart_time.blank?
      self.next_restart_at = nil
    elsif daily_restart_time_changed?
      msg = "must be of the form HH:MM"
      if daily_restart_time =~ /\d{2}:\d{2}/
        hh, mm = daily_restart_time.split(":")
        if (0..23).include?(hh.to_i) && (0..59).include?(mm.to_i)
          self.next_restart_at = calculate_next_restart_at
        else
          errors.add :daily_restart_time, msg
        end
      else
        errors.add :daily_restart_time, msg
      end
    end
  end

  def calculate_next_restart_at
    if daily_restart_time?
      if daily_restart_time >= (Time.now.utc + 1.minute).strftime("%H:%M")
        day_at_time(Time.now.utc.to_date)
      else
        day_at_time(Time.now.utc.to_date + 1)
      end
    else
      nil
    end
  end

  def day_at_time(date)
    hh, mm = daily_restart_time.split(":")
    DateTime.new(date.year, date.month, date.day, hh.to_i, mm.to_i, 0)
  end

  # We start up based on startup_sequence and class name.  Also display that way.
  def auto_start_sequence
    auto_start_sort = startup_type_automatic? ? "0" : "1"
    auto_start_sort + sprintf("%5d", startup_sequence) + name
  end

  def mcp_logger
    self.class.mcp_logger
  end

  def bp_logger
    background_process_class.logger
  end

  # Returns sql where fragment for finding tokens this process holds but shouldn't.  Should not
  # hold any token if not running (pid is nil) or if bp_tokens.holder_pid doesn't match bp pid.
  #
  #   # will return one of:
  #   "(holder = 'ClassName')"
  #   "(holder = 'ClassName' and holder_pid != pid)"
  def stale_token_where_clause
    holder = background_process_class.token_holder_name
    ary = [%(holder = '#{holder}')]
    ary << %(holder_pid != #{pid}) if pid
    "(#{ary.join(' and ')})"
  end

  def daily_restart_time_minute
    self[:daily_restart_time].to_s.gsub(/\d\d:/,'')
  end

  def daily_restart_time_hour
    self[:daily_restart_time].to_s.gsub(/:\d\d/,'')
  end

  def daily_restart_time_minute=(minute)
    self[:daily_restart_time] = "" if daily_restart_time.nil?
    self[:daily_restart_time] += ":#{minute}"
  end

  def daily_restart_time_hour=(hour)
    self[:daily_restart_time] = "#{hour}#{daily_restart_time.to_s}"
  end

  class << self

    def request_start
      update_attributes_for_stopped
      auto.each do |bp|
        mcp_logger.info "Requesting start for #{bp.name}"
        bp.request_start
      end
    end

    # FIXME: comments
    def request_stop
      # clear request column (i.e., 'start') for all stopped processes
      updates = ['request = ?', nil]
      conditions = {:status => STOPPED, :request => REQUEST_START}
      update_all updates, conditions

      # request stop for each process that isn't stopped (running, starting)
      child_processes.each do |bp|
        unless bp.stopped?
          mcp_logger.warn "... requesting stop for #{bp.name}"
          bp.request_stop
        end
      end
    end

    # Should only be called when all processes are stopped.  Makes sure all status-type columns
    # are set properly for stopped processes.
    def update_attributes_for_stopped
      mcp_logger.debug "Update ALL background processes with attributes for stopped"
      all.each(&:update_attributes_for_stopped)
      mcp_logger.debug "Finished updating ALL background processes with attributes for stopped"
    end

    # Master Control Process logger.  Use this logger in methods called by MCP.
    def mcp_logger
      AsperaBackground::Control::Master.logger
    end

    # All background processes other than MCP.
    def child_processes
      AsperaBackground::Control::Master.child_processes
    end

    # Monitor all processes.
    # FIXME: probably want to exclude MCP from the monitoring
    def monitor
      mcp_logger.info "monitoring each background process..."
      background_processes = child_processes.sort_by(&:auto_start_sequence)
      background_processes.each do |bp|
        mcp_logger.info "..monitoring #{bp.name} request=#{bp.request} status=#{bp.status} pid=#{bp.pid}"
        mcp_logger.info "..calling background_process_class.cleanup..."
        bp.background_process_class.cleanup
        mcp_logger.info "..calling background_process_class.monitor..."
        bp.monitor
      end
    end

    def processes_starting
      find(:all, :conditions => ["status = ?", STARTING])
    end

    def any_process_starting?
      count(:conditions => ["status = ?", STARTING]) != 0
    end

    # Returns the next auto-start process to be started.
    def auto_start_next
      conditions = {
          :startup_type => STARTUP_TYPE_AUTOMATIC,
          :pid_status   => PID_STATUS_STOPPED,
          :status       => STOPPED,
      }
      all(:conditions => conditions).
          sort_by(&:auto_start_sequence).
          first
    end

    def auto_start_starting?
      master_id = AsperaBackground::Control::Master.status_record.id
      auto.all(:conditions => ["id != ? and status = ?", master_id, STARTING]).present?
    end

    # FIXME: needs to be running (per pid) and not hung
    def running
      all.select { |bp| bp.running? }
    end

    # Release any tokens held by background processes if process is not running or if
    # process pid doesn't match token holder pid.
    # MCP already loaded background processes.  Save reloading in time-critical loop.
    def release_stale_tokens(background_processes)
      mcp_logger.info "..Looking for stale tokens to release..."
      st = stale_tokens(background_processes)
      if st.empty?
        mcp_logger.info "....No stale tokens found"
      else
        mcp_logger.info "....Found #{st.size} stale tokens"
      end
      st.map { |e| e.remove_holder }
    end

    # Returns array of tokens that should be released.
    def stale_tokens(background_processes)
      where_clause = background_processes.map(&:stale_token_where_clause).join(' or ')
      mcp_logger.info "......Search criteria: #{where_clause}"
      AsperaBackground::Control::Token.all(:conditions => where_clause)
    end

  end
end