# Status constants defined by Aspera Console.  
# See DbLogger class for other status constants.

require 'progress_log'

class Console
  
  # Exit code when set by Console (as opposed to dblogger).
  Console::EXIT_CODE = 600

  # For background processes that work with transfer data, the amount of time to wait between
  # marking a record for deletion and actually removing it from the database.
  # Record deletion is separated into two stages (marking and actual removal) primarily to accommodate
  # delays in file records arriving, but also allows decoupling time-critical display processing
  # (i.e. removing records from view) from the actual record deletion which can be time-consuming  
  TRANSFER_DATA_DELETION_WAIT = 1.hour
  
  # These values matter!  Used to calculate column names in:
  #
  # * JobEmail
  # * JobTemplateEmail
  # * FaspSession
  #---  
  EMAIL_TYPE_START   = 'start'
  EMAIL_TYPE_SUCCESS = 'success'
  EMAIL_TYPE_ERROR   = 'error'
  
  EMAIL_TYPES = [EMAIL_TYPE_START, EMAIL_TYPE_SUCCESS, EMAIL_TYPE_ERROR]
  
  # application names used for transfers started via console
  CONSOLE_APPLICATION = 'aspera.console' # this console
  OUTSIDE_CONSOLE_APPLICATION = 'aspera.console (outside)' # someone else's console

  # priority + policy values from db logger become a single 'priority' description in Console
  PRIORITY_FIXED = 'Fixed'
  PRIORITY_HIGH = 'High'
  PRIORITY_NORMAL = 'Normal'
  PRIORITY_LOW = 'Low'
  ALL_PRIORITIES = [PRIORITY_FIXED, PRIORITY_HIGH, PRIORITY_NORMAL, PRIORITY_LOW]
  
  #dblogger
  STATUS_STARTED    = DbLogger::STATUS_STARTED # deprecated, only applies to FaspSession and DbLogger
  STATUS_RUNNING    = DbLogger::STATUS_RUNNING
  STATUS_PAUSED     = DbLogger::STATUS_PAUSED
  STATUS_COMPLETED  = DbLogger::STATUS_COMPLETED
  STATUS_CANCELLED  = DbLogger::STATUS_CANCELLED
  STATUS_ERROR      = DbLogger::STATUS_ERROR
  STATUS_WILLRETRY  = DbLogger::STATUS_WILLRETRY
  STATUS_VALIDATING    = DbLogger::STATUS_VALIDATING
  STATUS_TO_BE_VALIDATED = DbLogger::STATUS_TO_BE_VALIDATED
  
  # console only
  STATUS_QUEUED        = 'queued'
  STATUS_SUBMITTED     = 'submitted'
  STATUS_SCHEDULED     = 'scheduled'
#  STATUS_NEVER_STARTED = 'never started'
  STATUS_INACTIVE      = 'inactive'
  STATUS_UNKNOWN       = 'unknown'
  
  # FIXME: shouldn't need STARTED since background prevents it from happening
  STATUSES_ACTIVE = [
    STATUS_STARTED,
    STATUS_RUNNING,
    STATUS_PAUSED,
    STATUS_VALIDATING,
    STATUS_TO_BE_VALIDATED,
    STATUS_UNKNOWN
  ]

  # Statuses that indicate a job or job_step may get further updates from the db logger.
  STATUSES_POSSIBLY_ACTIVE_FOR_JOB_AND_STEP = [
    STATUS_QUEUED,
    STATUS_SUBMITTED,
    STATUS_RUNNING,
    STATUS_PAUSED,
    STATUS_VALIDATING,
    STATUS_TO_BE_VALIDATED,
    STATUS_UNKNOWN,
  ]

  STATUSES_CONCURRENT = [
    STATUS_QUEUED, STATUS_SUBMITTED, STATUS_RUNNING, STATUS_PAUSED, STATUS_WILLRETRY, STATUS_VALIDATING, STATUS_TO_BE_VALIDATED
  ]

  CLOSED_STATUS_LIST = "'#{[STATUS_COMPLETED, STATUS_ERROR, STATUS_CANCELLED, STATUS_INACTIVE, STATUS_WILLRETRY].join("', '")}'"

  STATUSES_EMAIL_ACTIVE =  STATUSES_ACTIVE
  STATUSES_EMAIL_SUCCESS = [STATUS_COMPLETED]
  STATUSES_EMAIL_ERROR   = [STATUS_CANCELLED, STATUS_ERROR, STATUS_INACTIVE]
  STATUSES_EMAIL_STARTED = [STATUS_RUNNING, STATUS_PAUSED, STATUS_WILLRETRY] + STATUSES_EMAIL_SUCCESS + STATUSES_EMAIL_ERROR
  
  STATUSES_DASHBOARD_ACTIVE = [
    STATUS_QUEUED,
    STATUS_SUBMITTED,
    STATUS_RUNNING,
    STATUS_PAUSED,
    STATUS_WILLRETRY,
    STATUS_VALIDATING,
    STATUS_TO_BE_VALIDATED,
    STATUS_UNKNOWN
  ]
  
  STATUSES_DASHBOARD_PROBLEMS = [
    STATUS_ERROR,
#    STATUS_WILLRETRY,
#    STATUS_NEVER_STARTED,
    STATUS_INACTIVE,
    # cancelled jobs not really problems and not really successes
    # but we have to show them somewhere....
    STATUS_CANCELLED 
  ]
  
  SOAP_COMM_ERROR = 505
  SOAP_COMM_ERROR_MSG = "SOAP communication failure"
  
#  SESSION_NEVER_STARTED = 506
#  SESSION_NEVER_STARTED_MSG = "Session never started or node not reporting"

  EXCEPTION_ERROR = 507

  # DASHBOARD_RECENT
  # 
  # DASHBOARD_ERRROR
  # EMAIL_ERROR
  
  # See discussion at http://www.ruby-forum.com/topic/62553
  VALID_IP_REGEX = /\A(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3}\z/

  # Generalization of above to take *, n.*, n.m.*, n.m.o.*, n.m.o.p
  VALID_IP_FILTER_REGEXP = /\A(\*|(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){0,2}\.\*)|(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3})\z/
    
  # validation regular expression for input that could either be a valid IP address or a DNS name
  # the DNS expression is taken from http://regexlib.com/REDetails.aspx?regexp_id=391
  # VALID_DNS_TERM_REGEXP = "([a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?\.)+([a-zA-Z]){2,6}"
  # [mk@ FIX] added "_" as valid part of FQDN: [a-zA-Z0-9\-] -> [a-zA-Z0-9_\-]
  # VALID_DNS_TERM_REGEXP = "([a-zA-Z0-9]([a-zA-Z0-9_\-]{0,61}[a-zA-Z0-9])?\.)+([a-zA-Z]){2,6}"
  # VALID_IP_TERM_REGEXP = "(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3}"
  # VALID_IP_OR_DNS_STRING = "\A(#{VALID_DNS_TERM_REGEXP}|#{VALID_IP_TERM_REGEXP})\z"
  VALID_IP_OR_DNS_REGEXP = /\A(\*|([a-zA-Z0-9]([a-zA-Z0-9_\-]{0,61}[a-zA-Z0-9])?\.)+([a-zA-Z]){2,6}|(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3})((\/\w+)*)\z/
  VALID_IP_OR_DNS_OR_FQDN_REGEXP = /(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{0,62}[a-zA-Z0-9]\.)+[a-zA-Z]{2,63}$)|\A(\*|([a-zA-Z0-9]([a-zA-Z0-9_\-]{0,61}[a-zA-Z0-9])?\.)+([a-zA-Z]){2,6}|(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3})((\/\w+)*)\z/

  VALID_URL_REGEXP = /^(http|https):\/\/([a-z0-9]+([\-\.]{1}[a-z0-9]+)*\.[a-z]{2,5}| (?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)(?:\.(?:25[0-5]|(?:2[0-4]|1\d|[1-9])?\d)){3})(:[0-9]{1,5})?(\/.*)?/ix
  VALID_URL_REGEXP2 = /\A(http|https):\/\/\S+\z/  

  
  class << self

    # convert strings and nils to arrays that can be join()ed or passed into sanitize_sql
    # arrayize(nil) = []
    # arrayize(1) = [1]
    # arrayize([1,2,3]) = [1,2,3]
    def arrayize(value_or_value_array)
      if value_or_value_array.nil?
        []
      elsif value_or_value_array.kind_of? Array
        value_or_value_array
      else
        [value_or_value_array]
      end
    end

    # convert strings to single element arrays, initialize unsupplied options to empty arrays    
    def conform_options!(option_hash, keys)
      keys.each { |key|
        option_hash[key] = arrayize(option_hash[key])
      }
    end
    
    def execute_sql_string_or_array(sql_string_or_array)
      ActiveRecord::Base.logger.debug sql_string_or_array.join(' ')
      ActiveRecord::Base.connection.execute(arrayize(sql_string_or_array).join(' '))
    end

    def archive_one_table(opts)
      conform_options! opts, [
        :fields,
        :filename_prefix,
        :filename_suffix,
        :join,
        :order,
        :table,
        :where,
      ]

      # Note: using Job here in place of ActiveRecord::Base because an actual ActiveRecord is required.
      execute_sql_string_or_array %W[
        SELECT #{opts[:fields].join(',')}
        INTO OUTFILE '#{opts[:filename_prefix]}_#{opts[:filename_suffix]}.csv'
        FIELDS
          TERMINATED BY ','
          OPTIONALLY ENCLOSED BY '"' 
          ESCAPED BY '\\\\'
        LINES
          TERMINATED BY '\\n'
        FROM #{opts[:table]}
        #{"#{Job.public_sanitize_sql_for_conditions(opts[:join])}" unless opts[:join].empty?}
        #{"WHERE #{Job.public_sanitize_sql_for_conditions(opts[:where])}" unless opts[:where].empty?}
        #{"ORDER BY #{opts[:order].join(',')}" unless opts[:order].empty?}]
    end

    # Note: using Job here in place of ActiveRecord::Base because an actual ActiveRecord is required.
    def purge_one_table(opts)
      conform_options! opts, [:table, :fields, :join, :where]
      execute_sql_string_or_array %W[
        DELETE #{opts[:fields].join(',')}
        FROM #{opts[:table]}
        #{"#{Job.public_sanitize_sql_for_conditions(opts[:join])}" unless opts[:join].empty?}
        #{"WHERE #{Job.public_sanitize_sql_for_conditions(opts[:where])}" unless opts[:where].empty?}]
    end
    
    def archive_schema(filename_prefix, tables)
      archive_one_table({
        :table => 'information_schema.COLUMNS',
        :fields => %W[
          TABLE_SCHEMA
          TABLE_NAME
          COLUMN_NAME
          ORDINAL_POSITION
          COLUMN_DEFAULT
          IS_NULLABLE
          DATA_TYPE
          CHARACTER_MAXIMUM_LENGTH
          CHARACTER_OCTET_LENGTH
          NUMERIC_PRECISION
          NUMERIC_SCALE
          CHARACTER_SET_NAME
          COLLATION_NAME
          COLUMN_TYPE
          COLUMN_KEY
          EXTRA],
        :filename_prefix => filename_prefix,
        :filename_suffix => :schema,
        :order => [:TABLE_NAME, :ORDINAL_POSITION],
        :where => [
          "TABLE_SCHEMA=? AND TABLE_NAME IN (?)",
          ActiveRecord::Base.connection.current_database,
          tables],
      })
    end

    # archive data before a certain date and then remove from database
    # rake task 'archive' calls this. --cbn 10/4/11
    def archive_and_purge_data(before_date)
      filename_prefix = "console_backup_pre_#{before_date}_at_#{Time.now.utc.strftime('%Y-%m-%d_%H%M%S')}_utc"
      work_list = [
        {:table => :fasp_nodes,
          :archive => true,
          :fields => '*',
          :filename_suffix => :nodes,
          :purge => false,
        },
        {:table => :fasp_rates,
          :archive => true,
          :fields => 'fr.*',
          :filename_suffix => :rates,
          :join => 'fr JOIN fasp_sessions fs USING (node_uuid, session_id)',
          :order => 'fr.applied_at, fr.id',
          :purge => true,
          # use fasp_session.stopped_at rather than fasp_rates.applied at because we don't want to pick up rates for sessions that are still active
          :where => ['fs.stopped_at < ?', before_date],
        },
        {:table => :fasp_files,
          :archive => true,
          :fields => 'ff.*',
          :filename_suffix => :files,
          :join => 'ff JOIN fasp_sessions fs USING (node_uuid, session_id)',
          :order => 'ff.started_at, ff.id',
          :purge => true,
          # use fasp_session.stopped_at rather than fasp_files.stopped at because db logger can leave this blank in some cases
          # also, we don't want to pick up files for sessions that are still active
          :where => ['fs.stopped_at < ?', before_date],
        },
        {:table => :fasp_sessions,
          :archive => true,
          :fields => 'fasp_sessions.*',
          :filename_suffix => :sessions,
          :purge => true,
          :where => ['stopped_at < ?', before_date]
        },
        {:table => :sent_emails,
          :archive => true,
          :fields => 'se.*',
          :filename_suffix => :sent_emails,
          :join => 'se JOIN sent_email_events see ON se.sent_email_event_id=see.id JOIN job_steps js ON see.job_step_id=js.id JOIN jobs j ON js.job_id=j.id',
          :order => 'se.id',
          :purge => true,
          :where => ['j.ended_at < ?', before_date]
        },        
        {:table => :sent_email_events,
          :archive => true,
          :fields => 'see.*',
          :filename_suffix => :sent_email_events,
          :join => 'see JOIN job_steps js ON see.job_step_id=js.id JOIN jobs j ON js.job_id=j.id',
          :order => 'see.id',
          :purge => true,
          :where => ['j.ended_at < ?', before_date]
        },
        {:table => :job_step_email_events,
          :archive => true,
          :fields => 'jsee.*',
          :filename_suffix => :job_step_email_events,
          :join => 'jsee JOIN job_steps js ON jsee.job_step_id=js.id JOIN jobs j ON js.job_id=j.id',
          :order => 'jsee.id',
          :purge => true,
          :where => ['j.ended_at < ?', before_date]
        },
        {:table => :job_step_emails,
          :archive => true,
          :fields => 'jse.*',
          :filename_suffix => :job_step_emails,
          :join => 'jse JOIN job_steps js ON jse.job_step_id=js.id JOIN jobs j ON js.job_id=j.id',
          :order => 'jse.id',
          :purge => true,
          :where => ['j.ended_at < ?', before_date]
        },
        {:table => :job_steps,
          :archive => true,
          :fields => 'js.*',
          :filename_suffix => :job_steps,
          :join => 'js JOIN jobs j ON js.job_id=j.id',
          :order => 'js.id',
          :purge => true,
          :where => ['j.ended_at < ?', before_date]
        },
        {:table => :jobs,
          :archive => true,
          :fields => 'jobs.*',
          :filename_suffix => :jobs,
          :order => :id,
          :purge => true,
          :where => ['ended_at < ?', before_date]
        },
        
      ]

      # perform all archive steps first, don't purge until all have succeeded
      archive_worklist = work_list.find_all {|w| w[:archive]}
      archive_tablenames = archive_worklist.map {|w| w[:table].to_s}
      archive_schema(filename_prefix, archive_tablenames)
      
      archive_worklist.each {|w|
        w[:filename_prefix] = filename_prefix
        archive_one_table w
      }

      purge_worklist = work_list.find_all {|w| w[:purge]}
      purge_worklist.each {|w|
        w[:filename_prefix] = filename_prefix
        purge_one_table w
      }

    end

    def alternate_protocols_enabled
      console_config["alternate_protocols_enabled"]
    end

    def alternate_protocols_port
      (console_config["alternate_protocols_port"] || 40002).to_i
    end

    def console_config
      CONSOLE_CONFIG
    end

    def email_file_list_max_length
      (console_config[:email_file_list_max_length] || 1000).to_i
    end

    def default_database_ip
      console_config[:default_database_ip]
    end

    def nab_demo?
      console_config[:nab_demo] ? true : false
    end
    
    def database_config
      Rails.configuration.database_configuration[Rails.env]
    end
    
    def username
      database_config["username"]
    end
    
    def password
      database_config["password"]
    end
    
    def database
      database_config["database"]
    end
    
    def host
      database_config["host"]
    end
    
    def port
      database_config["port"]
    end

    # Developers can use mysql_bin_dir in console.yml to directly 
    # specify where their mysql binaries are located or this will
    # return the default "#{rubys_bin_dir}/../../mysql/bin"
    def mysql_bin_dir
      database_config['mysql_bin_dir'] ||
        "#{Aspera::Process.ruby_bin_dir}/../../mysql/bin"
    end

    # Returns +true+ if in Rails development environment
    def development?
      Rails.env == 'development'
    end

    # Returns +true+ if in Rails test environment
    def test?
      Rails.env == 'test'
    end

    # Returns +true+ if in Rails production environment
    def production?
      Rails.env == 'production'
    end

    # Returns _ip_address_ with leading 0's so it will sort properly.
    #
    #   Console.sortable_ip_address("1.22.33.212")  => "001.022.033.212"
    #   Console.sortable_ip_address("*")            => "*"
    #   Console.sortable_ip_address("1.22.*")       => "001.022.*"
    def sortable_ip_address(ip_address)
      ip_address.split(".").map do |octet|
        octet =~ /^\d{1,3}$/ ? sprintf( "%03d", octet) : (octet == "*" ? "999" : octet)
      end.join(".")
    end
    
    # Version number we show user in footer.
    def user_version
      version_txt.first.blank? ? "" : "(v#{version_txt.first})"
    rescue
      ""
    end
    
    # Build number we show in comments
    def comments_version
      "Version: #{version_txt.last}"
    rescue
      ""
    end
    
    def version_txt
      @version_txt ||= IO.read("#{Rails.root}/config/version.txt").split(/[\n\r]+/)
    rescue
      []
    end
    
    def version_path
      "#{RAILS_ROOT}/config/version.txt"
    end
    
    def uac_version
      'uac_version,"1.0"'
    end

    # Prints a message to screen for long-running tasks.  Prints _message_ at 
    # start of block passed in, then 'done' when done.
    def log_to_screen(message)
      if test?
        yield
      else
        # If the progress_file_path environment variable is set, we will write log data there too.
        ProgressLog.log message
        
        # print "#{message} ... "
        # STDOUT.flush
        yield
        # puts 'done'
        # STDOUT.flush
      end
    end

    # All database (and other?) set-up should be called from here.
    # Code should be idempotent, i.e., can be run multiple times with no ill effects.
    # This should preclude the need to put database updates/inserts in individual migrations
    # to initialize data
    def post_db_migrate
      Group.create_installation_default_groups
      TransferSecurity.set_transfer_path_parts_for_all_users
      TransferSecurity.create_all_security_views
      Canonicalizer.create_optimization_views
      Canonicalizer.create_optimization_table_records
      Canonicalizer.create_utility_functions
      AspRpt::ReportGenerator.delete_console_1_4_reports
      AspRpt::ReportGenerator.create_db_user
      AspRpt::ReportGenerator.create_database
      AspRpt::ReportGenerator.create_security_function
      AspRpt::ReportGenerator.create_security_views
      EmailTemplate.regenerate_template_files
      AspRpt::Template.restore_prepackaged_templates
      AspRpt::ReportGenerator.create_date_range_proc
    end
    
    # Returns the value stored in config/secret.yml
    def secret
      ActionController::Base.session_options[:secret]
    end
    
    def ip_127_0_0_1
      '127.0.0.1'
    end
    
    def ip_0_0_0_0
      '0.0.0.0'
    end
    
    def timed_operation(label, logger = nil, level = :debug)
      start = Time.now
      yield
      elapsed = Time.now - start
      message = "#{label} #{elapsed}"
      if logger
        
        logger.send(level, message)
      else
        puts message
      end
      elapsed
    end

    def proxy_config
      settings = ConsoleConfig::ConsoleProxy.load_settings

      {
          :address => settings.fasp_proxy_address,
          :port => settings.fasp_proxy_port,
          :use_ssl => settings.fasp_proxy_use_ssl,
          :login => settings.fasp_proxy_login,
          :password => settings.fasp_proxy_password,
      }
    end

    # Returns true if Console (actually the Master Control Process) is running.
    def running?
      AsperaBackground::Controller.running?
    end
  end
end
