require 'fileutils'

class DbuTask < ActiveRecord::Base
  #include AsyncTask
  include Schedulable
  include Common::ConstantHelper

  BACKUP_OPERATION = 'backup'
  PURGE_OPERATION  = 'purge'
  DATE_ONLY_FORMAT = "%Y-%m-%d"
  DATED_DIR_FORMAT = "%Y-%m-%d_%H.%M.%S_UTC"
  RATE_CHOICES = %w(day week month)
  PURGE_TYPE_JOBS = "jobs"
  PURGE_TYPE_TRANSFERS = "transfers"
  
  # This is the token name used when requesting (and holding) tokens. It is needed
  # so that we can later clean up tokens held by any failed tasks.
  DBU_TASK = 'dbu_task'
  
  INTERNAL_ERROR = "Internal error - see details page"

  before_save :before_save_hook

  named_scope :history, :conditions => ['status NOT IN (?,?,?)', Console::STATUS_QUEUED, Console::STATUS_SCHEDULED, Console::STATUS_RUNNING]
  named_scope :non_scheduled, :conditions => ['status != ?', Console::STATUS_SCHEDULED]
  named_scope :in_process, :conditions => ['status IN (?,?)',  Console::STATUS_QUEUED, Console::STATUS_RUNNING]
  
  # Pseudo-column for supporting the schedule-this-task and purge-all checkboxes
  attr_accessor :purge_all_closed, :proposed_started_at
  cattr_accessor :bg
  
  
  enum_field :status, %w(    scheduled queued running completed error cancelled    )
  STATUSES.each { |status| const_set("STATUS_#{status}".upcase, status) }
  constant_helpers('STATUS', :method_prefix => :none)
  default_value_for :status, Console::STATUS_QUEUED
  
  # TODO: move this view-related code into a helper
  # def rate_options
  #   option_list = ""
  #   RATE_CHOICES.each do |rel_rate|
  #     selected = (rel_rate == self.relative_rate) ? " selected='selected'" : ""
  #     option_list += "<option#{selected} value='#{rel_rate}'>#{rel_rate}(s)</option>"
  #   end
  #   option_list
  # end
    
  def before_validation
    if purging?
      if purge_all_closed?
        self.purge_all_completed = true
        self.purge_all_cancelled = true
        self.purge_all_errored   = true
        self.purge_all_inactive  = true
        self.purge_all_zero_byte = true
      end
    end
  end
  
  def validate_backup_base_dir
    if backup_base_dir.blank?
      errors.add("backup_base_dir","cannot be blank")
    elsif !Pathname.new(backup_base_dir).absolute?
      errors.add("backup_base_dir","must be an absolute path")
    elsif restricted_backup_directories_include?(backup_base_dir)
      errors.add("backup_base_dir","must not be a restricted directory")
    end
  end
  
  def validate
    if backing_up?
      validate_backup_base_dir
    elsif purging?
      if relative_period.blank?
        errors.add("relative_period",  "^Time frame cannot be blank")
      elsif !(relative_period.to_s =~ /^\d+$/)
        errors.add("relative_period", "^Time frame should not contain non-numeric characters") 
      else
        to_date = relative_to_date
        unless self.class.sane_date?(to_date, proposed_started_at)
          errors.add("to_date",  "^Time frame must be valid")
        end
        self.to_date = to_date.utc.strftime('%Y-%m-%d %H:%M:%S')
      end
      if backup_before_purge
        validate_backup_base_dir
      end
      # at least some transfers selected
      transfer_selectors = [ (purge_all_closed == 1), purge_all_completed, purge_all_cancelled, purge_all_errored, purge_all_inactive, purge_all_zero_byte]
      errors.add_to_base("You must select some data to purge") if transfer_selectors.none?
    end
  end
  
  def relative_to_date
    if status == Console::STATUS_SCHEDULED
      # schedule_to_date(schedule.scheduled_start_at)
      schedule_to_date
    else
      date = relative_period.to_i.send(relative_rate).ago
      #date.in_time_zone(time_zone)
    end
  end
  
  def before_save_hook
    self.backup_base_dir.strip!  if self.backup_base_dir.is_a? String
    self.to_date.strip!          if self.to_date.is_a? String
    self.description = build_description({
      :status => status, 
      :operation => operation, 
      :session_types => session_types, 
      :schedule => schedule,
    }) unless status == Console::STATUS_CANCELLED
  end
  
  def build_description(options)
    descr = ""
    if backing_up?
      location = best_final_dir_we_know
      descr += "Full backup to #{location}"
    else
      preface = "Purge"
      descr += "#{preface} #{options[:session_types]}"
      if options[:status] == Console::STATUS_SCHEDULED
        descr += " #{scheduled_purge_description_string(options[:schedule].scheduled_start_at)}"
      else
        descr += " #{purge_date_string}"
      end
    end
  end
  
  def schedule_to_date
    date = schedule.scheduled_start_at - relative_period.to_i.send(relative_rate)
    date.in_time_zone(schedule.time_zone)
  end
  
  def proposed_started_at
    if status == Console::STATUS_SCHEDULED || (schedule && schedule.initial_start_time) # we don't want to use actual time to calculate dates if this had any kind of associated schedule.
      schedule.initial_start_time
    else
      Time.zone.now
    end
  end
  
  def relative_rate
    # default to 1 month - which is a much less disastrous default than 1 day
    self[:relative_rate] ||= 'month'
  end
  
  def relative_period
    self[:relative_period] ||= 1
  end
  
  # def percentage_complete
  #   if total_count > 0
  #     percentage = current_count / total_count.to_f
  #     "#{(percentage * 100).round}%"
  #   else
  #     "Loading tables..."
  #   end
  # end
   
  def restricted_backup_directories
    dirs = ["#{Rails.root}"]
    if Aspera::OS.windows?
      dirs << "#{Rails.root}/../../Common Files/Aspera/Common"
    else
      dirs << "#{Rails.root}/../common"
    end
    dirs.map { |d| File.expand_path(d) }
  end
  
  def restricted_backup_directories_include?(absolute_backup_directory)
    backup_directory = absolute_backup_directory.gsub(/\\/,'/')
    backup_directory = File.expand_path(backup_directory)
    restricted_backup_directories.find do |d|
      if Aspera::OS.windows?
        backup_directory =~ /^#{Regexp.escape(d)}/i
      else
        backup_directory =~ /^#{Regexp.escape(d)}/
      end
    end
  end
  
  def self.add(hash, time_to_run = nil)
    return "Hash must contain some data" if hash.nil? or hash.empty?
    
    our_hash = hash.dup.symbolize_keys
    our_hash.delete :id
    task = self.new(our_hash)

    if task.save
      # Kick off this task, if it is ready, and move others on their path
      # (but not if recurrences cause an infinite loop by rescheduling again)
      # self.process_queue
      return task.id 
    else
      return task.errors.full_messages.join('; ')
    end
  end
  
  def cancellable?
    if purging?    
      ([Console::STATUS_SCHEDULED, Console::STATUS_QUEUED, Console::STATUS_RUNNING].include?(status))
    else
      ([Console::STATUS_SCHEDULED, Console::STATUS_QUEUED].include?(status))
    end
  end
  
  def cancel
    return false unless cancellable?
    self.status = Console::STATUS_CANCELLED
    self.stopped_at = Time.now
    save_with_validation(false)
    true
  end
  
  def self.process_scheduled
    ready_to_start.each do |task|
      # reobtain object because 'ready_to_start' named scope returns records that are read-only due to table join
      DbuTask.find(task.id).move_to_queued_state
    end    
  end
  
  def move_to_queued_state
    self.status = Console::STATUS_QUEUED
    self.queued_at = Time.now
    errors = nil
    begin
      unless self.save
        errors = self.errors.full_messages.join(", ")
      end
    rescue => e
      errors = "Caught exception: #{e}\n#{e.better_backtrace}"
    end
    if errors
      self.status = Console::STATUS_ERROR
      self.stopped_at = Time.now
      # self.err_desc = 'Task failed'
      self.error_text = "Converting from SCHEDULED to QUEUED failed for #{self.inspect}\n  #{errors}"
      save_with_validation(false)
    end
    next_schedule = schedule.next_occurrence
    if next_schedule
      next_task = self.clone
      next_task.schedule = next_schedule
      next_task.to_date =  schedule_to_date
      next_task.status = Console::STATUS_SCHEDULED
      next_task.save_with_validation(false)
    end
  end
  
  # NOTE: Per MP 12/17/11: Purge tasks can no longer be 'orphaned'. If 
  # they are 'running', just continue to purge.
  # -- mark as error anything marked as running 
  def self.mark_orphaned_tasks
    find_all_by_status(Console::STATUS_RUNNING).each do |task|
      if task.backing_up?
        task.status = Console::STATUS_ERROR
        task.error_text = "Console or Database Utility background process shut down while operation was running"
        task.save
      end
    end
  end
  
  # We're only concerned about adding items to the queue - the work of marking
  # them as completed is done in the notify_status() method.
  def self.process_queue(bg)
    self.bg = bg
    delete_all_except(100)
    mark_orphaned_tasks
    process_scheduled
    task = find_by_status(Console::STATUS_RUNNING, :order=>'started_at')
    if task.nil?
      task = find_by_status(Console::STATUS_QUEUED, :order=>'queued_at')
    end
    DbuTask.start_task(task.id) if task
  rescue StandardError, Timeout::Error
    logger.error "Exception: #{$!}\n#{$!.better_backtrace}"
    begin
      ActiveRecord::Base.connection.reconnect!
    rescue StandardError, Timeout::Error
      logger.error "Exception trying to reconnect to database: #{$!}\n#{$!.better_backtrace}"
      # do nothing, try to reconnect on next go around
    end
  end
  
  # backup_base_dir with Windows backslashes converted to forward slashes
  def backup_base_dir_fslash
    (Aspera::OS.windows? ? backup_base_dir && backup_base_dir.gsub(/\\/, '/') : backup_base_dir)
  end

  # def duration
  #   if end_time
  #     (end_time - started_at).to_s
  #   else
  #     ""
  #   end
  # end
  #

  def scheduled_purge_description_string(scheduled_start_at)
    datetime = scheduled_start_at -  relative_period.to_i.send(relative_rate)
    datetime = datetime.in_time_zone(schedule.time_zone)
    DbuTask.purge_cutoff_datetime_description(datetime)
  end
  
  def progress_or_description
    progress.blank? ? description : progress
  end
  
  def completion_text
    if status == STATUS_ERROR
      # This assumes that self.progress has been updated with any error synopsis 
      # when the task was completed in an error state. If we change how errors
      # get saved, update completion_text() accordingly.
      progress_or_description
    else
      description
    end
  end
  
  def best_final_dir_we_know
    # NOTE: final_full_dir won't be set until the task has started, to
    # reflect the actual start time instead of either the intended 
    # start time or the time it was scheduled.
    ### TODO: can change the name from final_full_dir to final_dir
    
    return self.final_full_dir unless self.final_full_dir.blank?

    ending_dir = [dir_suffix, ""].compact.join('_YYYY-MM-DD_hhmmss')
    if backup_base_dir
      return File.join(backup_base_dir_fslash, ending_dir )
    end
    ""
  end
      
  # include ActionView::Helpers::TextHelper  # for pluralize()
  # def relative_time_ago
  #   # pluralize is from ActionView::Helpers::TextHelper
  #   pluralize(self.relative_period, self.relative_rate)
  # end
  
  def session_types
    if purge_all_closed? or all_purge_options_selected?
      "all sessions"
    else
      array = []
      array << "completed" if purge_all_completed
      array << "cancelled" if purge_all_cancelled
      array << "errored"   if purge_all_errored
      array << "inactive"  if purge_all_inactive
      array << "zero byte" if purge_all_zero_byte
      "all #{array.to_sentence(:last_word_connector => " and ")} sessions"
    end
  end
  
  # tell the rake tasks what to do about cancelled, errorred or zero-byte transfers
  def session_selectors
    return [] if purge_all_closed? or all_purge_options_selected? # the default
    
    array = []
    array << "COMPLETED=no" unless purge_all_completed
    array << "CANCELLED=no" unless purge_all_cancelled
    array << "ERRORED=no"   unless purge_all_errored
    array << "INACTIVE=no"  unless purge_all_inactive
    array << "EMPTY=no"     unless purge_all_zero_byte
    return array
  end
  
  def all_purge_options_selected?
    purge_all_completed && purge_all_cancelled && purge_all_errored && purge_all_inactive && purge_all_zero_byte
  end
  
  def backup_dir
    dating = started_at.utc.strftime(DATED_DIR_FORMAT)
    ending_dir = [dir_suffix, dating].join('_')

    if backup_base_dir
      File.join( backup_base_dir_fslash, ending_dir )
    else
      nil
    end
  end

  def backup_dir_os_specific
    Aspera::OS.windows? ? backup_dir && backup_dir.gsub(/\//, '\\') : backup_dir
  end

  # def backup_dir_option
  #   if backup_dir
  #     "DIR=#{backup_dir}"
  #   else
  #     ""
  #   end
  # end

  # Case-sensitive comparisons were becoming a pain, so centralize here and ignore case.
  def backing_up?
    operation =~ Regexp.new( BACKUP_OPERATION, true )
  end
  
  def purging?
    operation =~ Regexp.new( PURGE_OPERATION, true )
  end

  # The purge_all_closed pseudo-column is not a true boolean column, so it needs this test function.
  def purge_all_closed?
    self.purge_all_closed == "1"
  end
  
  def repeating?
    schedule and schedule.repeat?
  end
  
  def schedule_description
    (schedule || Schedule.new).description
  end
    
  def can_edit?
    schedule.present?
  end

  def purge_date_string
    self.class.purge_date_string(
    {
      :relative_rate => relative_rate, 
      :relative_period => relative_period, 
      :time_zone => time_zone, 
      :proposed_started_at => proposed_started_at
      }
    )
  end
  
  # Delegate all ..._text() requests to the CleTask
  # def output_text
  #   cle_task.output_text rescue ""
  # end
  
  # def progress_text
  #   cle_task.progress_text(true) rescue ""
  # end
  # 
  # def get_result_code
  #   cle_task.get_result_code rescue nil
  # end

  def run
    self.class.wait_for_lock_on_all_tokens do
      run_operation
    end
  end

  def run_operation
    if backing_up?
      backup
    elsif purging?
      purge
    end
  end

  def backup
    self.status = Console::STATUS_RUNNING
    self.started_at = Time.now
    self.final_full_dir = backup_dir_os_specific
    save!
    Dbu::ExportImport::ConsoleApp.new(final_full_dir).export
    self.status = Console::STATUS_COMPLETED
    self.stopped_at = Time.now
    save_with_validation(false)
  rescue
    self.status = Console::STATUS_ERROR
    self.stopped_at = Time.now
    self.error_text = $!.to_s
    self.class.cleanup
    save_with_validation(false)
  end
  
  def check_in_and_check_for_cancel
    self.bg.check_in
    check_for_cancel
  end
  
  def check_for_cancel
    if DbuTask.find(self.id).status == Console::STATUS_CANCELLED
      raise AsperaException::CancelledTask
    end
  end

  def initialize_purge_task
    cleanup_tables
    self.status = Console::STATUS_RUNNING
    self.started_at = Time.now
    self.final_full_dir = backup_dir_os_specific if backup_before_purge
    save!    
    select_records_for_purge
    determine_value_used_for_batching
    self.total_count =  purge_task.count_of_all_rows_to_purge
    self.current_count =  purge_task.current_count_of_rows_purged
    save_with_validation(false)
    if backup_before_purge && total_count_of_values_to_purge > 0
      export_purged_records
    end
    purge_task
  end
  
  def select_records_for_purge
    purge_task.select_records
  rescue AsperaException::CancelledTask
    self.cancel  
  rescue AsperaBackground::Control::Stop
    self.bg.stop
  end
  
  def export_purged_records
    purge_task.export_database
  rescue AsperaException::CancelledTask
    self.cancel  
  rescue AsperaBackground::Control::Stop
    self.bg.stop 
  end

  # NOTE: Per MP 12/17/11: Purge tasks can no longer be 'orphaned'. If 
  # they are 'running', just continue to purge.  This allows tokens to be
  # freed and, thus, Transfer, Data Canonicalizer, etc have a chance to progress.
  def purge
    if total_count_of_values_to_purge == 0
      initialize_purge_task
    else
      Dbu::ExportImport::Transfer.truncate_purge_tables
      purge_task.purge
      self.current_count = current_count + purge_task.current_count_of_rows_purged
      self.current_count_of_values_purged = purge_task.batch_size + purge_task.current_count_of_values_purged
    end

    # only mark as  COMPLETE when the two value counts are equal
    # total_count may be less than current_count of purged values, since we 
    # use the batch_size to calculate # of records purged
    if total_count_of_values_to_purge <= current_count_of_values_purged
      self.status = Console::STATUS_COMPLETED
      self.stopped_at = Time.now
      cleanup_tables
    end
    save_with_validation(false)
  rescue AsperaException::CancelledTask
    # do nothing    
  rescue
    self.status = Console::STATUS_ERROR
    self.stopped_at = Time.now
    self.error_text = $!.to_s
    save_with_validation(false)
    cleanup_tables
  end
  
  def purge_task
    Dbu::ExportImport::Transfer.new(to_date, backup_dir, 
      :dbu_task => self,
      :timezone => time_zone,
      :total_count_of_values_to_purge => total_count_of_values_to_purge,
      :current_count_of_values_purged => current_count_of_values_purged,
      :value_used_for_batch_purge => value_used_for_batch_purge,
      :select_errored   => purge_all_errored,
      :select_inactive  => purge_all_inactive,
      :select_cancelled => purge_all_cancelled,
      :select_empty     => purge_all_zero_byte,
      :select_completed => purge_all_completed
    )
  end
  
  def cleanup_tables
    Dbu::ExportImport::Transfer.truncate_dbu_tables
    Dbu::ExportImport::Transfer.truncate_purge_tables
  end
  
  # total # of jobs to be purged and total # of transfers to be purged will not always match
  # We want the max count of one of these to use as our benchmark for batching...
  def determine_value_used_for_batching
    jobs_to_purge = purge_task.count_of_jobs_to_purge 
    transfers_to_purge = purge_task.count_of_transfers_to_purge
    self.total_count_of_values_to_purge = [transfers_to_purge, jobs_to_purge].max
    if transfers_to_purge < jobs_to_purge
      self.value_used_for_batch_purge = PURGE_TYPE_JOBS
    else
      self.value_used_for_batch_purge = PURGE_TYPE_TRANSFERS
    end
  end

  ###########################################
  # Class methods used from dbu.rake
  class << self
    
    def purge_date_string(options)
      # protect against injection of strange parameter
      options[:relative_rate] = 'month' unless %w(day week month).include?(options[:relative_rate])
      msg = "invalid date"
      purge_cutoff = nil
      Time.use_zone(options[:time_zone]) do
        begin
          purge_cutoff = options[:proposed_started_at] -  options[:relative_period].to_i.send(options[:relative_rate])
          #purge_cutoff = purge_cutoff.in_time_zone

          # check to make sure the number of days/weeks/months is present and a number
          return msg unless (options[:relative_period].to_s =~ /^\d+$/)
          # check reasonableness of date
          return msg unless sane_date?(purge_cutoff, options[:proposed_started_at])
        rescue ArgumentError, NoMethodError
          return msg
        end
        msg = purge_cutoff_datetime_description(purge_cutoff)
      end
      return msg
    end
    
    # we allow dates that are in the future, when the task occurs in the future
    def sane_date?(purge_cutoff, start_at)
      purge_cutoff.between?(20.years.ago, start_at)
    end
    
    def start_task(id)
      Dir.chdir Rails.root
      find(id).run
    end
    
    def rake_restore(rake_dir) 
      Dbu::ExportImport::ConsoleApp.new(rake_dir).import 
    end
  
    def wait_for_lock_on_all_tokens(&block)
      puts_and_log "Waiting for tokens..."
      success = AsperaBackground::Control::Token.wait_for_lock_on_tokens(
          DBU_TASK,
          AsperaBackground::Base::TOKEN_NAMES,
          :timeout => 60 ) do
        puts_and_log "Received tokens; performing requested action..."
        yield
      end
      puts_and_log "Tokens not received." unless success
      success
    end

    def delete_non_scheduled
      DbuTask.non_scheduled.destroy_all
    end

    # format for display of proposed purge cutoff
    def purge_cutoff_datetime_description(datetime)
      "before #{datetime.strftime '%B %d, %Y %H:%M:%S %Z'}"
    end

    def delete_all_except(keep)
      # :offset does not work without :limit in mysql, and we don't want to tie up the
      # bg process deleting an arbitrary number of items, so destroy only 100 old items at a time
      tasks = DbuTask.all(
        :conditions => [
          'status NOT IN (?,?,?)', 
          Console::STATUS_SCHEDULED, 
          Console::STATUS_QUEUED, 
          Console::STATUS_RUNNING
        ], 
        :order => "stopped_at DESC",
        :limit => 100,
        :offset => keep
      )
      tasks.each do |t|
        t.destroy
      end      
    end
    
    # puts will show up in the background task log; ProgressLog.log will show up in the task-specific log.
    def puts_and_log(msg)
      # puts msg
      ProgressLog.log msg
    end
    
    # Called from the controller when updating the UI to display the user's changed preference of task period.
    def calculate_proposed_started_at(hash)
      if hash[:status] == Console::STATUS_SCHEDULED
        hash[:proposed_started_at] = Schedule.new(hash[:schedule_attributes]).initial_start_time
      else
        hash[:proposed_started_at] = Time.zone.now
      end
      hash[:proposed_started_at]
    end
      
    def cleanup
      tokens = AsperaBackground::Control::Token.find_all_by_holder( DBU_TASK )
      tokens.each do |token|
        token.remove_holder
      end
    end
  end
end
