class EmailTemplate < ActiveRecord::Base
  include EmailMethods

  # Valid types in Console::EMAIL_TYPES

  named_scope :start,   :conditions => {:template_type => Console::EMAIL_TYPE_START}
  named_scope :success, :conditions => {:template_type => Console::EMAIL_TYPE_SUCCESS}
  named_scope :error,   :conditions => {:template_type => Console::EMAIL_TYPE_ERROR}
  named_scope :default, :conditions => {:is_default => true}

  validates_uniqueness_of :name, :scope => :template_type
  validates_presence_of :name
  validates_presence_of :plain_body, :html_body, :message => "^Body can't be blank"
  validates_length_of :name, :within => 1..25, :allow_blank => true
  validates_length_of :from_name, :within => 1..60, :allow_blank => true
  validates_length_of :reply_to_address, :within => 5..255, :allow_blank => true
  validates_format_of :reply_to_address, :with => VALID_EMAIL_REGEX, :message => 'is not a valid email address', :allow_blank => true
  validates_length_of :subject, :within => 1..255
  validate :subject_has_no_forbidden_variables

  before_save :sanitize_strings
  after_save :generate_template_files
  before_destroy :check_last_and_default
  after_destroy :reset_references_to_default
  after_destroy :erase_template_files

  DIRECTORY = File.dirname(__FILE__) + "/../views/notifier"
  TEMPLATE_DIRECTORY = "#{DIRECTORY}/templates"
  MAILER_DIRECTORY = 'notifier/templates'
  CLASSES_WITH_TEMPLATES = [JobStepEmail, JobTemplateEmail, JobTemplateStepEmail, User, LoginEmail]

  # 1, 2, or 3 in third element indicates if it is a start, success, or error variable
  VARS = [
     ['DESCRIPTION', 'Description of the transfer', '123'],
     ['LAST_SESSION_ID', 'Unique ID of the most recent transfer session', '123'],
     ['CONTACT', 'Login name of user who initiated the transfer', '123'],
     ['SESSION_ID', 'Unique ID for most recent session started for transfer', '123'],
     ['STARTED_AT', 'Time the transfer began', '123'],
     ['STOPPED_AT', 'Time the transfer stopped', '23'],
     ['SOURCE_NODE', 'Source node of the transfer', '123'],
     ['SOURCE_ENDPOINT', 'Source endpoint of the transfer', '123'],
     ['FIRST_5_SOURCE_PATHS', 'First five source paths being transferred', '123'],
     ['ALL_SOURCE_PATHS', 'All source paths being transferred (up to 4000 chars)', '123'],
     ['DESTINATION_NODE', 'Destination node of the transfer', '123'],
     ['DESTINATION_ENDPOINT', 'Destination endpoint of the transfer', '123'],
     ['DESTINATION_PATH', 'Target path on the destination node for the transfer', '123'],
     ['COMMENT', 'Comment for transfer', '123'],
     ['EXPECTED_TOTAL_BYTES', 'Total number of bytes at start of transfer (may be misleading for Windows push hot folders)', '123'],
     ['BYTES_TRANSFERRED', 'Total number of bytes transferred', '23'],
     ['ELAPSED_TIME', 'Time elapsed for transfer', '23'],
     ['AVERAGE_RATE', 'Average transfer rate', '23'],
     ['FILES_TRANSFERRED', 'Number of files transferred', '23'],
     ['EXPECTED_TOTAL_FILES', 'Total number of files at start of transfer (may be misleading for Windows push hot folders)', '123'],
     ['FILES_SKIPPED', 'Number of files skipped during transfer', '23'],
     ['FILES_FAILED', 'Number of files that failed to be transferred', '23'],
     ['FIRST_10_FILEPATHS_W_CHECKSUMS', 'Paths and checksums (if available) of the first 10 files transferred', '23'],
     ['ALL_FILEPATHS_W_CHECKSUMS', "Paths and checksums (if available) of all transferred files (up to configured limit: #{Console.email_file_list_max_length})", '23'],
     ['FIRST_10_FILENAMES', 'Names of the first 10 files transferred', '23'],
     ['ALL_FILENAMES', "Names of all transferred files (up to configured limit: #{Console.email_file_list_max_length})", '23'],
     ['FIRST_10_FILEPATHS', 'Paths of the first 10 files transferred', '23'],
     ['ALL_FILEPATHS', "Paths of all transferred files (up to configured limit: #{Console.email_file_list_max_length})", '23'],
     ['ERROR_MESSAGE', 'Error message for transfer', '3'],
     ['SOURCE_FAILOVER_GROUP', 'Failover group of the source node', '123'],
     ['DEST_FAILOVER_GROUP', 'Failover group of the destination node', '123'],
  ]
  
  FORBIDDEN_SUBJECT_SUBSTITUTION_VARIABLE_NAMES = %w(
    FIRST_5_SOURCE_PATHS
    ALL_SOURCE_PATHS
    COMMENT
    FIRST_10_FILENAMES
    ALL_FILENAMES
    FIRST_10_FILEPATHS
    ALL_FILEPATHS
    FIRST_10_FILEPATHS_W_CHECKSUMS
    ALL_FILEPATHS_W_CHECKSUMS
  )
  
  def short_name
    name[0, 26]
  end
  
  def default?
    is_default?
  end
  
  def set_as_default
    EmailTemplate.send(template_type).each { |e| e.update_attribute(:is_default, false) }
    update_attribute(:is_default, true)
	end

  # Absolute path to subject template file
  def subject_template_path
    template_path_for('subject')
  end
  
  # Absolute path to html template file
  def html_template_path
    template_path_for('html')
  end
  
  # Absolute path to plain template file
  def plain_template_path
    template_path_for('plain')
  end

  # Path required by ActionMailer to subject file
  def subject_mailer_path
    mailer_path_for('subject')
  end
  
  # Path required by ActionMailer to html file
  def html_mailer_path
    mailer_path_for('html')
  end
  
  # Path required by ActionMailer to plain file
  def plain_mailer_path
    mailer_path_for('plain')
  end

  def generate_template_files(force = false)
    generate_subject_template(force)
    generate_html_template(force)
    generate_plain_template(force)
  end

  def substitution_variables
    type_number = Console::EMAIL_TYPES.index(template_type) + 1
    # sort results by length descending so that if one substitution variable is a subset of another it will not supersede
    # the longer one and leave fragment of variable name behind
    VARS.select{ |var| var[2].include?(type_number.to_s) }.sort {|a,b| b[0].length <=> a[0].length}
  end

  def substitution_variable_names
    substitution_variables.map { |var| var[0] }
  end

  def generate_subject_template(force)
    if force || subject_changed? || !File.exists?(subject_template_path) || File.open(subject_template_path).stat.mode != 0644
      File.open(subject_template_path, 'w', 0644) { |f| f.write(subject_template_content) }
    end
  end

  def subject_template_content
    substitute_erb_tags(subject, 'h')
  end

  def generate_html_template(force)
    if force || html_body_changed? || !File.exists?(html_template_path) || File.open(html_template_path).stat.mode != 0644
      File.open(html_template_path, 'w', 0644) { |f| f.write(html_template_content) }
    end
  end

  def html_template_content
    substitute_erb_tags(html_body, 'h_html')
  end

  def generate_plain_template(force)
    if force || plain_body_changed? || !File.exists?(plain_template_path) || File.open(plain_template_path).stat.mode != 0644
      File.open(plain_template_path, 'w', 0644) { |f| f.write(plain_template_content) }
    end
  end

  def substitute_erb_tags(template, prefix)
    substitution_variable_names.each do |var|
      template.gsub!(var, "<%= #{prefix}_#{var.downcase} %>")
    end
    template
  end

  def plain_template_content
    substitute_erb_tags(plain_body, 'h').
      gsub(/<a.*?href.*?(["'])(.*?)\1.*?>(.*?)<\/a>/, '\3 (URL: \2)'). # urls
      gsub(/<[^%]*?>/, '')      # throw out html (non-ERb) tags
  end

  def erase_template_files
    [subject_template_path, html_template_path, plain_template_path].each do |file|
      FileUtils.rm_f(file)
    end
  end

  def check_last_and_default
    raise(AsperaException::TemplateDelete, "Cannot delete default template") if default?
    raise(AsperaException::TemplateDelete, "Cannot delete last remaining template") if
        EmailTemplate.find_all_by_template_type(template_type).size == 1
  end

  # after_delete callback.  Finds any references to this template and sets
  # them to the default for template_type.
  def reset_references_to_default
    logger.info 'Resetting template references to defaults after delete'
    default_template_id = EmailTemplate.default_template_id(template_type)
    column_name = "#{template_type}_template_id"
    updates = "#{column_name} = #{default_template_id}"
    conditions = "#{column_name} = #{id}"
    CLASSES_WITH_TEMPLATES.each do |klass|
      logger.info "Resetting templates for #{klass}"
      rows_updated = klass.update_all(updates, conditions)
      logger.info "Reset #{rows_updated} rows for #{klass}"
    end
  end

  class << self

    def clone_default(template_type)
      cloned_template = default_template(template_type).clone
      cloned_template.attributes = {
        :name => name_for_new(template_type),
        :is_default => false,
        :created_at => nil,
        :updated_at => nil,
      }
     cloned_template.save!
    end

    def name_for_new(template_type)
      i = 0
      while true
        i += 1
        name = "new_#{template_type}_template_#{i}"
        return name unless find_by_name(name)
      end
    end

    def options_for(type, include_none_option = true)
      options = send(type).sort_by(&:short_name)
      # Move default to front of array
      if default_template = options.select { |t| t.default? }.first
        default_template.name =  "(default) " + default_template.name[0..18]
        options.delete(default_template)
        options.unshift(default_template)
      end

      options.map! { |t| [t.short_name, t.id] }
      options.unshift(["No notification",0]) if include_none_option
      options
    end

    # Returns the id of the default template for notification _template_type_.
    def default_template_id(template_type)
      default_template(template_type).id rescue nil
    end

    def default_template(template_type)
      send(template_type).default.first
    end

    # returns template with id, or, if not found, the default for the specified type
    def find_or_default(id, template_type)
      EmailTemplate.find(:first, :conditions => ["id=?",id]) || EmailTemplate.find(EmailTemplate.default_template_id(template_type))
    end

    # Returns sql statements for inserting the delivered email templates.
    def delivered_templates_sql
      path = "#{DIRECTORY}/template_sql.txt"
      IO.read(path).reject { |line| line.starts_with?('#') }
    end

    # Loads delivered email templates to db, generates template files and sets them to default.
    def load_delivered_templates
      template_types = { 1 => 'start', 2 => 'success', 3 => 'error' }
      template_ids = template_types.keys

      ActiveRecord::Base.transaction do
        EmailTemplate.update_all('email_templates.is_default = false')
        EmailTemplate.delete(template_ids)
        template_ids.each do |id|
          insert_template(id, template_types[id])
          find(id).generate_template_files(true)
        end
      end
    end

    def insert_template(id, type)
      plain_body = connection.quote(IO.read("#{DIRECTORY}/#{type}.plain.erb"))
      html_body  = connection.quote(IO.read("#{DIRECTORY}/#{type}.html.erb"))
      sql = %W[
        insert into email_templates (
               id
             , template_type
             , name
             , is_default
             , subject
             , plain_body
             , html_body
             , created_at
             , updated_at
             ) values(
               #{id}
             , '#{type}'
             , 'Default #{type.capitalize}'
             , true
             , 'Aspera Console Transfer Notification - #{type.capitalize}'
             , #{plain_body}
             , #{html_body}
             , utc_timestamp()
             , utc_timestamp()
          ) ].join(' ')
      ActiveRecord::Base.connection.execute(sql)
    end

    # Regenerate template files from content in the database.
    def regenerate_template_files
      all.each { |template| template.generate_template_files }
    end

    # Delete template files generated during RSpec testing.
    def delete_test_template_files
      FileUtils.rm_f Dir["#{EmailTemplate::TEMPLATE_DIRECTORY}/test-*"]
    end

  end

  # Send a test email using the given template, using dummy data
  def send_test_email_from_template(recipient)
    job_step = EmailPreviewJobStep.new.job_step

    # Send email
    Notifier.deliver_email({
                               :template => self,
                               :time_zone => JobStepEmail.time_zone_for_email(recipient),
                               :recipients => recipient,
                               :job_step => job_step
                           })
    if $email_test_output =~ /EMAIL: EXCEPTION:/
      logger.error $email_test_output
    end

    return $email_test_output
  end

  private

  def subject_has_no_forbidden_variables
    FORBIDDEN_SUBJECT_SUBSTITUTION_VARIABLE_NAMES.each do |name|
      if subject.include?(name)
        msg = '^Subject cannot contain variables that produce multi-line output: ' +
            FORBIDDEN_SUBJECT_SUBSTITUTION_VARIABLE_NAMES.join(", ")
        errors.add(:subject, msg)
        return
      end
    end
  end

  def sanitize_strings
    [:subject, :html_body, :plain_body].each do |column|
      sanitize_column(column)
    end
  end

  def sanitize_column(column)
    if send("#{column}_changed?")
      string = send(column)
      send("#{column}=", ::ActionController::Base.helpers.sanitize(
          string,
          :tags => ActionView::Base.sanitized_allowed_tags + %w(!-- body head html meta style table td th tr),
          :attributes => ActionView::Base.sanitized_allowed_attributes + %w(align bgcolor border cellpadding cellspacing charset content http-equiv lang style type valign)
      ))
    end
  end

  def template_path_for(content_type)
    test_prefix = $TESTING ? 'test-' : ''
    "#{TEMPLATE_DIRECTORY}/#{test_prefix}#{template_type}_#{id}.text.#{content_type}.erb"
  end

  def mailer_path_for(content_type)
    test_prefix = $TESTING ? 'test-' : ''
    "#{MAILER_DIRECTORY}/#{test_prefix}#{template_type}_#{id}.text.#{content_type}.erb"
  end

end
