class FaspNode < ActiveRecord::Base
  include FaspNode::NodeConfiguration
  include ActionView::Helpers::NumberHelper

  # reserved values for node failover group name
  NODE_GROUP_NAME_NONE = '(none)'
  NODE_GROUP_NAME_NEW = 'enter new name...'

  REQUIRED_NODE_VERSION = '3.4.6.88391'
  REQUIRED_REPORTED_LOGGER_VERSION = '3.4.5.88375'

  include NodeBase

  concerned_with :background, :callbacks, :cluster, :connect, :health_check, :map_and_chart, :mysql, :security, :soap, :ssh_tunnel, :validations, :errors

  has_one :managed_node, :dependent => :destroy, :foreign_key => :foreign_key_id

  has_many :config_settings, :class_name => "Setting", :dependent => :delete_all
  has_many :account_settings, :class_name => "Setting", :dependent => :delete_all

  has_many :job_steps_to_submit, :class_name => "JobStep", :foreign_key => :submit_node_id, :conditions => ["to_submit=TRUE AND status='queued'"]

  has_many :source_template_steps, :class_name => "JobTemplateStep", :foreign_key => :source_node_id, :include => :job_template
  has_many :dest_template_steps, :class_name => "JobTemplateStep", :foreign_key => :dest_node_id, :include => :job_template
  has_many :sessions, :class_name => "FaspSession", :foreign_key => :node_id
  has_many :node_users # view
  has_many :users, :through => :node_users
  has_many :endpoints, :class_name => "Login", :as => :owner, :dependent => :destroy
  has_many :aspera_syncs, :dependent => :destroy
  has_many :aspera_watchfolders, :dependent => :destroy

  has_one :transfer_node, :foreign_key => :foreign_key_id, :conditions => ["type = 'ManagedNode'"]

  validates_presence_of :node_name, :node_address
  validates_uniqueness_of :node_address, :node_name
  validates_length_of :node_name, :in => 1..100, :allow_blank => true
  validates_format_of :node_address, :with => Console::VALID_IP_OR_DNS_OR_FQDN_REGEXP, :allow_blank => true, :unless => :atc?
  validates_port :ssh_port, :unless => :atc?

  validates_port :v3_node_daemon_port

  validates_presence_of :node_group_name, :if => :enable_node_group?,
                        :message => '^Failover group name can\'t be blank'

  validates_presence_of :v3_node_username, :on => :update, :if => :v3_edit_credentials?
  validates_presence_of :v3_node_password, :on => :update, :if => :v3_edit_credentials?
  validates_presence_of :v3_node_password_confirmation, :on => :update, :if => :v3_edit_credentials?
  validates_confirmation_of :v3_node_password, :on => :update, :if => :v3_edit_credentials?

  # before_validation_on_create :license_check_create

  after_save :update_login_addresses, :mirror_to_managed_node
  before_save :update_v3_node_password, :if => :v3_edit_credentials?
  before_save :update_v3_port, :if => :v3_node_daemon_port_changed?
  before_save :save_old_node_address_if_changed

  # Must wait until after save to homogenize
  before_save :set_homogenization_needed
  after_save :homogenize_cluster_endpoints!, :if => :to_be_homogenized?

  HUMANIZED_ATTRIBUTES = {
      :config_database_address => "Database Address",
      :config_database_port => "Database Port",
      :config_soap_address => "Transfer Server Address",
      :config_soap_port => "Transfer Server Port",
      :v3_node_daemon_port => "Node Port",
      :v3_node_username => "Node Username",
      :v3_node_password => "Node Password",
      :v3_node_password_confirmation => "Node Password Confirmation",
  }

  # to support node_file_data_collector
  attr_accessor :temp_session_node_uuid_hash

  # to support updating personal logins after changing node address
  attr_accessor :old_node_address

  # For #create() do we create default groups/permissions?
  attr_accessor :create_default_groups

  # to support using form builder to create and accept the 'covert to nonlegacy node' checkbox
  attr_accessor :convert_to_nonlegacy_node

  # Holds results of get from asconfigurator
  attr_accessor :raw_config_response

  # Holds asconfigurator error messages
  attr_accessor :asconfigurator_errors

  attr_accessor :configuration_password
  attr_accessor :node_api_password

  named_scope :managed, :conditions => "node_uuid IS NOT NULL AND node_uuid != ''"
  named_scope :placed_on_map, :conditions => ["map_visible = ?", true]
  named_scope :duplicate_uuid,
              :conditions => ["node_uuid IN (SELECT node_uuid FROM fasp_nodes WHERE node_uuid IS NOT NULL GROUP BY node_uuid HAVING COUNT(id)>1)"],
              :order => "node_uuid"

  named_scope :monitored, :conditions => "legacy_node=FALSE AND disabled=FALSE"

  default_value_for :ssh_port, 22

  def type_string
    self.atc? ? 'Cluster' : 'Managed'
  end

  def atc?
    self.is_a?(ClusterNode)
  end

  # whether this is a non-legacy node that cannot be used to start transfers due to lack of impersonation ACL
  def acl_problem?
    !legacy_node && !impersonation_enabled
  end

  def valid_login_types
    if atc?
      Login::LOGIN_TYPES_CLUSTER
    else
      Login::LOGIN_TYPES_MANAGED
    end
  end

  def legacy_conversion_problem_templates
    templates.find_all {|t| t.convert_legacy_acl_problem?(self)}
  end

  # did node supply us with a value for this setting when we queried asconfigurator?
  def setting_present_in_node_reply?(spec)
    config_settings.find_by_name(spec[:name].to_s).present?
  end


  def save_old_node_address_if_changed
    self.old_node_address=self.node_address_was
  end

  def update_login_addresses
    endpoints.each { |e|
      if e.address != self.node_address
        e.address = self.node_address
        e.endpoint_type = self.default_endpoint_type if e.star_at_node?
        e.save!
      end
    }

    if config_login_exists? && config_login.address != self.node_address
      config_login.address = self.node_address
      config_login.save!
    end
    if node_api_login_exists?
      node_api_login.address = self.node_address
      node_api_login.use_ssl = self.v3_node_use_ssl
      node_api_login.validate_ssl_cert = self.v3_validate_ssl_cert
      node_api_login.save!
    end

    #update personal logins for node
    if node_address != old_node_address
      personal_logins = Login.find(:all, :conditions => ["owner_type='User' AND address=?", old_node_address])
      personal_logins.each {|pl|
        pl.address = self.node_address
        pl.save!
      }
    end

  end

  # alias to allow common code with PartnerNode
  def address
    node_address
  end

  def soap_error_steps
    return [] unless soap_error?

    statuses = [JobStep::STATUS_RUNNING, JobStep::STATUS_PAUSED]
    conditions = ["status in (?) and (source_node_id = ? or dest_node_id = ?)", statuses, id, id]
    JobStep.all(:conditions => conditions, :order => 'started_at desc')
  end

  def active_sessions_for_user(user)
    @active_sessions_for_user ||= begin
      steps = user.job_steps.possibly_active
      steps.map { |step|
        (step.q_waiting? && !step.q_front?) ? nil : step.sessions.all(:conditions => {:node_id => id, :status => FaspSession::STATUSES_POSSIBLY_ACTIVE})
        #step.sessions.all(:conditions => {:node_id => id, :status => FaspSession::STATUSES_POSSIBLY_ACTIVE})
      }.flatten.compact
    end
  end

  def javascript_safe_node_name
    node_name.to_s.escape_javascript
  end

  def user_groups_with_vlinks
    user_groups.select(&:has_vlink?)
  end

  def active_vlink_sessions
    @active_vlink_sessions ||= sessions.active - user_groups_with_vlinks.map(&:active_sessions).flatten
  end

  # Define <=> method on #sortable_address.
  def <=>(other)
    Console.sortable_ip_address(node_address) <=> Console.sortable_ip_address(other.node_address)
  end

  def rate_totals(sessions = sessions.active)
    FaspSession.rate_totals sessions
  end

  def version_tuple
    reported_logger_version.nil? ? [] : reported_logger_version.split('.').map { |v| v.to_i }
  end

  def formatted_version_tuple
    vt = version_tuple
    return '' if vt.blank?

    "#{vt[0]}.#{vt[1]}.#{vt[2]}.#{vt[3]}"
  end

  def version_at_least?(other_version)
    return false if version_tuple.blank?
    other_version_tuple = other_version.split('.').map(&:to_i)
    other_version_tuple.each_with_index do |other_el, i|
      el = version_tuple[i]
      return false if el < other_el
      return true if el > other_el
    end
    true
  rescue
    false
  end

  def supports_encryption_at_rest?
    version_tuple.length > 1 && (version_tuple[0] > 2 || (version_tuple[0] == 2 && version_tuple[1] >= 2))
  end

  def private_key_filename
    File.expand_path(File.join(File.dirname(__FILE__), '..', '..', 'config', 'private_keys', self.node_address + '.putty'))
  end

  def private_key_present?
    FileTest.exists?(self.private_key_filename)
  rescue
    false
  end

  def request_stats_update
    AsperaActionRequest.request_stats_update_for_node(id)
  end

  def update_box_config(attributes = {})
    load_config_settings
    bc = box_config
    bc.database
    bc.attributes = attributes
    bc.save
  end

  def mirror_to_managed_node
    self.build_managed_node unless managed_node
    # puts "#{__LINE__} #{managed_node.inspect}."
    # puts "#{__LINE__} #{self.inspect}."

    managed_node.uuid = self.node_uuid
    managed_node.primary_address = self.node_address

    # Debug and fix an upgrade migration failure - which has two oddities:
    # 1) node_name is seen as an inaccessible private method for no apparent reason
    # 2) The two workarounds for overcoming its privacy give different answers:
    #    self.send(:node_name)  always gives "127.0.0.1"
    #    self[:node_name]       gives a node-specific answer (e.g. "linux test node100.103") -- which I think is the expected value!
    #
    # puts "#{__LINE__} getting nn..."
    # nn = self.send(:node_name)
    # puts "#{__LINE__} #{self.id}: #{nn} -- where does this one come from??? (and why is self.node_name seen as private!?)."
    nn = self[:node_name]
    # puts "#{__LINE__}    #{nn}."
    managed_node.name = nn

    managed_node.show_on_map = self.map_visible
    managed_node.map_x = self.map_x
    managed_node.map_y = self.map_y
    managed_node.ssh_port = self.ssh_port

    # managed_node.use_ssh_tunnel = !self.ssh_tunnel_port.nil?
    # managed_node.ssh_tunnel_port = self.ssh_tunnel_port
    managed_node.mgmt_username = self.mgmt_username
    managed_node.central_restart_needed = self.attributes[:central_restart_needed] || false
    #    managed_node.soap_active_sessions = self.reported_active_sessions

    managed_node.reported_logger_version = self.reported_logger_version
    managed_node.reported_os = self.reported_os
    managed_node.reported_os_version = self.reported_os_version
    managed_node.reported_hostname = self.reported_hostname
    managed_node.reported_central_uri = self.reported_central_uri
    managed_node.reported_at = self.reported_at
    managed_node.reported_by = self.reported_from

    begin
      ActiveRecord::Base.with_conditional_retry(:errors => [:deadlock, :lock_wait_timeout]) {
        managed_node.save!
      }
    rescue
      raise unless Rails.env == 'production'
      logger.error("Exception: #{$!} #{$!.backtrace.join("\n")}")
    end
  end

  def ssh_encryption_algorithm
    ssh_encryption
  end

  def v3_node_password_confirmation
    @v3_node_password_confirmation
  end

  def v3_node_password_confirmation=(value)
    @v3_node_password_confirmation = value
  end

  def v3_node_password
    @v3_node_password
  end

  def v3_node_password=(value)
    @v3_node_password = value
  end

  def v3_edit_credentials
    !!@v3_edit_credentials
  end

  def v3_edit_credentials?
    v3_edit_credentials
  end


  def v3_edit_credentials=(value)
    @v3_edit_credentials = !!value
  end


  def v3_node_protocol
    v3_node_use_ssl ? 'https' : 'http'
  end

  def v3_node_base_url
    "#{v3_node_protocol}://#{node_address}:#{v3_node_daemon_port}"
  end

  def v3_node_async_base_url
    "#{v3_node_base_url}/async"
  end

  def v3_node_info_url
      "#{v3_node_base_url}/info"
  end

  def v3_node_async_counters_url
    "#{v3_node_base_url}/url"
  end

  def v3_node_async_summary_url
    "#{v3_node_async_base_url}/summary"
  end

  def v3_node_async_list_url
    "#{v3_node_async_base_url}/list"
  end

  def v3_node_async_files_url
    "#{v3_node_async_base_url}/files"
  end

  def v3_node_entitlement_license_url
    "#{v3_node_base_url}/license"
  end

  def soap_connected?
    node_status.success?
  end

  def node_api_connected?
    node_api_connected
  end

  def node_api_available?
    node_api_connected? # && node_api_enabled?
  end

  def node_info_query_available?
    node_api_available? && soap_connected?
  end

  def async_reporting_enabled?
    async_reporting_enabled
  end

  def async_available?
    !legacy_node || (legacy_node_api_enabled && node_api_connected? && async_reporting_enabled?)
  end

  def watchfolders_available?
    !legacy_node
  end

  def has_node_api_credentials?(params=nil)
    v3_node_username_was.present?
  end

  def has_node_api_credentials_in_params?(params)
    (params && params[:node] && params[:node][:v3_edit_credentials]=="on")
  end

  def no_node_api_credentials?
    !has_node_api_credentials?
  end

  def deactivate_all_aspera_syncs
    self.connection.execute("UPDATE aspera_syncs SET active = 0 WHERE fasp_node_id = #{self.id}")
  end

  def deactivate_all_aspera_watchfolders
    self.connection.execute("UPDATE aspera_watchfolders SET active = 0 WHERE fasp_node_id = #{self.id}")
  end

  def cached_bitrate(direction)
    direction == 'inbound' ? total_inbound_bitrate : total_outbound_bitrate
  end

  def node_group_enabled?
    (node_group_name.present? && enable_node_group)
  end

  def in_group_with?(other_node)
    return false if other_node.nil?

    if node_group_enabled? && other_node.node_group_enabled?
      if node_group_name == other_node.node_group_name
        return true
      end
    end

      false
  end

  def to_be_homogenized?
    @to_be_homogenized
  end

  def set_homogenization_needed
    @to_be_homogenized = node_group_enabled? && (new_record? || node_group_name_changed? || enable_node_group_changed?)

    return true
  end

  def validate_connect(connected)
    if !connected
      if (!((v3_node_daemon_port.to_i.to_s == v3_node_daemon_port) && (v3_node_daemon_port.to_i < 65536)))
        errors.add(:v3_node_daemon_port, "must be numeric and less than 65536.")
      end
      if (v3_node_password.blank?||v3_node_password_confirmation.blank?)
        errors.add(:v3_node_password, "can't be blank")
        errors.add(:v3_node_password_confirmation, "can't be blank")
      elsif (v3_node_password!=v3_node_password_confirmation)
        errors.add(:v3_node_password, "doesn't match confirmation")
        errors.add(:v3_node_password_confirmation, "doesn't match confirmation")
      else
        # errors.add(:v3_node_username, "could not connect with this username/password combination.")
        # errors.add(:v3_node_password, "could not connect with this username/password combination.")
        # errors.add(:v3_node_password_confirmation, "could not connect with this username/password combination.")
      end
    end
  end

  def supports_newgen_node_api?
    self.version_at_least?("3.8.0")
  end

  class << self

    def request_stats_update(nodes)
      nodes.each { |node| node.request_stats_update }
    end

    # sd: Tried adding a version of this to AR::Base but forgot why I abandoned it.
    # See ConsoleConfig::Base.human_attribute_name for how it might look in AR::Base
    def human_attribute_name(attribute)
      HUMANIZED_ATTRIBUTES[attribute.to_sym] || attribute.to_s.humanize
    end

  end

# protected
#
#   def license_check_create
#     return true
#     if Utils.managed_node_limit <= FaspNode.count
#       self.errors.add_to_base("Creating another node would exceed your license limit (#{Utils.managed_node_limit}).")
#       return false
#     end
#   end

end
