class FaspSession
  #include NewRelic::Agent::Instrumentation::ControllerInstrumentation

  UNNAMED_APPLICATION = 'outside application'

  # class methods

  class << self
    #include NewRelic::Agent::Instrumentation::ControllerInstrumentation

    def time(label)
      start = Time.now
      yield
      "#{label}: #{Time.now - start}"
    end

    def process_new_logger_rows
      #TODO MP
      #AsperaBackground::DatabaseIngest.logger.info time("exclude_new_empty_sessions") {exclude_new_empty_sessions}

      AsperaBackground::DatabaseIngest.logger.info time("process_new_sessions") {process_new_sessions}
      AsperaBackground::DatabaseIngest.logger.info time("process_new_files") {process_new_files}
      AsperaBackground::DatabaseIngest.logger.info time("process_new_rates") {process_new_rates}
    end
    #add_transaction_tracer :process_new_logger_rows, :category => :task

    # preemptively marks new empty sessions for deletion (if console is configured to do so) to prevent doing
    # all the work to assign job_step / job etc if we're going to delete it anyway
    # (Note that this is only safe to do if there is not a JobStep waiting to be linked to this session.
    # If the transfer was started from the Console, or if this was a retry, then there is a JobStep record waiting)
    def exclude_new_empty_sessions
      raise "not implemented yet"
    end

    # Assigns FaspSession's without a JobStep to a JobStep, possibly creating a JobStep in the process.
    def process_new_sessions
      sessions = new_sessions
      sessions.each do |session|
        logger.info("new session FaspSession.id=#{session.id}")
        begin
          session.calculate_column_values
          session.assign_to_step
        rescue SignalException
          # allow interruption
          raise
        rescue Exception
          logger.error "#{$!}\n#{$!.backtrace.join("\n")}"
        end
      end
      logger.info "Processed #{sessions.size} new sessions"
      sessions.size
    end

    # Returns FaspSessions that have not been assigned to a JobStep.  Sessions are sorted
    # such that (usually) initiator (client) sessions are processed last because client sessions are
    # more authoritative than server sessions.
    def new_sessions
      limit = AsperaBackground::DatabaseIngest.batch_size || 100

      # since MySQL 5.1, primary key autoincrement field may be discontiguous, gaps can be wide

      # TODO MP
      # Possible future optimization: don't process sync sessions until they show some bytes transferred or they end
      # (alternate query for ids_to_process)
      #"SELECT
      #  id
      #FROM
      #  fasp_sessions USE INDEX (idx_job_step_id)
      #WHERE
      #  job_step_id IS NULL
      #  AND marked_for_delete = 0
      #  AND (
      #    bytes_transferred > 0
      #    OR status <> 'running'
      #    OR cookie IS NULL
      #    OR cookie NOT LIKE 'aspera.sync%'
      #  )"

      ids_to_process  = FaspSession.find(:all, :conditions => {:job_step_id => nil, :marked_for_delete => false}, :select => [:id], :limit => limit, :order => 'id').map(&:id)
      first_id = ids_to_process[0] || 0
      last_id = ids_to_process[-1] || 0

      conditions = {:id => (first_id..last_id), :job_step_id => nil, :marked_for_delete =>  false}
      order = 'IF(initiator="Local", 0, 1), id'
      sessions = all(:conditions => conditions, :order => order)
      logger.debug "#{sessions.size} new sessions yet to be processed"
      sessions
    end

    # Assigns Job, JobStep, FaspSession foreign keys to new FaspRate rows.
    def process_new_rates
      update_foreign_keys FaspRate
    end

    # Assigns Job, JobStep, FaspSession foreign keys to new FaspFile rows.
    def process_new_files
      #update_foreign_keys FaspFile
    end

    # Updates job, step, and session ids for fasp_files and fasp_rates.
    def update_foreign_keys(klass, limit = 333)
      table = klass.table_name

      id_sql = %W[
        SELECT t.id
          FROM #{table} t
          JOIN fasp_sessions fs
            USING (node_uuid, session_id)
          WHERE t.fasp_session_id IS NULL
         ORDER BY t.id
         LIMIT #{limit}
      ].join(' ')
      unprocessed_ids = connection.select_values(id_sql)

      update_sql = %W[
        UPDATE #{klass.table_name} t
          INNER JOIN fasp_sessions fs
            USING (session_id, node_uuid)
        SET t.fasp_session_id = fs.id,
          t.job_step_id = fs.job_step_id,
          t.job_id = fs.job_id
        WHERE t.id IN #{unprocessed_ids.in}
      ].join(' ')
      connection.execute_with_conditional_retry({:errors => [:deadlock, :lock_wait_timeout]}, update_sql)

      id_sql = %W[
        SELECT id
          FROM #{table}
         WHERE job_step_id IS NULL
         ORDER BY id
         LIMIT #{limit}
      ].join(' ')
      unprocessed_ids = connection.select_values(id_sql)

      update_sql = %W[
        UPDATE #{klass.table_name} t
          INNER JOIN fasp_sessions fs
            USING (session_id, node_uuid)
        SET t.fasp_session_id = fs.id,
          t.job_step_id = fs.job_step_id,
          t.job_id = fs.job_id
        WHERE t.id IN #{unprocessed_ids.in}
      ].join(' ')
      connection.execute_with_conditional_retry({:errors => [:deadlock, :lock_wait_timeout]}, update_sql)

      id_sql = %W[
        SELECT id
          FROM #{table}
         WHERE job_id IS NULL
         ORDER BY id
         LIMIT #{limit}
      ].join(' ')
      unprocessed_ids = connection.select_values(id_sql)

      update_sql = %W[
        UPDATE #{klass.table_name} t
          INNER JOIN fasp_sessions fs
            USING (session_id, node_uuid)
        SET t.fasp_session_id = fs.id,
          t.job_step_id = fs.job_step_id,
          t.job_id = fs.job_id
        WHERE t.id IN #{unprocessed_ids.in}
      ].join(' ')
      connection.execute_with_conditional_retry({:errors => [:deadlock, :lock_wait_timeout]}, update_sql)
  end

    def trace(msg)
      logger.info { ("%12.5f - " % Time.now.to_f) + msg }
    end

  end

  # instance methods

  # Some FaspSession columns are set by the database logger and others are calculated.
  # This method sets the values for the calculated columns as part of the background
  # process.
  def calculate_column_values
    calculate_columns_for_new unless job_step_id?
    logger.debug("Session values after calculation FaspSession.id=#{id}") { self.inspect }
  end

  # Set calculated columns that are set once when the record is first seen.
  def calculate_columns_for_new
    set_node_columns
    set_other_node_columns
    save_with_validation(false)
  end

  def set_node_columns
    node = FaspNode.find_by_node_uuid(node_uuid, :order => 'reported_active_sessions DESC, id') # in case we have duplicate node UUIDs, pick the most active node to improve chance of picking right node
    if node
      self.node_id = node.id
      self.node_address = node.node_address
    end

    # if a record was reported by other side, make sure it has proper other_node info
    # (in cases where there is NAT or VIP between the two sides of the transfer, the server_addr or client_addr
    #  of the other record will be populated with the NAT gateway address or VIP address)
    other_side = other_side_record
    if other_side && node && other_side.other_node_id != node.id
      other_side.other_node_id = node.id
      other_side.other_node_address = node.node_address
      other_side.save_with_validation(false)
    end

  end

  def set_other_node_columns(ignore_existing_other_node_id = false)
    if self.other_node_uuid.present?
      node = FaspNode.find_by_node_uuid(self.other_node_uuid)

      if node.present?
        # Other node is managed
        self.other_node_id = node.id
        self.other_node_address = node.node_address
      end
    end

    # if we already have other node info, it must have been populated by the other record, we can skip
    unless other_node_id || ignore_existing_other_node_id
      # if information from other side is available, use it
      other_side = other_side_record
      if other_side
        node = FaspNode.find_by_node_uuid(other_side.node_uuid, :order => 'reported_active_sessions DESC, id')
        if node
          self.other_node_id = node.id
          self.other_node_address = node.node_address
        end
      end

      # if we didn't succeed in getting info from other side, try to look up by address of other side as reported by this node
      unless other_node_id
        self.other_node_address = local? ? server_addr : client_addr
        other_node = FaspNode.find_by_node_address(other_node_address)
        if other_node
          self.other_node_id = other_node.id
        end
      end
    end
  end

  # Assigns a FaspSession to a JobStep by finding an existing JobStep
  # or creating a new one.
  def assign_to_step
    FaspSession.trace("assign_to_step FaspSession.id=#{id}")
    assign_to_step_by_session_id ||
      assign_to_step_by_cookie_uuid ||
      assign_sync_sessions ||
      assign_old_sync_sessions ||
      assign_by_whole_cookie ||
      assign_to_new_step
  end

  # Look for a JobStep by matching <tt>job_steps.uuid</tt> with the uuid found in
  # <tt>fasp_sessions.cookie</tt>.  A match should be found for Jobs started by
  # the console. JobStep.uuid is a random uuid created by Console for each JobStep.
  # We only expect to find it in the cookie if Console initiated the transfer.
  # Returned value matters!
  def assign_to_step_by_cookie_uuid
    FaspSession.trace("assign_to_step_by_cookie_uuid")
    FaspSession.trace("cookie_hash[:uuid]=#{cookie_hash[:uuid]}")
    FaspSession.trace("JobStep.find_by_uuid(cookie_hash[:uuid])=#{JobStep.find_by_uuid(cookie_hash[:uuid]).inspect}")
    if !cookie_hash[:uuid].blank? && (step = JobStep.find_by_uuid(cookie_hash[:uuid]))
      logger.debug("assign_to_step_by_cookie_uuid FaspSession.id=#{id}")
      exchange_values_with(step)
      return true
    else
      return false
    end
  end

  def assign_by_whole_cookie
    # attempt to find job_step for session by matching the whole cookie
    if found_job_step = job_step_for_cookie
      exchange_values_with(found_job_step)
      return true
    else
      return false
    end
  end

  def job_step_for_cookie
    # do this only for sessions where cookie = aspera.console, aspera.faspex, or aspera.scp
    # In these cases, we have a reasonable expectation that if the cookie is
    # the same, then the session is the same. (except in case where as_forwarder duplicates a cookie)

    # fail if no cookie
    return nil if cookie.blank?
    # fail if not a standard Aspera cookie
    if ![Console::CONSOLE_APPLICATION, 'aspera.faspex', 'aspera.scp', 'aspera.console', 'aspera.console (outside)'].include?(cookie_hash[:application_name])
      return nil
    end

    # look at existing JobSteps
    conditions = ["cookie = ? AND job_step_id IS NOT NULL", cookie]
    candidate_steps = FaspSession.find(:all, :conditions => conditions).collect(&:job_step).compact.uniq
    candidate_steps.each do |s|
      unless cookie_inconsistent_with_step?(s)
        return s
      end
    end
    return nil
  end

  # checks whether self.cookie is inconsistent with pre-existing fasp_sessions with same cookie,
  # i.e. see if self cannot be a retry for the pre-existing fasp_sessions with same cookie and job_id already assigned
  def cookie_inconsistent_with_step?(js)
    # if we've gotten to this point, session cannot already have a matching fasp_session reported from the other node (session_id match already has been checked)

    # CASE 1
    # look to see if we have a previous fasp_session record reported by the same node as the one that reported self
    if prev_session = js.sessions.find_by_node_uuid(self.node_uuid)
      # check to see if previous transfer succeeded
      # (can't check for failure of previous session due to possibility of db logger not having correctly marked previous fasp_session as failed)
      if prev_session.status == 'completed'
        return true
      end
      # check if any transfer parameters are different
      if same_reporter_fields_mismatch?(prev_session)
        return true
      end
    end

    sessions_reported_by_other_nodes = js.sessions.find(:all, :conditions => ["node_uuid != ?", self.node_uuid], :order => 'started_at DESC')


    # CASE 3 (in chronological development order) - we need to check this before CASE 2 because failover scenario will be flagged
    # by case #2 as inconsistent
    #
    # check for case were we have two server nodes that are load-balanced, and retry hit a different node than initial attempt
    # managed nodes must be in same failover group
    unless sessions_reported_by_other_nodes.empty?
      # check to see if the other records reported by same side as the current session record (i.e. server vs client)
      # were reported by other nodes in same failover group
      self_node = FaspNode.find_by_node_uuid(self.node_uuid)
      self_failover_group = self_node && self_node.enable_node_group && self_node.node_group_name

      # this managed node is not in a failover group (if other side is in failover group and is the side that changed nodes, then CASE 1 should have caught it)
      if self_failover_group.present?

        sessions_reported_by_other_nodes.each do |s|
          # check to see if previous transfer succeeded (if so, this isn't a retry)
          return true if s.status == 'completed'

          if s.initiator == self.initiator
            other_node = FaspNode.find_by_node_uuid(s.node_uuid)
            other_failover_group = other_node && other_node.enable_node_group && other_node.node_group_name
            # other node that reported in same role (server vs client) must have same failover group name
            return true unless self_failover_group == other_failover_group
            # check fields that should be the same
            return true if same_reporter_fields_mismatch?(s)
          else
            return true if different_reporter_fields_mismatch?(s)
          end
        end
        # failover group is present, and we've checked all the previous sessions reported by other nodes in same role
        # and they all are in same failover group
        return false
      end
    end


    # CASE 2
    # look to see if we have a previous fasp_session record reported by a different node as the one that reported self (other node)
    unless sessions_reported_by_other_nodes.empty?
      # if the job_step has two records reported by two other nodes, then this one cannot possibly be a retry
      if sessions_reported_by_other_nodes.map(&:node_uuid).uniq.size > 1
        return true
      end
      prev_session = sessions_reported_by_other_nodes[0]
      # check to see if previous transfer succeeded
      # (can't check for failure of previous session due to possibility of db logger not having correctly marked previous fasp_session as failed)
      if prev_session.status == 'completed'
        return true
      end
      # check fields that should be the same (or opposite) when the session was reported by the other node's db logger
      if different_reporter_fields_mismatch?(prev_session)
        return true
      end
    end


    # We've done our best to prevent cross-linking of session with previous session that this is not a retry for. Nothing precludes this job step from being linked to self.
    return false
  end



  # checks fields that should be the same for a retry when the two records are reported by the same node (or by two nodes in same failover group)
  def same_reporter_fields_mismatch?(prev_session)
    initiator != prev_session.initiator ||
      operation != prev_session.operation ||
      cipher != prev_session.cipher ||
      #server_addr != prev_session.server_addr ||
      client_addr != prev_session.client_addr ||
      source_paths != prev_session.source_paths ||
      dest_path != prev_session.dest_path ||
      user != prev_session.user ||
      token != prev_session.token
  end

  # checks fields that should be the same for a retry when the two records are reported by the same node
  def different_reporter_fields_mismatch?(prev_session)
    initiator == prev_session.initiator ||  # *** initiator should NOT match, they should be opposite, Remote vs. Local
      operation != prev_session.operation ||
      cipher != prev_session.cipher ||
      #server_addr == prev_session.server_addr || # should NOT match, but limited use to check this since 'initiator' should take care of this
      #client_addr == prev_session.client_addr || # should NOT match, but limited use to check this since 'initiator' should take care of this
      #source_paths #prev_session.source_paths || # fancy logic could try to check this to see if one is a subset of the other, but would be fragile
      #dest_path #prev_session.dest_path ||      # fancy logic could try to check this to see if one is a subset of the other, but would be fragile
      user != prev_session.user ||
      token != prev_session.token
  end

  # Looks for another FaspSession row with the same session_id that is already
  # linked to a JobStep.  This would be the case when a FaspSession row was created
  # by the other FaspNode participating in the transfer (session).  Returned value matters!
  def assign_to_step_by_session_id
    conditions = ["session_id = ? AND job_step_id IS NOT NULL", session_id]
    if session = FaspSession.find(:first, :conditions => conditions)
      exchange_values_with(session.job_step)
      return true
    end
  end

  def assign_old_sync_sessions
    return false unless aspera_old_sync?
    assign_to_new_step
  end

  # new aspera.sync2 FaspSessions are assigned new JobSteps unless both sides are managed nodes and the other side
  # has already reported the session (and already has a JobStep)
  def assign_sync_sessions
    return false unless aspera_sync?
    job_uuid = cookie_hash[:uuid]
    logger.info("new sync session, server:#{self.server_addr}")
    # find the most recent sync Job with this ID (since after 100 sessions we create a new
    # Job, there may be more than one with the same external id)
    hot_job = Job.find_by_external_id(job_uuid, :order => "id DESC")
    if hot_job
      FaspSession.trace("found hot job:#{hot_job.id}")

      # since the job may have been marked as completed when the prior step finished...
      if hot_job.status != DbLogger::STATUS_RUNNING
        hot_job.status = DbLogger::STATUS_RUNNING
        hot_job.save!
        FaspSession.trace("hot job updated")
      end

      unless hot_job.steps.empty?
        logger.debug ">>>"
        logger.debug ">>>>>>>>>>>>>>>>>> prevent email for sync job step: #{hot_job.steps.last.id}"
        hot_job.steps.last.prevent_email(JobEmail::TYPE_SUCCESS)
        hot_job.steps.last.prevent_email(JobEmail::TYPE_ERROR)
      end

    end

    unless hot_job && hot_job.steps.length < 100
      FaspSession.trace("new sync job spawned")
      hot_job = new_job!
    end
    sync_step = new_step!(hot_job)
    FaspSession.trace("adding step:#{sync_step.id} to job:#{hot_job.id}")
    logger.debug ">>>>>>>>>>>>>>>>>> new sync job step: #{sync_step.id}"
    logger.debug ">>>"

    true
  end

  # Links FaspSession to newly created JobStep and Job.
  def assign_to_new_step
    logger.debug("assign_to_new_step FaspSession.id=#{id}")
    new_job = new_job!
    new_step = new_step!(new_job)
    return true
  end

  def generate_contact
    if cookie_hash[:submitted_by].blank?
      "#{user} (ssh)"
    else
      cookie_hash[:submitted_by]
    end
  end

  def generate_job_name
    if aspera_sync? || aspera_old_sync?
      result = "Hot folder: #{cookie_hash[:description]}"
    elsif is_faspex?
      result = cookie_hash[:description] || "Faspex transfer"
    elsif aspera_watchfolder?
      wf_record = AsperaWatchfolder.find_by_uuid(cookie_hash[:uuid])

      result = wf_record.present? ? "Watchfolder: #{wf_record.meta_name}" : "Aspera Watchfolder"
    elsif !cookie_hash[:description].blank?
      result = cookie_hash[:description]
    else
      result = self.default_job_name
    end

    result && result[0,255]
  end

  def address_name(addr)
    node = FaspNode.find_by_node_address(addr)
    return node.node_name unless node.nil?
    node = PartnerNode.find_by_address(addr)
    return node.name unless node.nil?
    addr
  end

  # Creates and returns new Job instance for this FaspSession.
  # Raises error if Job can't be created.
  #---
  # FIXME: May want do change login so job is only saved once (after step creation)
  # so callbacks that set permissions are only run once.
  def new_job!
    new_job = Job.create!({
      :status      => Job::STATUS_RUNNING,
      :ssh_user    => user,
      :contact     => generate_contact,
      :description => generate_job_name,
      :started_via => cookie_hash[:application_name] || UNNAMED_APPLICATION,
      :external_id => cookie_hash[:uuid],
      :force_recalc => true,
    })
    logger.debug("new_job! FaspSession.id=#{id}") { new_job.inspect}
    new_job
  end

  # Creates and returns new JobStep instance for this FaspSession.
  # Raises error if JobStep can't be created.
  def new_step!(job)
    # do NOT set to new session's status (which might be 'completed')
    # to ensure this step is handled in step's background processing.
    new_step = job.job_steps.build :status => JobStep::STATUS_RUNNING, :source_initiated => self.initiated_by_source
    exchange_values_with(new_step)
    new_step.build_computed_notifications
    new_step.save!
    job.save!
    logger.debug("new_step! FaspSession.id=#{id}") { new_step.inspect }
    new_step
  end

  # FIXME: rename method to reflect fact that status is being propagated
  # to job_step and job.
  def exchange_values_with(step)
    # update self (FaspSession) with job and job_step associations
    self.job_step = step
    self.job = step.job
    save(false)

    step.update_attribute(:client_docroot, self.client_docroot) if self.client_docroot.present?
    step.update_attribute(:server_docroot, self.server_docroot) if self.server_docroot.present?

    # Record access keys and current names (if present)
    if self.client_access_key.present?
      step.update_attribute(:client_access_key, self.client_access_key)
      key = Login.access_keys.find_by_login(self.client_access_key, 'ClusterAccessKey')
      step.update_attribute(:client_access_key_name, key.name) if key.present? && key.name.present?
    end

    if self.server_access_key.present?
      step.update_attribute(:server_access_key, self.server_access_key)
      key = Login.access_keys.find_by_login(self.server_access_key, 'ClusterAccessKey')
      step.update_attribute(:server_access_key_name, key.name) if key.present? && key.name.present?
    end

    step.update_attribute(:marked_for_delete, false)
    job.update_attribute(:marked_for_delete, false)

    # see if there is already a previous session
    existing_session = (step.client_session || step.server_session)
    existing_session_uuid = existing_session && existing_session.session_id
    if existing_session_uuid && (existing_session_uuid != self.session_id)
      # replace the previous fasp session
      step.fasp_session_id = self.id
      if local?
        step.client_session = self
        step.server_session = nil
      else
        step.client_session = nil
        step.server_session = self
      end
      # set status, to suppress error notification for prior session
      step.status = self.status_for_job_and_step

      # update node info if needed (this session might be a a retry that hit a different node in a load-balancing cluster)
      update_node_related_fields_in_case_of_load_balancing(step)

      # make sure step gets updated
      step.force_recalc = true

      step.save_with_validation(false)

      # if timeout has expired, we need to reset email notifications for this job step
      # protect against case where stopped_at never got populated by db logger - it should get populated by Console eventually when status gets set to 'inactive'
      # i.e. don't send email out if stopped at is not populated - wait for console to populate it
      if existing_session.stopped_at &&
          (existing_session.status != DbLogger::STATUS_WILLRETRY ||
              (existing_session.stopped_at < existing_session.retry_timeout.to_i.seconds.ago)
          )

        step.create_email_events(true)
      end
    else
      # Trac #23288 - protect against simple/smart transfers that were marked as 'Transfer not started or not logged'
      need_to_create_email_events = (step.status == JobStep::STATUS_ERROR && step.err_desc == JobStep::NEVER_LOGGED_ERR_MSG)

      # update JobStep *_session associations
      step.fasp_session_id = self.id if step.session.nil? || local?
      step.client_session = self if  local?
      step.server_session = self if !local?
      # make sure step gets updated
      step.force_recalc = true
      step.save_with_validation(false)

      update_credentials(step) unless step.console?

      step.save_with_validation(false)

      step.create_email_events(true) if need_to_create_email_events

    end
  end

  # session might be a retry that hit a different node
  def update_node_related_fields_in_case_of_load_balancing(step)
    set_node_columns
    set_other_node_columns(true)
    if outbound?
      step.source_node_id = node_id
      step.source_node_address = node_address
      step.dest_node_id = other_node_id
      step.dest_node_address = other_node_address
    else
      step.dest_node_id = node_id
      step.dest_node_address = node_address
      step.source_node_id = other_node_id #if step.node_id.nil?
      step.source_node_address = other_node_address
    end
  end

  # Set source/dest node_id, node_address, ssh_login.  Not called if job was started
  # by console since we already have values.
  def update_credentials(step)
    if upload?
      step.dest_ssh_login = user
    else
      step.source_ssh_login = user
    end

    if outbound?
      step.source_node_id = node_id
      step.source_node_address = node_address
      step.dest_node_id = other_node_id
      step.dest_node_address = other_node_address
    else
      step.dest_node_id = node_id
      step.dest_node_address = node_address
      step.source_node_id = other_node_id #if step.node_id.nil?
      step.source_node_address = other_node_address
    end
    # make sure step gets updated
    step.force_recalc = true
  end


  def is_faspex?
    cookie_hash[:application_name] == 'aspera.faspex'
  end

end
