class JobStep

  NEVER_LOGGED_ERR_MSG = 'Transfer not started or not logged'

    # If 0, prefer client_session to server_session.  If > 0, use server_session
    # if its #usec column is larger by at least this amount.
    SESSION_USEC_THRESHOLD = 5000000

  # class methods

  class << self

    # amount of time to wait after job step has been submitted to wait for it to start before marking as error
    SUBMITTED_STEP_TIMEOUT_SECS = ConsoleConfig::ConsoleDefault.load_settings.submitted_job_timeout_seconds

    # if transfer is in queued state but initiating node no longer exists, mark as error.
    def fail_orphaned!
      sql = %Q[SELECT js.* FROM job_steps js
LEFT JOIN fasp_nodes n
ON js.submit_node_id = n.id
WHERE status="#{Console::STATUS_QUEUED}"
AND js.submit_node_id IS NOT NULL
AND n.id IS NULL]

      find_by_sql(sql).each do |js|
        begin
          js.status = Console::STATUS_ERROR
          js.started_at = Time.now
          js.ended_at = Time.now
          js.err_desc = "Initiating node deleted from Console"
          js.save(false)
          js.job.force_recalc = true
          js.job.save(false)
        rescue
          logger.error "Exception trying to fail orphaned JobStep id:#{js.id}"
        end
      end
    end


    def mark_invalid_scheduled_job_steps
      scheduled_start_boundary = ConsoleConfig::ConsoleDefault.scheduling_tolerance_minutes.minutes.ago
      conditions = ["status = ? and scheduled_start_at <= ?", STATUS_SCHEDULED, scheduled_start_boundary]
      JobStep.all(:conditions => conditions).each do |job_step|
        job_step.set_status_to_error("Scheduler was down or bad scheduled start time specified")
      end
    end

    def queue_scheduled_jobs
      # using Time.now seems to cause timezone problems, use 0.minutes.ago instead
      JobStep.scheduled.all(:conditions => ["scheduled_start_at <= ?", 0.minutes.ago], :include => :job).each do |step|
        next if step.job.needs_rescheduling?

        step.attributes = {
          :status    => STATUS_QUEUED,
          :queueable => true,
          :q_waiting => true
        }
        step.save(true)
      end
    end  
    
    def update_steps_from_sessions
      job_step_to_process = JobStep.to_recalc_sql
      logger.debug("#{job_step_to_process.present? ? job_step_to_process.size : 0} job step yet to be updated from sessions")
      job_step_to_process.each do |job_step|
        begin
          if job_step.session
            job_step.update_from_session
          end
          job_step.job.force_recalc = true
          job_step.job.save!
        rescue
          logger.error("Exception processing job_step #{job_step.id}: #{$!} (#{$!.class.name})\n#{$!.backtrace.join("\n")}")
        end
      end
    end

    # look for job steps that are stuck in the 'submitted' state
    def update_stalled_job_steps
      # because of the join on Jobs, the stalled_jobs collection yields read-only records
      stalled_job_steps.each do |step|
        # so we have to fetch the ones we want to modify again
        read_write_step = JobStep.find(step.id)
        read_write_step.set_status_to_error(NEVER_LOGGED_ERR_MSG)
      end
    end

    def process_retry_timeouts
      # because of the join on FaspSessions, the collection yields read only records
      retry_timeouts.each do |step|
        # so we have to fetch the ones we want to modify again
        read_write_step = JobStep.find(step.id)
        read_write_step.set_status_to_error(
            "#{RETRY_TIMEOUT_ERR_MSG} (Last error: #{read_write_step.err_desc})".to_s.mb_chars[0..254].limit_bytes(255),
            {:ended_at => step.retry_timeout_ended_at}
        )
      end
    end

    def retry_timeouts
      willretry.find(
        :all,
        :joins => :fasp_session,
        :conditions => ["fasp_sessions.stopped_at < TIMESTAMPADD(SECOND,-(fasp_sessions.retry_timeout+60),UTC_TIMESTAMP())"]
      )
    end

    def stalled_job_steps
      # old condition: cannot test for time since queued_at, things can now sit on the queue for a while before
      # getting submitted
      #%w[
      #      job_steps.started_at IS NULL AND
      #      (job_steps.status=? AND TIMESTAMPDIFF(SECOND,job_steps.submitted_at,UTC_TIMESTAMP()) > ? OR
      #      job_steps.status=? AND TIMESTAMPDIFF(SECOND,job_steps.queued_at,UTC_TIMESTAMP()) > ?)
      #    ].join(" "),
      # STATUS_QUEUED,
      # SUBMITTED_STEP_TIMEOUT_SECS

      find(:all, :joins=>:job,
           :conditions => [
               %w[
            job_steps.started_at IS NULL AND
            (job_steps.status=? AND TIMESTAMPDIFF(SECOND,job_steps.submitted_at,UTC_TIMESTAMP()) > ?)
          ].join(" "),
               STATUS_SUBMITTED,
               SUBMITTED_STEP_TIMEOUT_SECS
           ])
    end

    # Produce proper description on failover - "description (failover n)"
    def failover_description(desc)
      # See if description includes failover text
      failed_over = desc.match(/(\(failover (\d+)\))$/)

      if failed_over
        # Get current failover number
        failover_num = $2.to_i + 1

        desc.gsub(/(\(failover (\d+)\))$/, "(failover #{failover_num})")
      else
        "#{desc} (failover 1)"
      end
    end

    # Check each errored job to see if it can fail over
    def process_failovers
      JobStep.failover_possible_scope.all.each do |js|
        case js.status
          when JobStep::STATUS_ERROR
            if js.failover_possible && js.can_failover?
              # Record that this step has failed over
              js.failover_possible = false
              js.save(false)

              # Create a clone of failed job with alternate endpoints
              job_clone = js.job.request_rerun_clone(js.user)
              job_clone.description = failover_description(js.job.description)
              job_clone.save(false)

              step_clone = js.request_rerun_clone(job_clone)
              step_clone.attributes = {:source_endpoints_tried => js.source_endpoints_tried,
                                       :dest_endpoints_tried => js.dest_endpoints_tried,
                                       :failover_previous_id => js.id}

              step_clone.handle_failover
            end
          when JobStep::STATUS_COMPLETED, JobStep::STATUS_CANCELLED
            js.failover_possible = false
            js.save(false)
        end
      end
    end
  end
  
  # instance methods

    # calculates ended_at to use when updating transfer that did not retry as expected
    def retry_timeout_ended_at
      # protect against null values by using || even though we do not expect any
      (self.ended_at || Time.now) + (self.retry_interval_seconds || 0).seconds + 60.seconds
    end


  def set_status_to_error(error_message, options = {})
    logger.info "job step #{id}, setting status to error; msg: #{error_message}"

    new_values = {
        :status   => STATUS_ERROR,
        :ended_at => Time.now,
        :err_desc   => error_message
    }

    new_values[:started_at] = Time.now unless started_at?

    new_values.merge!(options)

    self.attributes = new_values
    save(false)

    prevent_email(Console::EMAIL_TYPE_START)
    job && job.update_attribute(:force_recalc, true)
  end

  def inspect_session_based_columns #:nodoc:
    self.insp(:status, :started_at, :ended_at, :exit_code, :err_desc, :eta, :progress_pct)
  end

  # Returns +true+ if server_session is newer than client_session by a certain margin.
  def server_session_recent_enough?
    microsecond_difference = server_session.usecs - client_session.usecs
    microsecond_difference > SESSION_USEC_THRESHOLD
  rescue
    false
  end

  # replaces 'paused' with 'queued' if item has been requeued by Console
  def queue_aware_status_from_session
    status_from_session = session.status_for_job_and_step
    # replace 'paused' with 'queued' if JobStep not manually paused and JobStep is waiting in queue
    if status_from_session == STATUS_PAUSED && !self.paused && self.q_waiting && !self.q_front
      STATUS_QUEUED
    else
      status_from_session
    end
  end

  # Compensate for ascp bug where source paths can be reported differently by each side of transfer
  # Take source side if available
  # Prefer console pre-supplied value if present; dblogger truncates + uses tabs making rerun fail
  def source_paths_from_session
    if self.source_paths.nil?
      # nothing set yet, and this was not a Console-initiated transfer
      session.docrooted_source_paths
    else
      # something has been set already. This might be a Console-initiated transfer, or we may have set from session previously
      if self.submitted_at
        # this was Console-initiated. Preserve the current source_paths value, we don't want to overwrite with a truncated value
        # from the node
        self.source_paths
      else
        # source_paths was set from session previously. Check to see if the current value is from dest side and we are source side,
        # in which case we want to overwrite
        if session.outbound? && session.docrooted_source_paths != self.source_paths
          session.docrooted_source_paths
        else
          self.source_paths
        end
      end
    end
  end

  # Compensate for ascp bug where dest path can be reported differently by each side of transfer
  # Take source side if available
  # Prefer console pre-supplied value if present
  def dest_path_from_session
    if self.dest_path.nil?
      # nothing set yet, and this was not a Console-initiated transfer
      session.docrooted_dest_path
    else
      # something has been set already. This might be a Console-initiated transfer, or we may have set from session previously
      if self.submitted_at
        # this was Console-initiated. Preserve the current dest_path value
        self.dest_path
      else
        # dest_path was set from session previously. Check to see if the current value is from dest side and we are source side,
        # in which case we want to overwrite
        if session.outbound? && session.docrooted_dest_path != self.dest_path
          session.docrooted_dest_path
        else
          self.dest_path
        end
      end
    end
  end


  def update_from_session
    #logger.debug("update_from_session JobStep.id=#{self.id} FaspSession.id=#{fasp_session_id}")
    #logger.debug("step before") { inspect_session_based_columns }
    self.status            = queue_aware_status_from_session
    self.marked_for_delete = sessions.all?(&:marked_for_delete)
    self.source_paths = source_paths_from_session
    self.retry_interval_seconds ||= session.retry_timeout
    self.started_at   ||= session.started_at
    self.ended_at     = session.stopped_at
    self.exit_code    = session.err_code_for_job_and_step
    self.err_desc     = session.err_desc
    self.eta          = session.calculated_eta
    self.encryption   = map_cipher_to_encryption
    self.dest_path    = dest_path_from_session

    session_rates_and_policy = session.current_target_rate_and_policy
    self.minimum_rate = session_rates_and_policy[0] if session_rates_and_policy[0]
    self.target_rate = session_rates_and_policy[1] if session_rates_and_policy[1]

    self.progress_pct = session.calculated_pct  # FIXME: set to 100% if 'completed'?
    self.force_recalc = false
    self.paused = session.no_further_activity_expected? ? false : self.paused

    save(false)
    job.force_recalc!
  end

  # Prefer node's encryption setting; node may require/enforce encryption.  But dblogger returns incorrect
  # value when encryption is used so map to correct value.
  def map_cipher_to_encryption
    if session.cipher?
      session.cipher == 'AES128' ? 'AES-128' : session.cipher
    else
      encryption
    end
  end
  
end
