class PartnerNode < ActiveRecord::Base
  include NodeBase

  has_one :unmanaged_node, :dependent => :destroy, :foreign_key => :foreign_key_id

  has_many :endpoints, :class_name => "Login", :as => :owner, :dependent => :destroy
  has_many :source_template_steps, :class_name => "JobTemplateStep", :finder_sql =>
      'select job_template_steps.* ' +
      '  from job_template_steps ' +
      ' join logins on job_template_steps.source_saved_login_id = logins.id ' +
      ' where logins.address = "#{address}" '
  has_many :dest_template_steps, :class_name => "JobTemplateStep", :finder_sql =>
      'select job_template_steps.* ' +
      '  from job_template_steps ' +
      ' join logins on job_template_steps.dest_saved_login_id = logins.id ' +
      ' where logins.address = "#{address}" '
  
  has_one :transfer_node, :foreign_key => :foreign_key_id, :conditions => ["type = 'UnmanagedNode'"]

  validates_presence_of   :address, :ssh_port
  validates_format_of     :address, :with => Console::VALID_IP_OR_DNS_OR_FQDN_REGEXP, :allow_blank => true, :message => 'must be an IP address or domain name'
  validates_uniqueness_of :address
  validates_length_of     :name, :within => 1..100
  validates_uniqueness_of :name
  validates_numericality_of :ssh_port, :allow_nil => true, :only_integer => true, :message => 'must be an integer'
  validates_numericality_of :v3_node_daemon_port, :allow_nil => true, :only_integer => true, :message => 'must be an integer'

  named_scope :placed_on_map, :conditions => ["map_visible = ?", true]
  
  default_value_for :ssh_port, 22
  default_value_for :v3_node_daemon_port, 9092
  
  before_save :update_login_fields

  after_save :mirror_to_unmanaged_node
  after_save :update_star_endpoint_type

  alias_attribute :node_name, :name
  alias_attribute :node_address, :address

  # Returns _name_.  So can reference #node_name() whether we have a FaspNode or a PartnerNode.
  #def node_name
  #  name
  #end
  #
  ## Returns _name_changed?_.  So can reference #node_name_changed?() whether we have a FaspNode or a PartnerNode.
  #def node_name_changed?
  #  name_changed?
  #end
  #
  ## Returns _address_changed?_.  So can reference #node_address_changed?() whether we have a FaspNode or a PartnerNode.
  #def node_address_changed?
  #  address_changed?
  #end
  #
  #def node_address
  #  address
  #end
  #
  #def node_address=(val)
  #  self.address = val
  #end

  def type_string
    'Partner'
  end

  def valid_login_types
    Login::LOGIN_TYPES_UNMANAGED
  end

  def map_node
    {
      :id       => "pn_#{id}", # so doesn't collide with fasp_node map node with same id
      :name     => name,
      :x        => map_x || MapUtility::DEFAULT_MAP_X,
      :y        => map_y || MapUtility::DEFAULT_MAP_Y,
      :type     => "unmanaged",
      :locked   => true,
      :selected => false,
    }
  end

  def chart_node_out(user, sessions = active_sessions_for_user(user)) #=sessions.active )
    _total_rate = rate_totals(sessions)
    _timestamp = AsperaDB.current_timestamp(self.class)
    {
        :chart_index => 1,
        :controllable => false,
        :k => _total_rate[:out],
        :plot_type => "line",
        :s => _timestamp,
        :rails_id => id,
        :rails_model => "PartnerNode",
        :series_name => "PartnerNode" + id.to_s + "TotalOut",
        :tip_prefix => "<b>" + javascript_safe_node_name + " outbound total:</b> "
    }
  end

  def chart_node_out_area(user, sessions = active_sessions_for_user(user))
    chart_node = chart_node_out(user, sessions)
    chart_node[:plot_type] = "area"
    chart_node[:series_name] = chart_node[:series_name]+"Area"
    chart_node
  end

  def chart_node_out_area_to_node(user, node)
    chart_node = chart_node_out_area(user)
    total_rate = FaspSession.rate_totals(active_sessions_to_node(user, node))
    chart_node[:k] = total_rate[:in] # FaspSession.rate_totals meaning is reversed for partner nodes
    # override caption
    chart_node[:tip_prefix] = "<b>#{javascript_safe_node_name} -&gt; #{node.javascript_safe_node_name} total:</b> "
    chart_node
  end

  def active_sessions_to_node(user, node)
    active_sessions_for_user(user).map { |s|
      (s.job_step.dest_fasp_or_partner_node && s.job_step.dest_fasp_or_partner_node == node) ? s : nil
    }.compact
  end

  def chart_node_in(user, sessions = active_sessions_for_user(user))
    _total_rate = rate_totals(sessions)
    _timestamp = AsperaDB.current_timestamp(self.class)
    {
        :chart_index => 0,
        :controllable => false,
        :k => _total_rate[:in],
        :plot_type => "line",
        :s => _timestamp,
        :rails_id => id,
        :rails_model => "PartnerNode",
        :series_name => "PartnerNode" + id.to_s + "TotalIn",
        :tip_prefix => "<b>" + javascript_safe_node_name + " inbound total:</b> "
    }
  end

  def chart_node_in_area(user, sessions = active_sessions_for_user(user))
    #raise 'foo'
    chart_node = chart_node_in(user, sessions)
    chart_node[:plot_type] = "area"
    chart_node[:series_name] = chart_node[:series_name]+"Area"
    chart_node
  end

  def active_sessions_for_user(user)
    @active_sessions_for_user ||= begin
      steps = user.job_steps.possibly_active.all(:conditions => ["source_node_address = ? OR dest_node_address = ?", address, address])
      steps.map {|step|
        (step.q_waiting? && !step.q_front?) ? nil : step.sessions.all(:conditions => {:status => FaspSession::STATUSES_POSSIBLY_ACTIVE})
        #step.sessions.all(:conditions => {:status => FaspSession::STATUSES_POSSIBLY_ACTIVE})
      }.flatten.compact
    end
  end

  def javascript_safe_node_name
    node_name.to_s.escape_javascript
  end

  def rate_totals(sessions) # = sessions.active)
    totals = FaspSession.rate_totals(sessions)
    # Need to reverse directions because they are computed relative to managed nodes
    {:in => totals[:out], :out => totals[:in]}
  end

  def mirror_to_unmanaged_node
    self.build_unmanaged_node unless unmanaged_node

    unmanaged_node.primary_address = self.address
    unmanaged_node.name = self.name
    unmanaged_node.show_on_map = self.map_visible
    unmanaged_node.map_x = self.map_x || MapUtility::DEFAULT_MAP_X
    unmanaged_node.map_y = self.map_y || MapUtility::DEFAULT_MAP_Y
    unmanaged_node.ssh_port = self.ssh_port

    unmanaged_node.save!
  end


  class << self
    def map_nodes(node_to_edit = nil)
      nodes = placed_on_map
      nodes << node_to_edit unless node_to_edit.blank? || nodes.include?(node_to_edit)
      nodes.map do |node|
        map_node = node.map_node
        if node == node_to_edit
          map_node[:locked]   = false
          map_node[:selected] = true
        end
        map_node
      end
    end
  end
  
  # Creates default end point.
  def after_create
    endpoints.find_or_create_by_login_and_address("*", address)
  end

  # Update endpoint type and port of * endpoint
  def update_star_endpoint_type
    e = endpoints.find_by_login('*')
    if e.present? && e.endpoint_type != default_endpoint_type
      e.update_attributes(
        :endpoint_type => default_endpoint_type,
        :port => default_endpoint_type == Login::LOGIN_TYPE_SSH ? ssh_port : v3_node_daemon_port
      )
    end
  end

  def after_destroy
    destroy_templates_and_steps
  end
  
  def update_login_fields
    update_logins(address_was, address, ssh_port)
  end
  
  private
  
  def validate
    validate_address_not_managed_node
    validate_name_not_managed_node
  end
  
  def validate_address_not_managed_node
    if FaspNode.find_by_node_address(address)
      errors.add(:address, "is in use by a managed node")
    end
  end

  def validate_name_not_managed_node
    if FaspNode.find_by_node_name(name)
      errors.add(:name, "is in use by a managed node")
    end
  end
  
end
