class TransferJob < ActiveRecord::BaseWithoutTable
  
  column :comment, :string
  column :description, :string
  column :console_tags, :string

  column :source_end_point_id, :integer
  column :dest_end_point_id, :integer
  
  column :source_address, :string
  column :source_initiated, :boolean
  column :source_login, :string
  column :source_password, :string
  column :source_paths, :string
  column :source_port, :string
  column :source_use_ssh_key, :boolean
  column :source_ssh_key_id, :integer

  column :source_endpoint_type
  column :source_use_ssl
  column :source_validate_ssl_cert

  column :source_base_enable, :boolean
  column :source_base, :string
  
  column :dest_address, :string
  column :dest_login, :string
  column :dest_password, :string
  column :dest_path, :string
  column :dest_port, :string
  column :dest_use_ssh_key, :boolean
  column :dest_ssh_key_id, :integer

  column :dest_endpoint_type
  column :dest_use_ssl
  column :dest_validate_ssl_cert

  column :use_ascp4, :boolean

  column :protocol, :string

  column :udp_port, :integer
  column :use_fasp_proxy, :boolean
  column :proxy_host, :string
  column :proxy_port, :integer
  column :proxy_secure, :boolean
  column :proxy_user, :string
  column :proxy_password, :string
  column :prefer_dest_initiated, :boolean

  column :target_rate, :integer
  column :minimum_rate, :integer
  column :priority_and_policy, :string
  column :enable_retry, :boolean # form cbx
  column :max_retry_attempts, :integer
  column :retry_interval_seconds, :integer

  column :file_exclude_by_timestamp, :boolean
  column :file_exclude_newer_than, :integer

  column :encrypt_at_rest_enable, :boolean # form cbx
  column :encryption_password, :string
  column :encryption_password_confirm, :string  
  column :show_plain_text # form cbx
  column :encryption, :string # transport encryption

#  column :file_resume_policy, :string # no longer used?
  column :resume_incomplete, :boolean # form cbx
  column :file_overwrite_policy, :string
  column :file_compare_type, :string

  column :preserve_file_timestamps, :boolean
  column :preserve_file_uids, :boolean
  column :preserve_file_gids, :boolean
  column :preserve_file_create_times, :boolean
  column :preserve_file_modify_times, :boolean
  column :preserve_file_access_times, :boolean
  column :preserve_file_source_access_times, :boolean

  column :symbolic_link_handling, :string
  column :delete_before_transfer, :boolean
  column :archive_after_transfer, :boolean
  column :archive_directory, :string
  column :delete_source_files, :boolean
  column :delete_source_dirs_after_transfer, :boolean
  column :delete_empty_directories, :boolean
  column :exclude_file_filter, :string
  column :include_file_filter, :string
  
  column :datagram_size_cbx, :boolean # form cbx
  column :datagram_size, :integer
  column :read_block_size_cbx, :boolean # form cbx
  column :read_block_size, :integer
  column :write_block_size_cbx, :boolean # form cbx
  column :write_block_size, :integer

  column :run_when
  column :scheduled_start_at, :string

  column :test_mode_skip_soap, :boolean
  column :test_mode_skip_files, :boolean # passed to remote file browser for testing

  validates_presence_of :description, :message => "^Transfer name cannot be blank"
  validates_presence_of :source_paths, :message => "^Please select one or more items to transfer"
  validates_presence_of :dest_path, :message => "^Transfer destination cannot be blank"
  validates_length_of :description, :maximum => 255, :message => "^Transfer name is too long (maximum is 255 characters)"
  validates_length_of :include_file_filter, :maximum => 512
  validates_length_of :exclude_file_filter, :maximum => 512
  validates_length_of :comment, :maximum => 1000, :message => "^Comment is too long (maximum is 1000 characters)"
  validates_numericality_of :max_retry_attempts, :only_integer => true, :greater_than => 0, :if => :retry_enabled?, :message => "^Number of retries is not a number"
  validates_numericality_of :retry_interval_seconds, :only_integer => true, :greater_than => 0, :if => :retry_enabled?, :message => "^Seconds between retries is not a number"
  validates_numericality_of :target_rate, :greater_than => 0, :message => "^Target rate must be a positive number"
  validates_numericality_of :minimum_rate, :greater_than_or_equal_to => 0, :allow_nil => true, :message => "^Minimum rate is not a number"
  
  validates_numericality_of :udp_port, :allow_blank => true, :only_integer => true, :greater_than => 0, :less_than => 65536, :message => "^fasp UDP Port must be a valid port number"

  validates_presence_of :proxy_host, :if => :use_fasp_proxy?, :message => "^fasp proxy host cannot be blank"
  validates_numericality_of :proxy_port, :allow_blank => false, :only_integer => true, :greater_than => 0, :less_than => 65536, :if => :use_fasp_proxy?, :message => "^fasp proxy port must be a valid port number"
  #validates_presence_of :proxy_user, :if => :use_fasp_proxy?, :message => "^fasp proxy username cannot be blank"
  validates_presence_of :proxy_password, :if => :proxy_password_required?, :message => "^fasp proxy password cannot be blank if username is specified"

  validates_presence_of     :file_exclude_newer_than, :if => :file_exclude_by_timestamp
  validates_numericality_of :file_exclude_newer_than, :if => :file_exclude_by_timestamp,
                            :allow_blank => true,
                            :only_integer => false,
                            :greater_than_or_equal_to => 0,
                            :less_than_or_equal_to => 1000000000
  validates_numericality_of :file_exclude_newer_than, :if => :file_exclude_by_timestamp,
                            :only_integer => true,
                            :message => 'must be an integer value'

  validates_numericality_of :datagram_size,
    :allow_blank => true, 
    :greater_than_or_equal_to => 296, 
    :less_than_or_equal_to => 10_000,
    :message => "^Datagram size must be an integer between 296-10000"
  validates_numericality_of :read_block_size, :allow_blank => true, :greater_than => 0, :message => "^Read block size must be a positive number"
  validates_numericality_of :write_block_size, :allow_blank => true, :greater_than => 0, :message => "^Write block size must be a positive number"
  validates_presence_of     :archive_directory, :if => :archive_after_transfer
  validate                  :valid_tags?
  validate                  :unique_source_paths?

  validate                  :initiator_acl_valid?, :if => :both_end_points_present?

  # If one or both endpoints are not specified, other validations will fail
  # A non-specified endpoint will be present, but calling .node will fail
  def both_end_points_present?
    begin
      return source_end_point.valid? && dest_end_point.valid?
    rescue => ex
      return false
    end
  end

  def initiator_acl_valid?
    return unless source_endpoint_type == Login::LOGIN_TYPE_SSH || dest_endpoint_type == Login::LOGIN_TYPE_SSH

    # check to make sure that initiator is not a non-legacy node that lacks 'impersonation' ACL
    initiator = initiating_node
    initiator_acl_problem = initiator && initiator.acl_problem?

    if initiator_acl_problem
      other_node = non_initiating_node
      other_node_acl_problem = other_node && other_node.acl_problem?

      other_node_can_initiate = other_node && !other_node_acl_problem

      if !other_node_can_initiate
        err_msg_1 = "Console cannot use node '#{initiator.node_address}' as a transfer initiator because its Node API login lacks impersonation capability."
        err_msg_2 = if other_node
          "The other node (#{other_node.node_address}) cannot initiate the transfer either for the same reason."
        else
          "The other node (#{non_initiating_address}) cannot initiate the transfer because it is not a managed node."
        end
        errors.add_to_base(err_msg_1)
        errors.add_to_base(err_msg_2)
      end

    end
  end

  def both_sides_managed
    (self.source_end_point.node && self.dest_end_point.node).present?
  end

  def non_initiating_node
    src = self.source_end_point.node
    dest = self.dest_end_point.node

    if initiating_node == src
      dest
    else
      src
    end
  end

  def non_initiating_address
    src = self.source_end_point.node
    dest = self.dest_end_point.node

    if initiating_node == src
      self.dest_end_point.address
    else
      self.source_end_point.address
    end
  end

  def valid_tags?
    return true if console_tags.nil?

    names = {}

    # Check for blank names and values, and ensure uniqueness of tag names
    JSON.parse(console_tags).each do |k, v|
      errors.add_to_base("All tags must have a name") if v['name'].blank?
      errors.add_to_base("All tags must have a value") if v['value'].blank?

      errors.add_to_base("Tag names must be unique (Duplicate tag name: #{v['name']})") if names.has_key?(v['name'])

      names[v['name']] = v['value']
    end
  end

  def unique_source_paths?
    # Split and strip
    paths = source_paths.split("\n").collect{|p| p.strip.gsub(/\/+$/, "")}

    errors.add(:source_paths, " must be unique") if paths.detect{ |e| paths.count(e) > 1 }.present?
  end

  attr_accessor :user, :job, :step, :source_end_point, :dest_end_point, :tag_hash
  attr_reader :emails
 
  toggle_state_field_with_validation :show_connection, [:udp_port, :use_fasp_proxy, :proxy_host, :proxy_port, :proxy_secure, :proxy_user, :proxy_password]
  toggle_state_field_with_validation :show_transfer, [:target_rate, :minimum_rate, :max_retry_attempts, :retry_interval_seconds]
  toggle_state_field_with_validation :show_security, [:encrypt_at_rest_enable, :encryption_password]
  toggle_state_field_with_validation :show_advanced, [:datagram_size, :read_block_size, :write_block_size]
  toggle_state_field_with_validation :show_transfer_time, [:scheduled_start_at]
  
  toggle_state_field :show_file_handling, :show_notifications
  
  def initialize(user, attributes = {}, emails = nil)
    raise(ArgumentError, "Expected to be passed a User, got a #{user.class}: #{user.inspect}") unless user.instance_of?(User)
    defaults = ConsoleConfig::ConsoleDefault.transfer_defaults.merge(simple_transfer_defaults)
    super(defaults.merge!(attributes))
    self.user = user
    self.source_end_point = TransferJobEndPoint.new(
      source_endpoint_type,
      source_end_point_id,
      source_address,
      source_port,
      source_login,
      source_password,
      source_ssh_key_id,
      source_use_ssl,
      source_validate_ssl_cert
    )
    self.dest_end_point = TransferJobEndPoint.new(
      dest_endpoint_type,
      dest_end_point_id,
      dest_address,
      dest_port,
      dest_login,
      dest_password,
      dest_ssh_key_id,
      dest_use_ssl,
      dest_validate_ssl_cert
    )
    self.emails = emails
  end

  def proxy_password_required?
    use_fasp_proxy? && proxy_user.present?
  end

  # Returns has of defaults for simple transfer to be layered on to general 
  # transfer defaults.
  def simple_transfer_defaults
    {
      :run_when => 'now',
      :proxy_secure => true,
      :proxy_port => 9092,
      :use_fasp_proxy => false,
      :prefer_dest_initiated => false,
      :protocol => 'FASP',
      :preserve_file_gids => false,
      :preserve_file_uids => false,
      :preserve_file_create_times => false,
      :preserve_file_modify_times => false,
      :preserve_file_access_times => false,
      :preserve_file_source_access_times => false,
      :file_exclude_by_timestamp => false,
    }
  end
  
  def emails=(emails)
    @emails = if emails.nil?
      [JobStepEmail.new_from_user(user)].compact
    else
      JobStepEmail.new_emails(emails.values)
    end
  end
  
  # saves personal logins, if applicable, for transfer job and user
  def save_logins!
    if valid?
      @source_login_id = save_one_login!(source_endpoint_type, source_end_point, source_address, source_login, source_password, source_port, source_ssh_key_id, source_use_ssl, source_validate_ssl_cert)
      @dest_login_id   = save_one_login!(dest_endpoint_type, dest_end_point, dest_address, dest_login, dest_password, dest_port, dest_ssh_key_id, dest_use_ssl, dest_validate_ssl_cert)
    end
  end
  
  def save_one_login!(endpoint_type, endpoint, address, login, password, port, ssh_key_id, use_ssl, validate_ssl_cert)
    if endpoint.end_point && endpoint.end_point.wildcards?
      eplogin = endpoint.end_point
      address = eplogin.has_address? ? eplogin.address : address
      user    = eplogin.has_login?   ? eplogin.login : login
      return save_personal_login_data(endpoint_type, address, user, password, port, ssh_key_id, use_ssl, validate_ssl_cert).id
    else
      unless address.blank? || login.blank? || password.blank? || port.blank?
        return save_personal_login_data(endpoint_type, address, login, password, port, ssh_key_id, use_ssl, validate_ssl_cert).id
      end

      endpoint.end_point.endpoint_type = endpoint_type
      endpoint.end_point.use_ssl = use_ssl
      endpoint.end_point.validate_ssl_cert = validate_ssl_cert

      # If we entered in a new password or ssh key for existing endpoint, it must be saved
      if password.present? && endpoint.end_point.unencrypted_password != password
        endpoint.end_point.password = password
        endpoint.end_point.save
      elsif ssh_key_id.present? && endpoint.end_point.ssh_key_id != ssh_key_id
        endpoint.end_point.ssh_key_id = ssh_key_id
        endpoint.end_point.save
      end

      return endpoint.end_point.id
    end
  end

  def set_source_initiated
    self.source_initiated = Login.choose_source_initiated(source_end_point, dest_end_point, prefer_dest_initiated)
  end
  
  def save
    if super
      set_source_initiated
      begin
        save_logins!
        create_job!
        create_step!
      rescue # various types of exceptions here including ActiveRecord::RecordNotSaved,
        logger.error { "Exception while starting simple transfer: #{$!}\n#{$@.join("\n")}" }
        errors.add_to_base("Exception while starting simple transfer") if errors.empty?
        job && job.destroy
      end
    end
    errors.empty?
  end

  def create_job!
    self.job = Job.create!({
      :description        => description,
      :comment            => comment,
      :user               => user,
      :ssh_user           => source_initiated ? source_login : dest_login,
      :contact            => contact,
      :started_via        => Console::CONSOLE_APPLICATION,
      :status             => status_at_job_and_step_save,
      :scheduled_start_at => scheduled_start_at_at_job_step_and_save,
    })
  end
  
  def create_step!

    if emails
      emails.reject! {|x| x.no_notifications_chosen?}
    end

    # populate time zone for emails if needed
    emails.each do |one_email|
      one_email.time_zone = JobStepEmail.time_zone_for_email(one_email.email) unless one_email.time_zone
    end

    self.step = job.job_steps.build({
      :queueable             => queueable_at_step_save,
      :status                => status_at_job_and_step_save,
      :dest_path             => dest_path,
      :source_initiated      => source_initiated,
      :source_paths          => source_paths,
      :source_base           => (source_base_enable? && source_base.present? ? source_base : nil),

      :scheduled_start_at => scheduled_start_at_at_job_step_and_save,

      :dest_saved_login_id   => @dest_login_id,
      :dest_node_id          => dest_end_point.node_id,
      :dest_node_address     => actual_dest_address,
      :dest_ssh_login        => actual_dest_login,
      :dest_ssh_port         => dest_end_point.port,

      :source_saved_login_id => @source_login_id,
      :source_node_id        => source_end_point.node_id,
      :source_node_address   => actual_source_address,
      :source_ssh_login      => actual_source_login,
      :source_ssh_port       => source_end_point.port,

      :use_ascp4             => use_ascp4 || false,

      :protocol              => (Console.alternate_protocols_enabled ? protocol : 'FASP') || 'FASP',

      :udp_port              => udp_port,
      :use_fasp_proxy        => use_fasp_proxy,
      :proxy_host            => proxy_host,
      :proxy_port            => proxy_port,
      :proxy_secure          => proxy_secure,
      :proxy_user            => proxy_user,
      :proxy_unencrypted_password => proxy_password,
      :prefer_dest_initiated => prefer_dest_initiated,
      
      :target_rate           => target_rate,
      :minimum_rate          => minimum_rate,
      :priority_and_policy   => priority_and_policy,
      :max_retry_attempts    => (enable_retry ? max_retry_attempts : 0),
      :retry_interval_seconds=> (enable_retry ? retry_interval_seconds : 0),

      :file_exclude_by_timestamp => file_exclude_by_timestamp,
      :file_exclude_newer_than => file_exclude_newer_than,

      :ear_unencrypted_password => encryption_password,
      :encryption            => encryption,

      :console_tags          => console_tags,

      :resume_partially_transferred_files => resume_incomplete,
      :file_overwrite_policy => file_overwrite_policy,
      :file_compare_type     => file_compare_type,

      :preserve_file_timestamps => preserve_file_timestamps,
      :preserve_file_uids => preserve_file_uids,
      :preserve_file_gids => preserve_file_gids,

      # Should only be used if preserve_file_timestamps is active
      :preserve_file_create_times => (preserve_file_timestamps ? preserve_file_create_times : false),
      :preserve_file_modify_times => (preserve_file_timestamps ? preserve_file_modify_times : false),
      :preserve_file_access_times => (preserve_file_timestamps ? preserve_file_access_times : false),
      :preserve_file_source_access_times => (preserve_file_timestamps ? preserve_file_source_access_times : false),

      :symbolic_link_handling => symbolic_link_handling,
      :delete_before_transfer => delete_before_transfer || false,
      :archive_after_transfer => archive_after_transfer || false,
      :archive_directory => archive_directory,
      :delete_source_after_transfer => delete_source_files,
      :delete_source_dirs_after_transfer => delete_source_dirs_after_transfer || false,
      :delete_empty_directories => delete_empty_directories || false,
      :exclude_file_filter   => exclude_file_filter,
      :include_file_filter   => include_file_filter,
      
      :datagram_size         => (datagram_size_cbx ? datagram_size : nil),
      :read_block_size       => (read_block_size_cbx ? read_block_size : nil),
      :write_block_size      => (write_block_size_cbx ? write_block_size : nil),
      
      :emails             => emails,
   })
    self.step.build_computed_notifications
    self.step.save!
  end
  
  def status_at_job_and_step_save
    run_when == 'later' ? Console::STATUS_SCHEDULED : Console::STATUS_QUEUED
  end

  # transfers initiated by Console are Console-queueable, but if this is scheduled to start later
  # then don't mark as queueable yet, wait until it is due
  def queueable_at_step_save
    run_when == 'later' ? false : true
  end

  def scheduled_start_at_at_job_step_and_save
    (run_when == 'later' ? Time.zone.parse(scheduled_start_at) : nil)
  end
  
  def contact
    "#{user.login} (console)"
  end
  
  def actual_source_address
    source_end_point.has_address? ? source_end_point.address : source_address
  end
  
  def actual_source_login
    source_end_point.has_login? ? source_end_point.login : source_login
  end
  
  def actual_dest_address
    dest_end_point.has_address? ? dest_end_point.address : dest_address
  end
  
  def actual_dest_login
    dest_end_point.has_login? ? dest_end_point.login : dest_login
  end
  
  def initiating_node
    source_initiated ? source_end_point.node : dest_end_point.node
  end

  def server_end_point
    if @server_full_end_point.nil?
      if source_initiated
        @server_full_end_point = TransferJobEndPoint.new(dest_end_point.endpoint_type, nil, actual_dest_address, dest_end_point.port, actual_dest_login, dest_end_point.password, dest_end_point.ssh_key_id, dest_end_point.use_ssl, dest_end_point.validate_ssl_cert)
      else
        @server_full_end_point = TransferJobEndPoint.new(source_end_point.endpoint_type, nil, actual_source_address, source_end_point.port, actual_source_login, source_end_point.password, source_end_point.ssh_key_id, source_end_point.use_ssl, source_end_point.validate_ssl_cert)
      end
    end
    @server_full_end_point
  end

  def client_end_point
    if @client_full_end_point.nil?
      if source_initiated
        @client_full_end_point = TransferJobEndPoint.new(source_end_point.endpoint_type, nil, actual_source_address, source_end_point.port, actual_source_login, source_end_point.password, source_end_point.ssh_key_id, source_end_point.use_ssl, source_end_point.validate_ssl_cert)
      end
    else
        @client_full_end_point = TransferJobEndPoint.new(dest_end_point.endpoint_type, nil, actual_dest_address, dest_end_point.port, actual_dest_login, dest_end_point.password, dest_end_point.ssh_key_id, dest_end_point.use_ssl, dest_end_point.validate_ssl_cert)
    end
    @client_full_end_point
  end
    
  def cookie
    CGI.escapeHTML("#{Console::CONSOLE_APPLICATION}:#{step.uuid}:#{job.description}")
  end
  
  def validate
    set_source_initiated

    validate_file_filters
    validate_metadata
    validate_end_points
    validate_nodes
    validate_scheduled_start_at
  end
  
  def validate_metadata
    if target_rate && minimum_rate && minimum_rate > target_rate
      errors.add(:minimum_rate, "^Minimum rate cannot exceed target rate")
    end
  end

  def validate_file_filters
    if exclude_file_filter && exclude_file_filter.to_s.split(",").length > 16
      errors.add(:exclude_file_filter, "^Exclude filter cannot contain more than 16 items")
    end

    if include_file_filter && include_file_filter.to_s.split(",").length > 16
      errors.add(:include_file_filter, "^Include filter cannot contain more than 16 items")
    end
  end
  
  def validate_end_points
    ep = source_end_point
    ep.valid?
    ep.errors.each { |e| errors.add(:transfer_source_end_point_id, "^Source end point #{e}") }
    if encrypt_at_rest_enable
      if ep.node.nil?
        errors.add(:encrypt_at_rest_enable, "File protection only supported if the transfer source is a managed node")
      elsif ! ep.node.supports_encryption_at_rest?
        errors.add(:encrypt_at_rest_enable, 
          "Content protection requires node to be running 2.2 or higher ")
      end
      if encryption_password != encryption_password_confirm
        errors.add(:encryption_password, "File protection password does not match password confirmation")
      elsif encryption_password.blank?
        errors.add(:encryption_password, "File protection password cannot be blank") 
      end
    end
    ep = dest_end_point
    ep.valid?
    ep.errors.each { |e| errors.add(:transfer_dest_end_point_id, "^Destination end point #{e}") }

    unless initiating_node.is_a?(FaspNode)
      # Initiating node is not managed

      initiating_login_type = source_initiated ? source_endpoint_type : dest_endpoint_type
      non_initiating_login_type = source_initiated? ? dest_endpoint_type : source_endpoint_type

      if initiating_login_type == Login::LOGIN_TYPE_NODE_API
        errors.add(:base, "Only managed nodes may use Node API credentials for transfers")
      elsif non_initiating_login_type == Login::LOGIN_TYPE_ACCESS_KEY
        errors.add(:base, "SSH node must be managed for SSH + Access Key transfers")
      end
    end

    if source_end_point.present? && dest_end_point.present?
      if source_endpoint_type == Login::LOGIN_TYPE_SHARES && dest_endpoint_type == Login::LOGIN_TYPE_SSH ||
          dest_endpoint_type == Login::LOGIN_TYPE_SHARES && source_endpoint_type == Login::LOGIN_TYPE_SSH
        errors.add(:base, 'Shares endpoints may not be used with SSH endpoints')
      end
    end
  end
  
  def validate_nodes
    return unless errors.blank?
    validate_has_two_nodes

    validate_at_least_one_managed_node

    #can't continue if missing an endpoint
    return unless both_end_points_present?

    # ascp now allows transfers from node to self
    #validate_nodes_not_equal
    validate_transfer_path
  end
  
  def validate_has_two_nodes
    if source_end_point.address.blank? && dest_end_point.address.blank?
      errors.add_to_base "Select a source and a destination"
      false
    else
      true
    end
  end

  def validate_at_least_one_managed_node
    if no_managed_nodes?
      errors.add_to_base "Neither source nor destination node is a managed node with SSH credentials"
    end
  end

  def validate_nodes_not_equal
    if actual_source_address == actual_dest_address && ! actual_dest_address.blank?
      errors.add_to_base "Source and destination addresses must be different"
    end
  end
  
  def validate_transfer_path
    return if user.admin?
    msg = "You aren't allowed to transfer between your chosen source and destination"
    errors.add_to_base(msg) unless no_managed_nodes? ||
      user.transfer_paths_for_start_simple.find {|path| path.include? transfer_path }
  end
  
  def validate_scheduled_start_at
    if run_when == 'later'
      if scheduled_start_at.blank?
        errors.add(:scheduled_start_at, "^Later transfer time cannot be blank")
      else
        begin
          dt = Time.zone.parse(scheduled_start_at)
          if dt < DateTime.now
            errors.add(:scheduled_start_at, "^Later transfer time must be in the future")
          end
        rescue
          errors.add(:scheduled_start_at, "^Later transfer time is not a valid date/time.")
        end
      end
    end
  end
  
  def retry_enabled?
    self.enable_retry
  end

  def no_managed_nodes?
    (user.nodes.managed & [source_end_point.node, dest_end_point.node]).empty?
  end
  
  def transfer_path
    begin
      @transfer_path ||= TransferPath.new(source_end_point.transfer_path_end_point, dest_end_point.transfer_path_end_point)
    rescue
      nil
    end
  end
  
  def save_personal_login_data(endpoint_type, address, login, password, port, ssh_key_id, use_ssl, validate_ssl_cert)
    login = user.logins.find_or_create_by_address_and_login(address, login)
    login.endpoint_type = endpoint_type
    login.password = password
    login.port = port
    login.ssh_key_id = ssh_key_id
    login.use_ssl = use_ssl
    login.validate_ssl_cert = validate_ssl_cert
    login.save!
    login
  end
  
  class << self
  end
end
