class ActiveRecord::Base
#  include Common::ConstantsFor

  DEFAULT_PAGE_SIZE = 10
  CONDITIONAL_RETRY_ERRORS = {
      :deadlock =>  /Deadlock found when trying to get lock/,
      :lock_wait_timeout => /Lock wait timeout exceeded; try restarting transaction/
  }

  before_save :sanitize_values

  SKIP_SANITIZATION = false

  def sanitize_values
    return if self.class::SKIP_SANITIZATION
    self.attributes.each do |key, value|
      next if key.nil? || value.nil? || !(value.is_a?(String)) || value.to_s.strip.empty?
    
      self.write_attribute(key, self.class.public_sanitize_sql(value))
    end
  end

  def self.with_conditional_retry(options, &block)
    errors_to_retry = options[:errors]
    max_retries = options[:max_retries] || 3
    try_again = true
    while try_again && max_retries > 0
      begin
        yield
        try_again = false
      rescue
        if errors_to_retry.detect {|e| $!.to_s =~ CONDITIONAL_RETRY_ERRORS[e]}
          max_retries -= 1
          sleep 1
          logger.info("Exception: #{$!} #{$!.backtrace.join("\n")}") if logger          
        else
          raise
          try_again = false
        end
      end
    end
  end
  
  def sql_comment(text, options = {})
    border_char = options[:major] ? "=" : "-"
    comment_width = [40, text.split("\n").map(&:length).max + 2].max
    [
      '',
      sql_comment_border(comment_width, border_char),
      "# " + text.split("\n").join("\n# "),
      sql_comment_border(comment_width, border_char),
      ''
    ].join("\n")
  end

  def sql_comment_border(length = 40, border_char = '-')
    "#" + (border_char * length)
  end

  def indent_sql(sql_string)
    line_patterns_not_to_indent = /^  (;|#.*|CREATE|INSERT|SELECT|UPDATE|DELETE|SET|FROM|WHERE|GROUP BY|ORDER BY)(\s|\b|$)(.*)$/i
    line_patterns_to_double_indent = /^  (ON|AND|OR|USING)(\s|\b|$)(.*)$/i
    sql_string.split("\n").collect {|line|
      line.strip.insert(0, '  ').gsub(line_patterns_not_to_indent){|match| "#{$1}#{$2}#{$3}"}.gsub(line_patterns_to_double_indent){|match| "    #{$1}#{$2}#{$3}"}
    }.join("\n").gsub("  \n  \n", "  \n").gsub("  \n  \n", "  \n")
  end

  def db_time_now
    self.class.db_time_now
  end

  # Like <tt>inspect()</tt> with the following differences:
  #
  # * returned string will only include _method_names_ and <tt>:id</tt>.
  # * _method_names_ can be column names or any method the instance repsonds to
  # * if called with no _method_names_ it behaves like <tt>inspect</tt>
  # * angle brackets (<>) replaced with square brackes ([]) to be HTML-friendly
  #
  # Example:
  #
  #   user = User.new(:email => "a@b.com", :name => "Scott Brown")
  #
  #   user.insp(:email, :wants_email?)  => #[User id: nil, email: "a@b.com", wants_email: true]
  def insp(*method_names)
    return inspect.sub(/^#<(.*)>$/, '#[\1]') if method_names.empty?
    names = method_names.unshift(:id)
    attrs_as_nice_string = names.map do |name|
      "#{name}: #{method_for_inspect(name)}"
    end.join(", ")
    "#[#{self.class} #{attrs_as_nice_string}]"
  end

  # Formats _method_name_ nicely for insp.
  def method_for_inspect(method_name)
    value = send(method_name)

    if value.is_a?(String) && value.length > 50
      "#{value[0..50]}...".inspect
    elsif value.is_a?(Date) || value.is_a?(Time)
      %("#{value.to_s(:db)}")
    else
      value.inspect
    end
  end

  def sql_in_clause(array, place_holder = '--place-holder--')
    self.class.sql_in_clause(array, place_holder)
  end

  class << self

    def without_root_in_json(&block)
       saved = ActiveRecord::Base.include_root_in_json
       ActiveRecord::Base.include_root_in_json = false
       return_val = yield
       ActiveRecord::Base.include_root_in_json = saved
       return_val
    end

    # return native db representation of current utc time
    def translated_now
      sanitize_sql_array [":now", {:now => Time.now.utc}]
    end

    def truncate
      sql = "TRUNCATE TABLE #{quoted_table_name} "
      connection.execute(sql)
    end

    def db_time_now
      time_string = ActiveRecord::Base.connection.select_one('select utc_timestamp()').values.first
      DateTime.parse(time_string)
    end

    #make protected method publicly available
    def public_sanitize_sql(sql_array_or_string)
      return self.sanitize_sql(sql_array_or_string)
    end

    #make protected method publicly available
    def public_sanitize_sql_for_conditions(condition)
      sanitize_sql_for_conditions(condition)
    end

    # Defines specialized getter and setter methods for specifying instance
    # methods for maintaining the state of a toggle_section.
    def toggle_state_field(*args)
      args.each do |state_field|
        define_method state_field do
          instance_variable_get "@#{state_field}"
        end

        define_method "#{state_field}=" do |value|
          instance_variable_set "@#{state_field}", (value == "true" ? true : false)
        end
      end
    end

    def toggle_state_field_with_validation(state_field, validated_fields)
      # Getter for state field.  We want to return true (open toggle section) if
      # the state field is true or if any active record columns in the section have errors.
      define_method state_field do
        state = instance_variable_get "@#{state_field}"
        validated_fields.each do |validated_field|
          state = state || errors.on(validated_field)
        end
        state
      end

      # Setter for state field
      define_method "#{state_field}=" do |value|
        instance_variable_set "@#{state_field}", (value == "true" ? true : false)
      end
    end

    def sql_in_clause(array, place_holder = '--place-holder--')
      unquoted_array = array.empty? ? [place_holder] : array
      quoted_array = unquoted_array.map { |element| "'#{element}'" }
      "(#{quoted_array.join(', ')})"
    end

    def paginated_records_get(options, params)
      _page_number = params[:page].to_i
      # fix page number if invalid
      if _page_number == 0
        _page_number = 1
      end
      _per_page = params[:per_page].to_i
      if _per_page < 1
        _per_page = DEFAULT_PAGE_SIZE
      end

      _result = paginate(
        {
          :conditions => options[:conditions],
          :order      => options[:order],
          :page       => _page_number,
          :per_page   => _per_page,
        }
        )

      # fix page number if too high
      if _page_number > _result.total_pages
        if _result.total_pages == 0
          _page_number = 1
        else
          _page_number = _result.total_pages
        end
        # fix in our params
        params[:page] = _page_number.to_s
        # reretrieve records
        _result = paginate(
          {
            :conditions => options[:conditions],
            :order      => options[:order],
            :page       => _page_number,
            :per_page   => options[:per_page],
          }
          )
      end
      return _result
    end

    # Duplicate of function defined in ApplicationHelper - need to move to lib
    # clone parameter to protect against inadvertently changing the original
    # also protect against nil
    def param_clone(params, key)
      if params[key].nil?
        return nil
      else
        return params[key].clone
      end
    end

    ## generate a SQL 'ORDER BY' clause
    #def param_sort_clause(params, prefix, default_clause = '')
    #  _sort_clause = default_clause.clone
    #
    #  # clone anything we get from params[] to avoid inadvertently changing the original
    #  _param_sort_field = param_clone(params, (prefix + '_sort').to_sym)
    #  _param_sort_order = param_clone(params, (prefix + '_order').to_sym)
    #
    #  unless _param_sort_field.nil? || _param_sort_field == ""
    #    _sort_clause = AsperaDB.field_name_sanitize(_param_sort_field)
    #    if _param_sort_order == "d"
    #      _sort_clause << " DESC"
    #    end
    #  end
    #
    #  return _sort_clause
    #end

    # def human_attribute_name(attribute)
    #   (const_defined?(:HUMANIZED_ATTRIBUTES) && const_get(:HUMANIZED_ATTRIBUTES)[attribute.to_sym]) || super
    # end

    def validates_ip_address(attribute, options = {})
      options.merge!(:message => "is not a valid ip address")
      options.reverse_merge!(:with => Console::VALID_IP_REGEX)
      validates_format_of attribute, options
    end

    def validates_port(attribute, options = {})
      vno_options = options.reverse_merge({
          :greater_than_or_equal_to => 1,
          :less_than_or_equal_to => 65535,
      }).merge({
          :only_integer => true,
      })

      unless options[:message]
        first = vno_options[:greater_than_or_equal_to]
        last = vno_options[:less_than_or_equal_to]
        vno_options[:message] = "must be an integer between #{first}-#{last}"
      end

      validates_numericality_of attribute, vno_options
    end

    # Issues sql locks for _tables_.  If block given will lock, execute block, and ensure release
    # of the locks.
    def lock_tables_for_write(*tables, &block)
      tables = [tables].flatten
      sql = "lock tables " + tables.map { |e| "#{e} write" }.join(", ")
      connection.execute sql
      if block_given?
        begin
          yield
        ensure
          unlock_tables
        end
      end
    end

    # Unlocks any tables that have been locked.
    def unlock_tables
      sql = "unlock tables"
      connection.execute sql
    end


  end

end
