require 'csv'

module NodeConfiguration
 class AsconfiguratorError < StandardError; end
 class AdminUserDocrootedError < StandardError; end
end

module FaspNode::NodeConfiguration

  module ClassMethods

    # Various strings that might be returned when we attempt to run a command on a node
    # that doesn't support the command.
    def not_found_messages
      @not_found_messages ||= [
        'not found',
        'No such file or directory',
      ]
    end

  end

  module InstanceMethods

    # Returns +true+ if node is configurable.
    def is_configurable?
      config_login && config_login.connectable?
    end

    def remove_credentials
      Login.delete_all :owner_type => "NodeConfig", :owner_id => id
    end

    def remove_node_api_credentials
      login = Login.find(:first, :conditions => { :owner_type => "NodeAPIConfig", :owner_id => id })
      login.destroy if login
    end

    # legacy node only - update node api info based on 'Node API' subtab
    def update_api_credentials(params)
      raise 'Method only applies to legacy nodes' unless legacy_node
      self.node_api_username = params[:node_api_username]
      self.node_api_password = params[:node_api_password]
      self.v3_node_daemon_port = params[:v3_node_daemon_port]
      self.v3_node_use_ssl = params[:v3_node_use_ssl]

      if node_api_login_fields_valid?
        ActiveRecord::Base.transaction do
          # test Node API
          begin
            test_node_api_credentials(params[:node_api_username], params[:node_api_password], params[:v3_node_daemon_port], params[:v3_node_use_ssl], false)
          rescue Exception => e
            errors.add_to_base("Failed to connect via Node API: #{e.message}")
            return false
          end
          update_node_api_login
          update_attribute(:node_api_connected, true)
          update_attribute(:v3_node_use_ssl, params[:v3_node_use_ssl])
        end
      end
    end

    # Updates config id/pw - both ssh and node API (unless legacy node, which only updates ssh)  Won't save unless can connect to node.
    def update_credentials(params)
      self.attributes = params
      if config_login_fields_valid? && (legacy_node || node_api_login_fields_valid?)
        ActiveRecord::Base.transaction do
          # Only test SSH credentials if present
          if params[:configuration_login].present?
            # test SSH
            begin
              test_config_credentials(params[:configuration_login], params[:configuration_password], params[:configuration_ssh_key_id])
                # if we successfully connected, try to get asws info
                #query_node
            rescue Exception => e
              errors.add_to_base("Failed to connect via SSH: #{e.message}")
              return false
            end
          end

          unless legacy_node
            # test Node API
            begin
              test_node_api_credentials(params[:node_api_username], params[:node_api_password], v3_node_daemon_port, v3_node_use_ssl, v3_validate_ssl_cert)
            rescue Exception => e
              errors.add_to_base("Failed to connect via Node API: #{e.message}")
              return false
            end

            if !version_at_least?(FaspNode::REQUIRED_NODE_VERSION)
              errors.add_to_base("Node must be at least version #{FaspNode::REQUIRED_NODE_VERSION} (Node is using version #{formatted_version_tuple})")
              return false
            end
          end

          update_config_login unless params[:configuration_login].blank?
          update_node_api_login unless legacy_node
        end

        if params[:configuration_login].blank? && config_login.present?
          destroy_config_login
        end

        return true
      else
        return false
      end
    end

    def update_credentials_without_validation(params)
      self.attributes = params
      update_config_login
      update_node_api_login unless legacy_node
    end

    def update_credentials!(params)
      if !update_credentials(params)
        raise errors.full_messages.first
      end

      true
    end

    def update_v3_port
      node_api_login.port = v3_node_daemon_port
      node_api_login.save
    end

    def update_v3_node_password
      node_api_login.attributes = {
        :owner_type => "NodeAPIConfig",
        :owner_id => id,
        :address => node_address,
        :port => v3_node_daemon_port,
        :password => v3_node_password
      }
      node_api_login.save
    end

    def update_node_api_login
      node_api_login.attributes = {
          :owner_type => "NodeAPIConfig",
          :owner_id => id,
          :address => node_address,
          :port => v3_node_daemon_port,
          :login => node_api_username,
          :password => node_api_password,
          :ssh_key_id => nil
      }
      node_api_login.save!
    end

    def update_config_login
      config_login.attributes = {
        :owner_type => "NodeConfig",
        :owner_id => id,
        :address => node_address,
        :port => ssh_port,
        :login => configuration_login,
        :password => configuration_password,
        :ssh_key_id => configuration_ssh_key_id
      }
      config_login.save
      # sync_ssh_tunnels
    end

    def destroy_config_login
      config_login.destroy
    end

    def node_api_login_fields_valid?

      if node_api_username.blank?
        errors.add(:node_api_username, "^Node API username can't be blank")
      end

      if !node_api_username.blank? && node_api_username.length > 60
        errors.add(:node_api_username, "^Node API username must be 60 characters or less")
      end

      if node_api_password.blank?
        errors.add(:node_api_password, "^Node API password required.")
      end

      errors.empty?
    end

    def config_login_fields_valid?
      return true if atc?

      # ensure only console ssh keys are used.
      unless configuration_ssh_key_id.blank?
        self.configuration_ssh_key_id = SshKey.console.find(configuration_ssh_key_id).id
      end

      #if configuration_login.blank?
      #  errors.add(:configuration_login, "^Admin login is required")
      #end

      if !configuration_login.blank? && configuration_login.length > 60
        errors.add(:configuration_login, "^Admin Login must be 60 characters or less")
      end

      if !configuration_login.blank? && configuration_password.blank? && configuration_ssh_key_id.blank?
        errors.add_to_base("Password or SSH Key is required.")
      end

      if !configuration_password.blank? && !configuration_ssh_key_id.blank?
        errors.add_to_base("Enter a Password OR SSH Key (not BOTH).")
      end

      #if node_api_username.blank?
      #  errors.add_to_base("Node API login is required")
      #end
      #
      #if node_api_password.blank?
      #  errors.add_to_base("Node API password is required")
      #end

      errors.empty?
    end

    def test_node_api_login
      api_login = node_api_login
      test_node_api_credentials(api_login.login, api_login.unencrypted_password, api_login.port, v3_node_use_ssl, v3_validate_ssl_cert)
    end

    def test_node_api_credentials(login, password, port, use_ssl, require_signed_ssl)
      response=nil
      # AC - 827 : To Check for Console Proxy and retrieve Proxy Details
      proxy_fields = ConsoleConfig::ConsoleProxy.retrieve_proxy_data("http_proxy", node_address, ConsoleConfig::ConsoleProxy::PROXY_SOURCE[5])

      if proxy_fields
        http = Net::HTTP.new(node_address, port, proxy_fields[:host], proxy_fields[:port], proxy_fields[:user], proxy_fields[:pass])
      else
        http = Net::HTTP.new(node_address, port)
      end

      if use_ssl
        http.use_ssl = use_ssl
        http.verify_mode = require_signed_ssl ? OpenSSL::SSL::VERIFY_PEER : OpenSSL::SSL::VERIFY_NONE
      end

      req = Net::HTTP::Get.new('/info')
      req.basic_auth login, password

      req['accept'] = 'application/xml'
      response = nil
      begin
        response = http.start { |request| request.request(req) }
      rescue Exception => e
        logger.error "node api credentials failure: #{e.message}"
        raise
      end

      raise "Blank response from node" unless response

      raise "Bad login or password" if response.code == "401"
      raise "Unexpected response code from node: #{response.code}" unless response.code == "200"

      parsed = ActiveSupport::JSON.decode(response.body)
      if parsed['error']
        raise parsed['error']['user_message']
      end

      update_attribute(:reported_logger_version, parsed['version'])

      if !self.legacy_node? && self.version_at_least?('3.6')
        acls = parsed['acls']
        unless atc? || (acls.present? && acls.include?('impersonation'))
          raise "Node user impersonation ACL not set on node"
        end
      end

      #update node uuid / os / version
      # For clusters, use cluster_id instead of node_id
      if self.atc?
        update_attribute(:node_uuid, parsed['cluster_id'])
      else
        update_attribute(:node_uuid, parsed['node_id'])
      end
      update_attribute(:reported_hostname, self.node_address)
      update_attribute(:reported_os, parsed['os'].to_s.split(' ')[0])
      update_attribute(:reported_os_version, parsed['os'])
      update_attribute(:reported_central_uri, 'http://0.0.0.0:40001')
      update_attribute(:license_max_rate, parsed['license_max_rate'])

      new_date = parsed['license_expiration_date'] != 'none' ? Time.parse("#{parsed['license_expiration_date']} 23:59:59Z") : nil
      update_attribute(:license_expiration_date, new_date)

      true
    end

    # Tests config credentials.  Raises exception with some-what friendly message.
    # See AsperaRemote::Connection#new_session() for details.
    def test_config_credentials(
        login = config_login.login,
        password = config_login.unencrypted_password,
        ssh_key_id = config_login.ssh_key_id)

      AsperaRemote::Connection.test!(node_address, login, :port => ssh_port, :password => password, :ssh_key_id => ssh_key_id, :encryption => ssh_encryption)
    end

    # Override setter so we can clear cached return value from #node_status()
    def asws_node_response=(v)
      write_attribute(:asws_node_response, v)
      @node_response = nil
    end

    def query_node
      response = run_remote('asws  --url  "/node"  --accept "application/xml"')
      update_attribute(:asws_node_response, response)
    end

    def node_status
      @node_response ||= AswsResponse::NodeQuery.new(asws_node_response)
    end

    def run_remote(command)
      credentials_copy = config_credentials.dup
      credentials_copy[2].merge!(:password => "[filtered]")
      logger.info "Running remote command #{command}"
      logger.debug "On #{credentials_copy.inspect}"
      AsperaRemote::Connection.start(*config_credentials) do |ssh|
        response = ssh.exec!(command)
        logger.debug "Response: #{response}"
        return response
      end
    rescue
      logger.error "Run remote error: #{$!}\n#{$!.backtrace.join("\n")}"
      raise
    end

    def both_logins_exist?
      config_login_exists? && node_api_login_exists?
    end

    def config_login_exists?
      Login.find_by_owner_type_and_owner_id("NodeConfig", id)
    end

    # Returns an instance of Login used to connect to node for node configuration.
    def config_login
      @config_login ||= Login.find_or_initialize_by_owner_type_and_owner_id("NodeConfig", id)
    end

    def node_api_login_exists?
      Login.find_by_owner_type_and_owner_id("NodeAPIConfig", id)
    end

    def node_api_login
      @n_api_login ||= Login.find_or_initialize_by_owner_type_and_owner_id("NodeAPIConfig", id)
    end

    def configuration_ssh_key_id
      config_login.ssh_key_id
    end

    def configuration_ssh_key_id=(id)
      config_login.ssh_key_id = id
    end

    def configuration_ssh_key_name
      config_login.ssh_key_name
    end

    def node_api_username
      node_api_login.login
    end

    def node_api_username=(login)
      node_api_login.login = login
    end



    # Returns the configuration login username.
    def configuration_login
      config_login.login
    end

    # Sets the configuration login username.
    def configuration_login=(login)
      config_login.login = login
    end

    def user_configs
      get_user_list.map { |login| user_config(login) }
    end

    def group_configs
      get_group_list.map { |group| group_config(group) }
    end

    def vlink_configs
      get_vlink_list.map { |trunk_id| vlink_config(trunk_id) }
    end

    def user_config(login)
      @user_config ||= {}
      @user_config[login] ||= ::NodeConfiguration::Level::User.new(self, login)
    end

    def group_config(name)
      @group_config ||= {}
      @group_config[name] ||= ::NodeConfiguration::Level::Group.new(self, name)
    end

    def box_config
      @box_config ||= ::NodeConfiguration::Level::Box.new(self)
    end

    def vlink_config(trunk_id)
      @vlink_config ||= {}
      @vlink_config[trunk_id] ||= ::NodeConfiguration::Section::Vlink.new(self, trunk_id)
    end
  end

  def get_user_list_command
    [
        'get_user_list',
        '"source", "conf"',
    ].join("\n")
  end

  def get_user_data_command(user)
    [
        'get_user_data',
        '"user_name", "' + user + '"',
    ].join("\n")
  end

  def get_group_list_command
    [
        'get_group_list',
        '"source", "conf"',
    ].join("\n")
  end

  def get_group_data_command(group)
    [
        'get_group_data',
        '"group_name", "' + group + '"',
    ].join("\n")
  end

  def get_vlink_list_command
    [
        'get_trunk_list',
        '"source", "conf"',
    ].join("\n")
  end

  def get_vlink_data_command(id)
    [
        'get_trunk_data',
        '"id", "' + id + '"',
    ].join("\n")
  end

  def get_node_data_command
    'get_node_data'
  end

  def get_db_data_command
    'get_database_data'
  end

  def get_server_data_command
    'get_central_server_data'
  end

  def get_http_server_data
    'get_http_server_data'
  end

  def get_node_config_command
    [
      'get_all',
      Console.uac_version,
      'option_mask,"!get_uac_spec"',
      'option_mask,"!*default*"',
      'option_mask,"*"',
      'user_options',
      'trunk_mask,""',
      'end_list',
      'group_options',
      'trunk_mask,""',
      'end_list',
      'node_options',
      'trunk_mask,""',
      'end_list',
    ].join("\n")
  end

  def restart_central
    command = [
      'app_restart',
      '"app","asperacentral"',
      '',
      "terminate\n\n",
    ].join("\n")
    asconfigurator(command)
  end


  def get_user_list
    raw_user_list = asconfigurator(get_user_list_command)

    lines = raw_user_list.split(' ')
    users = []

    for line in lines
      if (line =~ /user_name/)
        user_name = line[/\"(.*?)\"/m, 1]
        users.push(user_name)
      end
    end

    users || []
  end

  def get_group_list
    raw_group_list = asconfigurator(get_group_list_command)

    lines = raw_group_list.split(' ')
    groups = []

    for line in lines
      if (line =~ /group_name/)
        group_name = line[/group_name\",\"(.*?)\"/m, 1]
        groups.push(group_name)
      end
    end

    groups || []
  end

  def get_vlink_list
    # cache this so we don't retrieve multiple times in a single call
    return @cached_get_vlink_list unless @cached_get_vlink_list.nil?

    raw_vlink_list = asconfigurator(get_vlink_list_command)

    lines = raw_vlink_list.split(' ')
    @cached_get_vlink_list = []

    for line in lines
      if(line =~ /trunk_id/)
        link_name = line[/trunk_id\",\"(.*?)\"/m, 1]
        @cached_get_vlink_list.push(link_name)
      end
    end

    @cached_get_vlink_list || []
  end


  # Reads configuration from node and stores parsed results in +settings+ table.
  def load_config_settings
    read_settings_from_node
    ActiveRecord::Base.transaction do
      delete_settings_from_console_db
      load_node_settings_to_console_db
    end
#    @box_config = nil
  end

  def load_user_config_settings(user)
    read_user_settings_from_node(user)
    ActiveRecord::Base.transaction do
      delete_settings_from_console_db
      load_node_settings_to_console_db
    end
#    @box_config = nil
  end

  def load_group_config_settings(group)
    read_group_settings_from_node(group)
    ActiveRecord::Base.transaction do
      delete_settings_from_console_db
      load_node_settings_to_console_db
    end
#    @box_config = nil
  end

  def raw_config_response_all_vlinks
    result = ""
    vlinks = get_vlink_list

    for id in vlinks
      result += asconfigurator(get_vlink_data_command(id))
    end
    result
  end


  def load_all_vlink_config_settings
    self.raw_config_response = raw_config_response_all_vlinks

    ActiveRecord::Base.transaction do
      delete_settings_from_console_db
      load_node_settings_to_console_db
    end
#    @box_config = nil
  end

  def load_node_settings_to_console_db
    sql_start = %(insert into settings(#{sql_column_names}) values )
    response_rows_of_interest.in_groups_of(5000, false) do |group|
      sql = sql_start + group.map { |row| "(#{sql_column_values(row)})" }.join(",\n")
      connection.execute(sql)
    end
  end

  def sql_column_names
    @sql_column_names ||= %w(fasp_node_id owner name override_value overridden inherited_value inherited_source).join(', ')
  end

  def sql_column_values(row)
    p = ::NodeConfiguration::SettingParser::Base.new_parser(row)
    [
      id,
      q(p.owner),
      q(p.name),
      q(p.override_value),
      (p.overridden ? 1 : 0),
      q(p.inherited_value),
      q(p.inherited_source)
    ].map { |value| value.to_s }.join(',')
  end

  def q(string)
    string.nil? ? 'NULL' : "'#{Mysql.escape_string(string)}'"
  end


  def response_rows_of_interest
    parsed_raw_response.select do |row|
      row.size >= 3                                &&
      row[1] !~ /in_trunk|out_trunk/               &&
      row[2] !~ /in_trunk|out_trunk/               &&
      row[1] != "file_exclude"                     &&
      row[2] != "file_exclude"                     &&
      true
    end
  end

  def parsed_raw_response
    # raise raw_config_response
    raw_config_response.split("\r\n").each_with_index{|r, i| logger.info "#{i}: #{r}"}
    CSV.parse(raw_config_response)
  end

  def delete_settings_from_console_db()
    Setting.delete_all(["fasp_node_id = ? AND name != ?", id, 'node_api'])
  end

  def read_settings_from_node
    self.raw_config_response = asconfigurator(get_node_data_command)
    self.raw_config_response += asconfigurator(get_db_data_command)
    self.raw_config_response += asconfigurator(get_http_server_data)
    self.raw_config_response += asconfigurator(get_server_data_command)
    self.raw_config_response += raw_config_response_all_vlinks
  end

  def read_user_settings_from_node(user)
    self.raw_config_response = asconfigurator(get_user_data_command(user))
    self.raw_config_response += raw_config_response_all_vlinks
  end

  def read_group_settings_from_node(group)
    self.raw_config_response = asconfigurator(get_group_data_command(group))
    self.raw_config_response += raw_config_response_all_vlinks
  end

  def write_settings_to_node
    asconfigurator(box_config.set_node_config_command)
  end

  def unencrypted_v3_node_api_password
    node_api_login.unencrypted_password
  end

  def config_credentials
    options = {
      :password   => config_login.unencrypted_password,
      :port       => ssh_port,
      :ssh_key_id => configuration_ssh_key_id
    }
    [node_address, config_login.login, options]
  end

  # Connect to node and run _command_, returning the results from STDOUT.
  def asconfigurator(command)
    response = ''
    AsperaRemote::Connection.start(*config_credentials) do |ssh|
      channel = ssh.open_channel do |channel|
        channel.exec "asconfigurator" do |ch, success|
          logger.info "sending\n======\n#{command_reload_terminate(command)}\n===="
          channel.send_data(command_reload_terminate(command))

          channel.on_data do |ch, data|
            response << data
          end

          channel.on_extended_data do |ch, type, data|
            asconfigurator_error(data)
          end
        end
        channel.wait
        ssh.loop
      end
    end
    logger.info "Response:\n=====\n#{response}\n====="
    set_central_restart_needed_flag(command, response)
    check_response_for_errors(response)
    response
  # rescue Timeout::Error
  #   raise "Timeout while trying to contact #{node_address}"
  end

  # Returns complete asconfigurator command with reload (if needed) and terminate included.
  def command_reload_terminate(command)
    return command if command.include?("\nterminate\n")
    r = [command]
    r << reload_command if true && update_command?(command)
    r << "terminate"
    r.join("\n\n") + "\n\n"
  end

  # Returns +true+ if _command_ is an update (set or delete) command.
  def update_command?(command)
    command =~ /^(set|delete)/
  end

  # Returns the string needed to reload/restart http fallback and central.
  def reload_command
    [
      %(app_reload),
      Console.uac_version,
      %("app","asperahttpd"),
      %("app","central")
    ].join("\n")
  end

  # Raises exception when data is written to STDERR by asconfigurator.
  def asconfigurator_error(data)
    return if data.include?("Could not chdir to home directory") # windows user never logged in to account
    if asconfigurator_not_installed?(data)
      raise NodeConfiguration::AsconfiguratorError, "Node must be running Aspera 2.2 or higher to do remote configuration."
    else
      raise NodeConfiguration::AsconfiguratorError, "Error running asconfigurator: #{data}"
    end
  end

  def asconfigurator_not_installed?(data)
    # not_installed_messages = [
    #   'asconfigurator: command not found', # windows linux
    #   'command not found: asconfigurator', # isilon
    #   'asconfigurator: not found',         # solaris
    # ]
    FaspNode.not_found_messages.any? { |m| data =~ /#{m}/ }
  end

  def check_response_for_errors(response)
    self.asconfigurator_errors = []
    check_response_for_docrooted_user_message(response)
    response.split("\n").each_with_index do |line, i|
      if line.starts_with?('error')
        if defined?(FasterCSV)
          level, code, what, why = FasterCSV.parse_line(line)
        else
          level, code, what, why = CSV.parse_line(line)
        end
        self.asconfigurator_errors << "#{what}: #{why}"
      end
    end
    asconfigurator_errors.uniq!
  end

  def check_response_for_docrooted_user_message(response)
    docrooted_user_messages = [
      'Command not accepted: asconfigurator', # windows
    ]
    if docrooted_user_messages.any? { |m| response =~ /#{m}/ }
      raise NodeConfiguration::AdminUserDocrootedError
    end
  end

  def set_central_restart_needed_flag(command, response)
    if update_command?(command) || command.include?('app_restart')
      update_attribute(:central_restart_needed, false)
      if response.include?('error,14,')
        update_attribute(:central_restart_needed, true)
      end
    end
  end

  def settings_for_set(owner)
    config_settings(true).select { |setting| setting.owner == owner && ok_to_send?(setting) }
  end

  # We don't want to send any Docroot rows that only represent defaults from
  # a higher level.  Assumption is that not sending rows with NULL override
  # doesn't break anything.
  # AC-683, not sending logging params to set_node_data
  def ok_to_send?(setting)
    excludeParams = ["classes","level","log_size","directory","log_count"]
    (!(setting.name.starts_with?("Docroot #") && setting.override_value.nil?)) && (!excludeParams.include?setting.name)
  end

  def self.included(receiver)
    receiver.extend         ClassMethods
    receiver.send :include, InstanceMethods
  end
end
