
require File.expand_path(File.dirname(__FILE__))+"/lib/asctl_env.rb"
require 'fileutils'
require 'asctl'
require 'asctl/exit_codes'
require 'ext/platform'

def windows?
  Platform.name == :windows
end

# Copied the guts of Controller.running_as_root?() because 
# we can't actually load Controller if we're not root.
unless windows?
  unless `id` =~ Regexp.new('^'+Regexp.escape('uid=0(root)'))
    puts "Only the 'root' user is allowed to use asctl"
    exit(Asctl::ExitCodes::USER_IS_NOT_ROOT)
  end
end

require 'asctl/log'
if defined?(Debugger)
  def log(*args)
    puts(*args)
  end
end
require 'asctl/fileutils'
require "object_factory"
require "pp"
require "ext/core"
require "aspera/version"

# FileUtils is included here as a convenience because we call its methods
# all over the place.  The Asctl::FileUtils is different from FileUtils
# in that it logs to a file then calls the FileUtils function.
include Asctl::FileUtils

class AsctlControl
  attr_accessor :exit_code
  @@verbose     = false
  @@force       = 0
  @@debug       = 0
  @@depth_first = true
  @@reload_component = nil

  def initialize( argv = [] )
    # Assume we'll get an error - but we don't know what kind yet
    @exit_code = Asctl::ExitCodes::UNKNOWN_ERROR
    @args = argv.dup
    log "==="
    log "=== %s: '%s'" % [Time.now, @args.join(' ')]
    log "==="

    if extract_arg(true,false,["--version"])
      puts "asctl version "+version
      exit(0)
    end
    @@verbose     = extract_arg(true,false,["-v", "--trace", "--verbose"])
    @@nolog       = extract_arg(true,false,["--no-log"])
    @@force       = extract_arg(1,0,["-f"])
    @@debug       = extract_arg(1,0,["-d"])
    @@depth_first = extract_arg(false,true,["-b"])
    @@reload_component = extract_arg(nil, nil, ["--reload-component=", '-c='])
    # TODO: eventually replace these globals
    $debug, $force, $verbose, $depth_first, $nolog, $reload_component = @@debug, @@force, @@verbose, @@depth_first, @@nolog, @@reload_component

    begin
      load_components

      # Only print help after loading all the components so we have something to say
      if @args.empty?
        print_full_usage
      else
        # Top-level command delegation
        command_segments = @args[0].downcase.split(/:/)   # e.g. "faspex:ALL:Start" => [faspex,all,start]
        command_arguments = @args[1..-1]
        specified_component = find_component(command_segments.first)
        if specified_component
          specified_component.execute( command_segments[1..-1], command_arguments )
          # Assume any error cases will throw an exception - so if we get here we were successful.
          @exit_code = Asctl::ExitCodes::SUCCESS
        else
          puts "Unrecognized component '#{command_segments.first}'. Valid component choices:"
          component_names.sort.each do |c_name|
            puts "  #{c_name}"
          end

          @exit_code = Asctl::ExitCodes::UNRECOGNIZED_COMPONENT
        end
      end
    rescue ArgumentError
      puts $!.to_s
      log_error($!)
      @exit_code = Asctl::ExitCodes::ARGUMENT_ERROR
    rescue YmlLoadError
      # TODO: verify we can actually trigger this
      puts $!.to_s
      log_error($!)
      puts "You may either"
      puts "  - fix the error or"
      puts "  - delete this configuration file and re-configure"
      @exit_code = Asctl::ExitCodes::EXCEPTION_LOADING_YML
    rescue Exception
      # It was a requirement that asctl not print stack traces
      # after an error (under normal use).  So we log it instead:
      puts $!.to_s
      log_error($!)
      @exit_code = Asctl::ExitCodes::EXCEPTION
    end
  end

  def find_component(name)
    log "Searching for '#{name}'..."
    @components.each do |component|
      log "#{component.name} (#{component.cmd})"
      return component if component.cmd == name
    end
    return nil
  end

  def component_names
    @components.collect {|c| c.cmd }
  end

  def self.force
    @@force
  end
  def self.debug
    @@debug
  end
  def self.depth_first
    @@depth_first
  end

  def self.verbose
    @@verbose
  end
  def self.verbose= value
    @@verbose = value
  end
  def self.nolog
    @@nolog
  end
  def self.without_verbosity
    v = @@verbose; @@verbose = false
    begin
      yield
    ensure
      @@verbose = v
    end
  end

  def self.reload_component= value
    @@reload_component = value
  end

  def self.reload_component
    @@reload_component
  end

  def extract_arg(found_value, not_found_value, triggers)
    triggers.each do |flag|
      if @args.include?(flag)
        @args.delete(flag)
        return found_value
      elsif !(flag =~ /[=\s]/).nil?
          @args.each do |arg|
            next if arg.nil? || !arg.include?(flag)
            matches = arg.match(/#{flag}[^\s]+/)[0].partition(/[=\s]/)
            next if matches.nil?
            found_value = matches[2]
            @args.delete(arg)
            return found_value
          end
          
      end
    end
    return not_found_value
  end

  def version
    data = ""
    fn = File.dirname(__FILE__)+"/VERSION"
    begin
      data = File.read(fn)
    rescue => e
      puts "Exception attempting to read #{fn}: #{e}"
    end
    return data
  end

  ASPERA_ROOT = File.dirname(__FILE__) + (windows? ? "/../../../../" : "/../../")

  # These are components that are *outside* of the Common area (since the common ones will all be updated together)
  VERSIONED_COMPONENTS = {
      :console => {:minimum_acceptable_version => "1.5.0.0",
                   :version_file_path => "config/version.txt",
                   :windows_path => "Aspera/Management Console",
                   :linux_path => "console"
      },
      :orchestrator => {:minimum_acceptable_version => "2.0.0",
                        :version_file_path => "config/version.txt",
                        :windows_path => "Aspera/Orchestrator",
                        :linux_path => "orchestrator"
      },
      :faspex =>  {:minimum_acceptable_version => "1.7.0.0",
                   :version_file_path => "config/version.txt",
                   :windows_path => "Aspera/Faspex",
                   :linux_path => "faspex"
      },
  }

  def component_path(hash)
    if windows?
      hash[:windows_path]
    else
      hash[:linux_path]
    end
  end

  def load_components
    # We always load our own files
    @components = []
    found_component_files =  Dir.glob(File.dirname(__FILE__)+"/../../common/*/*_controller.rb") #mysql and apache
    found_component_files += Dir.glob(File.dirname(__FILE__)+"/../../common/all_controller.rb") #all_controller

    # Load any non-common components that are up to the acceptable version level.
    # # "VERSIONED_COMPONENTS" is how we handle things that might be out of sync with asctl.
    # # mysql and apache are installed with asctl, so their asctl files cannot be out of sync.
    VERSIONED_COMPONENTS.each do |application,hash|
      version = Aspera::Version.from_file(ASPERA_ROOT + component_path(hash) + "/"+hash[:version_file_path])
      if version > hash[:minimum_acceptable_version]
        found_component_files += Dir.glob(ASPERA_ROOT + component_path(hash) + "/*_controller.rb")
      end
    end

    found_component_files.map!{ |f|
      File.expand_path(f)
    }
    found_component_files.uniq.each { |f|
      require f
      k = ObjectFactory.class_name(f)
      puts "Loading from #{f}" if $debug > 0
      @components << ObjectFactory.load(Object.const_get(k))
      # if we want to support calling a controller's dependencies
      # from the top level like "asctl ad_background:stop" instead
      # of "asctl faspex:ad_background:stop"
      # @components += @components.last.dependencies if @components.last.respond_to?(:dependencies)
    }
    @components.uniq!

    log "components (#{@components.length}): #{@components.collect {|c| c.class }.join(', ')}"
  end

  def print_full_usage
    @components.each do |component|
      component.show_help_messages
    end
  end

  at_exit do
    if !@@reload_component.nil?
      ObjectFactory.reload_one(@@reload_component)
    end
    
    ObjectFactory.save_all
    $stdout.flush
    $stderr.flush
  end
end
asctl_control = AsctlControl.new(ARGV)
exit( asctl_control.exit_code )
