require 'rubygems'
require 'yaml'
require 'mysql'
require 'db_encrypt_decrypt.rb'

include DbEncryptDecrypt

class AsperaDbSpaceWatcher
  
  DEBUG = false
  WATCH_TABLE_NAME='aspera_db_disk_space_free'
  MYSQL_SCHEMA_DB = 'information_schema'
  MYSQL_USERNAME = 'root'
  SLEEP_TIME_SECONDS = 60

  def aspera_database_config_reload
    @aspera_database_config = YAML.load_file(aspera_database_yaml_file_path)
  end

  def aspera_database_yaml_file_path
    File.join(mysql_base_path, 'database.rb.yml')
  end

  def databases_to_watch
    database_names = []
    result_set = @mysql_connection.query(sql_to_find_databases_to_watch)
    result_set.each do |row|
      database_names << row[0]
    end
    database_names
  end

  def db_path(db_name)
    print_if_debug{File.join(@mysql_data_dir, db_name)}
  end

# takes a path in standard Ruby format (forward slashes)
  def free_space_in_bytes(path)
    if windows?
      #replace forward slashes with backslashes
      win_path = path.gsub('/','\\')
      #wrap in quotes in case there are spaces
      win_path = %Q["#{win_path}"]
      command_line = print_if_debug{"dir /-C #{win_path}"}
      space_free_line = print_if_debug{`#{command_line}`.split("\n")[-1]}
      bytes_free = print_if_debug{space_free_line.match(/(\d+) bytes free/)[1]}
      if bytes_free.nil?
        nil
      else
        bytes_free.to_i
      end
    else
      #wrap in quotes in case there are spaces
      linux_path = %Q["#{path}"]
      command_line = print_if_debug{"df -k -P #{linux_path}"}
      space_free_line = `#{command_line}`.split("\n")[-1]
      kb_free = space_free_line.match(/^.+? +\d+ +\d+ +(\d+) +\d+% +.+?/)[1]
      if kb_free.nil?
        nil
      else
        kb_free.to_i * 1024
      end
    end
  rescue
    print_if_debug{"Exception: #{$!} #{$@.join("\n")}"}
    nil
  end

  def main_loop
    print_if_debug{"\n===================================== main_loop =====================================\n"}
    reload_cached_items
    databases_to_watch.each do |db_name|
      begin
        update_watch_table(db_name, free_space_in_bytes(db_path(db_name)))
      rescue
        # nothing, just protect against failure of one update to prevent others
        puts $! if DEBUG
      end
    end
  rescue
    # nothing, just protect against failure to connect to MySQL
    puts $! if DEBUG
  end

  def aspera_base_path
    if windows?
      # list of directories to test
      test_dirs = [
          File.join('C:','Program Files (x86)','Common Files','Aspera'),
          File.join('C:','Program Files','Common Files','Aspera'),
          File.join('D:','Program Files (x86)','Common Files','Aspera'),
          File.join('D:','Program Files','Common Files','Aspera')
      ]
      test_dirs.find {|dir| File.directory?(dir)}
    else
      File.join('opt','aspera')
    end
  end

  def mysql_base_path
    if windows?
      # list of directories to test
      test_dirs = [
          File.join('C:','Program Files (x86)','Common Files','Aspera','Common','mysql'),
          File.join('C:','Program Files','Common Files','Aspera','Common','mysql'),
          File.join('D:','Program Files (x86)','Common Files','Aspera','Common','mysql'),
          File.join('D:','Program Files','Common Files','Aspera','Common','mysql')
      ]
      test_dirs.find {|dir| File.directory?(dir)}
    else
      File.join('opt','aspera','common','mysql')
    end
  end

  def mysql_config_file_path
    if windows?
      File.join(mysql_base_path, 'my.ini')
    else
      File.join(mysql_base_path, 'my.cnf')
    end
  end

  def mysql_connection_close
    @mysql_connection.close if @mysql_connection
    @mysql_connection = nil
  end

  def mysql_connection_reopen
    mysql_connection_close
    @mysql_connection = Mysql.real_connect('127.0.0.1', mysql_username, mysql_password, MYSQL_SCHEMA_DB, mysql_port)
  end

  def mysql_data_dir_reload
    regex = /^\s*datadir=(.+)\s?$/
    data_dir_line = File.readlines(mysql_config_file_path).find { |line| line.match(regex)}
    @mysql_data_dir = data_dir_line.match(regex)[1]
  end

  def mysql_password
    pass = @aspera_database_config[:password]
    db_ecrypt_decrypt = DbEncryptDecrypt.new(@aspera_database_config[:path_to_secret])

    if !@aspera_database_config[:path_to_secret] || !db_ecrypt_decrypt.encrypted?(pass)
      return pass
    end
    
    db_ecrypt_decrypt.decrypt(pass)
  end

  def mysql_port
    @aspera_database_config[:port]
  end

  def mysql_username
    MYSQL_USERNAME
  end

  def print_if_debug(&block)
    result = yield
    if DEBUG
      puts "#{Time.now}  -- #{caller.first.match(/`(.+?)'/)[1]}: #{result}"
      puts
    end
    result
  end

  def reload_cached_items
    aspera_database_config_reload
    mysql_data_dir_reload
    mysql_connection_reopen
  end

  def sql_to_find_databases_to_watch
    print_if_debug {%Q[SELECT TABLE_SCHEMA
FROM `#{MYSQL_SCHEMA_DB}`.TABLES
WHERE TABLE_NAME='#{WATCH_TABLE_NAME}'
  AND ENGINE="InnoDB"
  AND TABLE_TYPE="BASE TABLE"]}
  end

  def sql_to_update_watch_table(db_name, bytes_free)
    print_if_debug{%Q[INSERT INTO `#{db_name}`.`#{WATCH_TABLE_NAME}`
(id, bytes_free, last_reported_at)
VALUES (1, #{bytes_free}, UTC_TIMESTAMP())
ON DUPLICATE KEY UPDATE
bytes_free=#{bytes_free},
last_reported_at=UTC_TIMESTAMP()
]}
  end

  def update_watch_table(db_name, bytes_free)
    @mysql_connection.query(sql_to_update_watch_table(db_name, bytes_free)) if bytes_free
  end

  def windows?
    RbConfig::CONFIG["host"] =~ /win32|ming/
  end
end
